/**
 * @author R. Diez
 */

package main;

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

import vcdEngine.*;


class CompareVcd
{
    public static final String appName    = "comparevcd";
    public static final String appVersion = "alpha 0.12";
    
    private static final String COMPARE_VARIABLES_OPTION = "compare-variables";
    
    public enum OperationResultEnum
    {
        OPRES_CARRY_ON,  // Only used internally.
        
        OPRES_NO_DIFFERENCES,
        OPRES_DIFFERENCES_FOUND,
        OPRES_SUCCESS,
        OPRES_ERROR,
    }

    // This set automatically removes duplicate variable names.
    private TreeSet< String > vcdVariablesToCompare = new TreeSet< String >();
    boolean isComparingAllVariables = false;
    
    private boolean listVcdVariables = false;    

    private VcdFileParser vcdFile1;
    private VcdFileParser vcdFile2;
    
    private class VcdVariablePair
    {
        VcdVariableManager.VcdVariable v1;
        VcdVariableManager.VcdVariable v2;
    }
    
    private VcdVariablePair[] varCompTable;

    
    public OperationResultEnum main ( String[] args )
    {
        ArrayList<String> filenames = new ArrayList<String>();
        
        OperationResultEnum ec1 = processCommandLineArguments( args, filenames );
        
        if ( ec1 != OperationResultEnum.OPRES_CARRY_ON )
            return ec1;
        
        try
        {
            assert filenames.size() == 2 : myutils.ErrUtils.assertionFailed();
            
            vcdFile1 = new VcdFileParser( filenames.get( 0 ) );
            vcdFile2 = new VcdFileParser( filenames.get( 1 ) );
            
            System.out.printf( "Comparing files \"%s\" and \"%s\" ...\n",
                               vcdFile1.getFilenameForUserMessages(),
                               vcdFile2.getFilenameForUserMessages() );
            
            OperationResultEnum ec2 = compareVcdFiles();
            
            switch ( ec2 )
            {
            case OPRES_DIFFERENCES_FOUND:
                System.out.printf( "The files have different contents.\n");
                break;                
                
            case OPRES_NO_DIFFERENCES:
                if ( isComparingAllVariables )
                {
                    // Alternative message text: "The file contents match."
                    System.out.printf( "The files have the same contents.\n");
                }
                else
                {
                    System.out.printf( "The given variables have the same values in both files.\n");
                }
                break;
            
            default:
                assert false : myutils.ErrUtils.assertionFailed();
                break;
            }
            
            return ec2;
        }
        finally
        {
            if ( vcdFile1 != null )
                myutils.FileUtils.closeIgnoreErrors( vcdFile1 );
            
            if ( vcdFile2 != null )
                myutils.FileUtils.closeIgnoreErrors( vcdFile2 );
        }
    }


    private OperationResultEnum processCommandLineArguments ( String[] args, ArrayList<String> filenames )
    {
        ArrayList< LongOpt > longopts = new ArrayList< LongOpt >();
        
        longopts.add( new LongOpt( "help", LongOpt.NO_ARGUMENT, null, 'h' ) );
        longopts.add( new LongOpt( "license", LongOpt.NO_ARGUMENT, null, 2 ) );
        longopts.add( new LongOpt( "version", LongOpt.NO_ARGUMENT, null, 3 ) );
        longopts.add( new LongOpt( "list-variables", LongOpt.NO_ARGUMENT, null, 4 ) );
        longopts.add( new LongOpt( COMPARE_VARIABLES_OPTION, LongOpt.REQUIRED_ARGUMENT, null, 5 ) );
    
        Getopt g = new Getopt( appName, args, "-h", longopts.toArray( new LongOpt[longopts.size()] ) );
    
        int c;
    
        while ( (c = g.getopt()) != -1 )
        {
            switch ( c )
            {
            case 1:
                filenames.add( g.getOptarg() );
                break;
    
            case 2:
                printLicenseText();
                return OperationResultEnum.OPRES_SUCCESS;
    
            case 3:
                printVersion();
                return OperationResultEnum.OPRES_SUCCESS;
    
            case 4:
                listVcdVariables = true;
                break;
                
            case 5:
                processVarNames( g.getOptarg() );
                break;
    
            case 'h':
                printHelp();
                return OperationResultEnum.OPRES_SUCCESS;
    
            case '?':
                // getopt should have printed its own error message.
                return OperationResultEnum.OPRES_ERROR;
                
            default:
                assert false : myutils.ErrUtils.assertionFailed();
            
                // I'm not sure I've got it all covered, so always generate a user-visible error message.
                throw new RuntimeException( "Internal error parsing the command-line arguments." );
            }
        }
    
        // These are the command-line arguments after a possible "--" marker.
        for ( int i = g.getOptind(); i < args.length; i++ )
        {
            filenames.add( args[i] );
        }
    
        if ( filenames.size() != 2 )
            throw new RuntimeException( "Invalid number of arguments, exactly 2 filenames are required." );
        
        return OperationResultEnum.OPRES_CARRY_ON;
    }    
    
    
    private void processVarNames ( String varsStr )
    {
        // System.out.printf( "Variables to compare: %s\n", varsStr );
        
        // This routine must cope with text like this:
        //    --compare-variables "  main.clk1,   main.clk2  ,  , "
        //    --compare-variables \\.main.,.clk1.
        
        if ( !varsStr.isEmpty() && varsStr.charAt( 0 ) == '\\' )
        {
            String rest = varsStr.substring( 1 );
            
            if ( rest.isEmpty() )
            {
                throw new RuntimeException( String.format("The argument \"%s\" to option --%s contains no variable names.",
                                                          varsStr,
                                                          COMPARE_VARIABLES_OPTION ) );
            }
            
            addVarName( rest );
            return;
        }
    
        String[] vars = varsStr.split( "," );
        
        boolean atLeastOne = false;
        
        for ( int i = 0; i < vars.length; ++i )
        {
            String str = vars[i].trim();
            
            if ( ! str.isEmpty() )
            {
                addVarName( str );
                atLeastOne = true;
            }
        }
        
        // If vcdVariablesToCompare remains empty, all variables found will be compared,
        // and that is probably not what the user intended if he tried to specify
        // which variables should be compared.
        if ( !atLeastOne )
        {
            throw new RuntimeException( String.format("The argument \"%s\" to option --%s contains no variable names.",
                                                      varsStr,
                                                      COMPARE_VARIABLES_OPTION ) );
        }
    }
    
    
    private void addVarName ( String varName )
    {
        for ( int i = 0; i < varName.length(); ++i )
        {
            if ( Character.isWhitespace( varName.charAt( i ) ) )
            {
                throw new RuntimeException( String.format("Invalid variable name \"%s\": whitespace characters are not allowed in variable names.",
                                                          varName ) );
            }
        }
        
        vcdVariablesToCompare.add( varName );
    }
  

    private OperationResultEnum compareVcdFiles ()
    {
        // Reuse these objects in the hope of more performance.
        VcdTimePeriod tp1 = new VcdTimePeriod();
        VcdTimePeriod tp2 = new VcdTimePeriod();    
        
        // Read the VCD header until the first timestamp is reached (or until the end of file is reached).
        vcdFile1.parseUntilNextTimestamp( tp1 );
        vcdFile2.parseUntilNextTimestamp( tp2 );
        
        assert tp1.getCurrentTimestamp() == VcdTimePeriod.TIMESTAMP_INITIAL_VALUES : myutils.ErrUtils.assertionFailed(); 
        assert tp2.getCurrentTimestamp() == VcdTimePeriod.TIMESTAMP_INITIAL_VALUES : myutils.ErrUtils.assertionFailed(); 

        assertTimestampInsideTimePeriod( VcdTimePeriod.TIMESTAMP_INITIAL_VALUES, tp1 );            
        assertTimestampInsideTimePeriod( VcdTimePeriod.TIMESTAMP_INITIAL_VALUES, tp2 );            
        
        VcdVariableManager vcdVarManager1 = vcdFile1.getVarManager();
        VcdVariableManager vcdVarManager2 = vcdFile2.getVarManager();

        if ( listVcdVariables )
        {
            vcdVarManager1.printVariableFullRefs( vcdFile1.getFilenameForUserMessages() );
            vcdVarManager2.printVariableFullRefs( vcdFile2.getFilenameForUserMessages() );
        }
        
        
        VcdTimescale ti1 = vcdFile1.getTimescale();
        VcdTimescale ti2 = vcdFile2.getTimescale();

        // The VCD parser will generate an error if the timescale is missing or if it is malformed.
        assert ti1.isValid() && ti2.isValid() : myutils.ErrUtils.assertionFailed();
        
        if ( ! ti1.isEqualTo( ti2 ) )
        {
            System.out.printf( String.format( "The file timescales are different:\n  %s for file 1 and\n  %s for file 2.\n",
                                              ti1.format(),
                                              ti2.format() ) );
            return OperationResultEnum.OPRES_DIFFERENCES_FOUND;
        }
        
        
        OperationResultEnum ecbt = buildVcdVarCompTable();
        
        if ( ecbt != OperationResultEnum.OPRES_CARRY_ON )
            return ecbt;

        
        // Compare the VCD variable values at initialisation. I'm not sure whether this step
        // is always useful, especially if the first timestamp in the file is zero.
        // We could add a command-line option to skip this check.
        
        long comparisonTimestamp = VcdTimePeriod.TIMESTAMP_INITIAL_VALUES;
        
        if ( compareVcdVariables( comparisonTimestamp ) )
        {
            return OperationResultEnum.OPRES_DIFFERENCES_FOUND;
        }
        
        for ( ;; )
        {
            // Step 1) Check if the end of file has been reached on one or both of the files. 
            
            if ( tp1.getNextTimestamp() == VcdTimePeriod.TIMESTAMP_END_OF_FILE )
            {
                if ( tp2.getNextTimestamp() == VcdTimePeriod.TIMESTAMP_END_OF_FILE )
                {
                    // We could warn here if the files have no timestamps
                    // or no variable data at all.
                    return OperationResultEnum.OPRES_NO_DIFFERENCES;
                }
                else
                {
                    System.out.printf( "File 1 ends at timestamp %s, but file 2 carries on further.\n",
                                       VcdTimePeriod.formatTimestamp( tp1.getCurrentTimestamp(),
                                                                      vcdFile1.getTimescale().getUnit() )  );
                    
                    return OperationResultEnum.OPRES_DIFFERENCES_FOUND;
                }
            }
            else if ( tp2.getNextTimestamp() == VcdTimePeriod.TIMESTAMP_END_OF_FILE )
            {
                System.out.printf( "File 2 ends at timestamp %s, but file 1 carries on further.\n",
                                   VcdTimePeriod.formatTimestamp( tp2.getCurrentTimestamp(),
                                                                  vcdFile2.getTimescale().getUnit() ) );
                
                return OperationResultEnum.OPRES_DIFFERENCES_FOUND;
            }
            

            // Step 2) At this point, both files have a future. Move the time
            // forwards on one or on both files.
                    
            if ( tp1.getNextTimestamp() == tp2.getNextTimestamp() )
            {
                moveForwards( vcdFile1, tp1 );
                moveForwards( vcdFile2, tp2 );
                
                assert tp1.getCurrentTimestamp() == tp2.getCurrentTimestamp() : myutils.ErrUtils.assertionFailed();
                comparisonTimestamp = tp1.getCurrentTimestamp();
            }
            else if ( tp1.getNextTimestamp() < tp2.getNextTimestamp() )
            {
                moveForwards( vcdFile1, tp1 );
                comparisonTimestamp = tp1.getCurrentTimestamp();
            }
            else
            {
                moveForwards( vcdFile2, tp2 );
                comparisonTimestamp = tp2.getCurrentTimestamp();
            }
            
            assertTimestampInsideTimePeriod( comparisonTimestamp, tp1 );            
            assertTimestampInsideTimePeriod( comparisonTimestamp, tp2 );
            
            
            // Step 3) Perform the actual comparison.
            
            if ( compareVcdVariables( comparisonTimestamp ) )
            {
                return OperationResultEnum.OPRES_DIFFERENCES_FOUND;
            }
        }
    }

    
    private static void moveForwards ( VcdFileParser vf, VcdTimePeriod tp )
    {
        long next = tp.getNextTimestamp();
        
        vf.parseUntilNextTimestamp( tp );
        
        assert tp.getCurrentTimestamp() == next : myutils.ErrUtils.assertionFailed();
    }


    private static void assertTimestampInsideTimePeriod ( long ts,
                                                          VcdTimePeriod tp )
    {
        // Assert that the comparison time intersects with the current time period
        // of both files.
        
        if ( tp.getCurrentTimestamp() == VcdTimePeriod.TIMESTAMP_INITIAL_VALUES )
        {
            if ( ts == VcdTimePeriod.TIMESTAMP_INITIAL_VALUES )
                return;
            
            assert tp.getNextTimestamp() != VcdTimePeriod.TIMESTAMP_END_OF_FILE : myutils.ErrUtils.assertionFailed();
            
            assert ts < tp.getNextTimestamp() : myutils.ErrUtils.assertionFailed();
            
            return;
        }

        assert ts >= tp.getCurrentTimestamp() : myutils.ErrUtils.assertionFailed();
        
        if ( tp.getNextTimestamp() != VcdTimePeriod.TIMESTAMP_END_OF_FILE )
            assert ts < tp.getNextTimestamp() : myutils.ErrUtils.assertionFailed();
    }

    
    private boolean compareVcdVariables ( long timestamp )
    {
        boolean traceTimestamp = false;
        if ( traceTimestamp )
        {
            System.out.printf( "Comparing variables at timestamp %s.\n",
                               VcdTimePeriod.formatTimestamp( timestamp,
                               vcdFile1.getTimescale().getUnit() ) );
        }
        
        boolean traceVariableValues = false; 
        if ( traceVariableValues )
        {
            vcdFile1.getVarManager().printVariableValues( timestamp, vcdFile1.getTimescale().getUnit(), vcdFile1.getFilenameForUserMessages() );
            vcdFile2.getVarManager().printVariableValues( timestamp, vcdFile2.getTimescale().getUnit(), vcdFile2.getFilenameForUserMessages() );
        }

        
        for ( int i = 0; i < varCompTable.length; ++i)
        {
            VcdVariablePair pair = varCompTable[ i ];
            
            String val1 = pair.v1.getValue(); 
            String val2 = pair.v2.getValue(); 
            
            if ( ! val1.equals( val2 ) )
            {
                System.out.printf( "Variable \"%s\" differs at timestamp %s, values are:\n  %s in file 1 and\n  %s in file 2.\n",
                                   pair.v1.getFullRef(),
                                   VcdTimePeriod.formatTimestamp( timestamp, vcdFile1.getTimescale().getUnit() ),
                                   pair.v1.formatValue(),
                                   pair.v2.formatValue() );

                // At the moment, this tool stops at the first difference.
                return true;
            }
        }
            
        return false;
    }

    
    private OperationResultEnum buildVcdVarCompTable ()
    {
        VcdVariableManager vcdVarManager1 = vcdFile1.getVarManager();
        VcdVariableManager vcdVarManager2 = vcdFile2.getVarManager();
        
        // If we are to compare all variables, check if the variable set is the same
        // in both files.
        
        isComparingAllVariables = vcdVariablesToCompare.size() == 0; 
        
        if ( isComparingAllVariables )
        {
            // If no variables are defined in the VCD headers, do not parse the rest of the file,
            // as it can only be empty of meaningful contents or malformed.
            // I'm not sure whether this should be considered an error. At the moment,
            // the files are just considered to be equal, although one could have
            // more timestamps than the other.

            if ( vcdVarManager1.isEmpty() && vcdVarManager2.isEmpty() )
            {
                System.out.print( "None of the files define any variables that could be compared.\n" );
                return OperationResultEnum.OPRES_NO_DIFFERENCES;
            }

            
            // Check that both files contain the same variables.
            //
            // Note that we check a variable-name level, like a human would do.
            // This means that these example files would be considered different,
            // although they have basically the same data:
            //  - File 1, variables 'a' and 'b' sharing ID '2'.
            //  - File 2, variable 'a' with ID '2', but no variable 'b'.
            
            boolean thereAreVarWithoutCounterparts1 = checkVariableSets( vcdFile1,
                                                                         vcdVarManager1.getAllVarFullRefs(), 
                                                                         vcdVarManager2.getAllVarFullRefs() );
            
            boolean thereAreVarWithoutCounterparts2 = checkVariableSets( vcdFile2, 
                                                                         vcdVarManager2.getAllVarFullRefs(), 
                                                                         vcdVarManager1.getAllVarFullRefs() );
            if ( thereAreVarWithoutCounterparts1 ||
                 thereAreVarWithoutCounterparts2 )
            {
                // There could be users that wish we would compare only those variables that
                // are in both files. But I am not sure that we should default to such behaviour.
                return OperationResultEnum.OPRES_DIFFERENCES_FOUND;
            }
            
            vcdVariablesToCompare.addAll( vcdVarManager1.getAllVarFullRefs() );
        }
        
            
        boolean traceVarNames = false;
        if ( traceVarNames )
        {
            System.out.println("Variables to compare:");
            
            Iterator< String > it = vcdVariablesToCompare.iterator();

            assert it.hasNext() : myutils.ErrUtils.assertionFailed();
                
            do
            {
                String varFullRef = it.next();

                System.out.printf( "  - Variable \"%s\".\n",
                                   varFullRef );
            }
            while ( it.hasNext() );
        }

        
        // Build the variable comparison table.
        //
        // POSSIBLE OPTIMISATION:
        //   Several VcdVariable objects can share a single VcdValue object,
        //   therefore this table could contain VcdValue objects instead of VcdVariable objects.
        
        assert vcdVariablesToCompare.size() > 0 : myutils.ErrUtils.assertionFailed();
        
        varCompTable = new VcdVariablePair[ vcdVariablesToCompare.size() ];
        
        Iterator< String > it = vcdVariablesToCompare.iterator();
        int i;

        for ( i = 0; it.hasNext(); ++i )
        {
            String varFullRef = it.next();
            
            VcdVariableManager.VcdVariable v1 = vcdVarManager1.getVarByFullRef( varFullRef );
            VcdVariableManager.VcdVariable v2 = vcdVarManager2.getVarByFullRef( varFullRef );
            
            if ( v1 == null )
            {
                throw new RuntimeException( String.format( "Variable \"%s\" not found in file \"%s\".",
                                                           varFullRef,
                                                           vcdFile1.getFilenameForUserMessages() ) );
            }
                
            if ( v2 == null )
            {
                throw new RuntimeException( String.format( "Variable \"%s\" not found in file \"%s\".",
                                                           varFullRef,
                                                           vcdFile2.getFilenameForUserMessages() ) );
            }
            
            varCompTable[ i ] = new VcdVariablePair();
            varCompTable[ i ].v1 = v1;
            varCompTable[ i ].v2 = v2;
        }
        
        assert i == vcdVariablesToCompare.size() : myutils.ErrUtils.assertionFailed();
        
        return OperationResultEnum.OPRES_CARRY_ON;        
    }

    
    // Returns true if there are variables in vcdFileA without a counterpart in the other file.
    
    private static boolean checkVariableSets ( VcdFileParser vcdFileA,
                                               Set<String> varSetA,
                                               Set<String> varSetB )
    {
        final String margin = "  ";
        
        TreeSet< String > hs = new TreeSet< String >( varSetA );
        hs.removeAll( varSetB );
        
        Iterator< String > it = hs.iterator();

        if ( !it.hasNext() )
            return false;
            
        System.out.printf( "In file \"%s\":\n", vcdFileA.getFilenameForUserMessages() );
        
        do
        {
            String varFullRef = it.next();

            System.out.printf( "%s- Variable \"%s\" does not have a counterpart in the other file.\n",
                               margin,
                               varFullRef );
        }
        while ( it.hasNext() );

        return true;
    }
    
    
    private void printLicenseText ()
    {
        String licenseText = myutils.Misc.loadClassResourceAsString( this.getClass(), "Gpl3License.txt" );
        System.out.println( licenseText );
    }


    private static void printVersion ()
    {
        System.out.println( appVersion );
    }


    private void printHelp ()
    {
        String helpText = myutils.Misc.loadClassResourceAsString( this.getClass(), "Help.txt" );
        System.out.println( String.format( helpText, appName, appVersion ) );
    }
}
