
package myutils;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.PrintStream;

/**
 * @author R. Diez
 */
public class ErrUtils
{
    /* STRATEGY FOR EXCEPTION HANDLING
     *
     * In my opinion, error handling is hard enough without having to worry
     * about catching different exceptions classes at different locations.
     * I also don't think it's a good idea having to specify what exceptions
     * classes a particular routine can throw. Unfortunately, this is
     * exactly what you must do in Java for all exceptions classes
     * not derived from RuntimeException the routine might throw.
     *
     * In my experience, there's usually nothing to be gained
     * by being able to tell different error classes apart.
     * From what I've seen on the Internet, quite a few people share
     * this point of view.
     * 
     * Therefore all the code I've written converts all exceptions
     * to RuntimeException objects where necessary, in order to minimize
     * the number of try/catch statements overall and to reduce.
     *
     * I find Java's idea of class 'Error' indicating "a serious problem
     * that a reasonable application should not try to catch" rather naive.
     * Java's own Swing GUI classes and Thread classes tend catch
     * those exceptions and ignore them, so the user will never see
     * a hint that something went wrong unless you do catch those
     * "special errors" yourself, at which point the distinction
     * becomes pretty irrelevant. Therefore, the code I've written
     * always catches the base Throwable class when dealing with exceptions.
     */

    public static void simulateError ( )
    {
        throw new RuntimeException( "Simulated error, used only during development for testing and debugging purposes." );
    }

    // Usage example:
    //   assert false : myutils.ErrUtils.assertionFailed();
    public static boolean assertionFailed ()
    {
        // Is there a way to detect whether we are running under a debugger,
        // and, if so, break at this point?
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();

        PrintStream printStream = new PrintStream(outputStream);

    	new Exception("Stack trace").printStackTrace( printStream );

        printStream.flush();

        try
        {
            outputStream.flush();
        }
        catch ( IOException ex )
        {
            // Ignore.
        }

        String msg = "Assertion failed.\n" + outputStream.toString();

        System.err.println( msg );

        System.exit(1);
        return false;
    }


    // This method used to be called rethrowAsRuntimeException(),
    // but there were always silly compilation warnings by the callers,
    // because Java has no "#pragma noreturn" construct.
    // The easiest way to prevent this sort of compilation issues
    // is to make this method return the new exception object,
    // and let the caller throw it.
    public static RuntimeException asRuntimeException ( Throwable ex, String prefixMsg )
    {
        if ( ex instanceof RuntimeException && prefixMsg.isEmpty() )
        {
            return (RuntimeException) ex;
        }

        return new RuntimeException( prefixMsg + getExceptionMessage( ex ) );
    }


    public enum ErrFileOp
    {
        FE_OPENING,
        FE_CREATING,
        FE_READING,
        FE_WRITING,
        FE_FLUSHING,
        FE_CLOSING
    }

    public static RuntimeException asRuntimeExceptionWithFilename ( Throwable ex,
                                                                    File f,
                                                                    ErrFileOp operation )
    {
        // The file name is sometimes present in the error messages, but not always.
    	// If it is already there, do not mention it again in the prefix.

        final String errMsg = getExceptionMessage( ex );

        String errWithFilename = "", errWithoutFilename = "";

        switch ( operation )
        {
            case FE_OPENING:
                errWithoutFilename = "Error opening file: ";
                errWithFilename = "Error opening file \"%1$s\":\n%2$s";
                break;

            case FE_CREATING:
                errWithoutFilename = "Error creating file: ";
                errWithFilename = "Error creating file \"%1$s\":\n%2$s";
                break;

            case FE_READING:
                errWithoutFilename = "Error reading from file: ";
                errWithFilename = "Error reading from file \"%1$s\":\n%2$s";
                break;

            case FE_WRITING:
                errWithoutFilename = "Error writing to file: ";
                errWithFilename = "Error writing to file \"%1$s\":\n%2$s";
                break;

            case FE_FLUSHING:
                errWithoutFilename = "Error flushing file: ";
                errWithFilename = "Error flushing file \"%1$s\":\n%2$s";
                break;

            case FE_CLOSING:
                errWithoutFilename = "Error closing file: ";
                errWithFilename = "Error closing file \"%1$s\":\n%2$s";
                break;

            default:
                assert( false ) : assertionFailed();
        }

        if ( errMsg.contains( f.getAbsolutePath() ) )
        {
            throw asRuntimeException( ex, errWithoutFilename );
        }
        else
        {
            RuntimeException re = new RuntimeException( String.format( errWithFilename,
                                                                       f.getAbsolutePath(),
                                                                       errMsg ) );
            throw re;
        }
    }

    public static String getExceptionMessage ( Throwable ex )
    {
        // The Java libraries throw sometimes exceptions which
        // return an empty error message. This utility function will make sure
        // that you always get some error message back.

        String msg = ex.getMessage();

        // getMessage() can return null. For example, this is the case
        // with exceptions of type NullPointerException.
        if ( msg != null && !msg.isEmpty() )
            return msg;

        // Java adds a silly prefix like "java.lang.RuntimeException: " to the exception message,
        // but not always (!). With this assertion I'm trying to find out exactly where that happens.
        if ( msg != null )
            assert !msg.contains("java.lang") : assertionFailed();

        return "Error of type " +  ex.getClass().getName();
    }

}
