#!/usr/bin/env python3

import unittest      
import subprocess
import time
from translator.intermediateBuilder.intermediateBuilder import IntermediateBuilder
from translator.visitor.DEVSlibrary.devsTranslatorVisitor import DEVSTranslatorVisitor
from translator.visitor.DEVSlibrary.devsConnect import DEVSConnect
from translator.visitor.intermediateVisitor.intermediateNamingVisitor import IntermediateNamingVisitor
from translator.visitor.DEVSlibrary.helper import *

import sys
import os.path
#sys.path.append(os.path.expanduser('~/UA/devs-eindwerk/'))
#import pydevs

class TestDEVSSim(unittest.TestCase):
	def setUp(self):
		# Configuration
		# Needed to copy our own testing framework
		setConfig("testing.ini")
		setConfigLine("experiment", "circuit", "main")
		
	def tearDown(self):
		pass

	def makeOutput(self, filename, bool=True, testing=True):
		try:
			subprocess.check_output(["python3", "translate.py", "--devs", "--config", "testing.ini", filename])
		except subprocess.CalledProcessError as e:
			self.fail()
		self.dirname = filename.split("/")[-1].split(".")[0]

	def compare(self, expectedFile):
		# PyDEVS will save files in devstrace.vcd and devstrace.xml
		# This saves the required files
		"""
		import shutil
		shutil.copy("devstrace.vcd", "testDEVS/vcd/" + expectedFile + ".vcd")
		shutil.copy("devstrace.xml", "testDEVS/xml/" + expectedFile + ".xml")
		"""

		# First compare the VCD files
		# But check whether java is found to run the comparator
		try:
			subprocess.check_output(["which", "java"], stderr=subprocess.STDOUT)
		except subprocess.CalledProcessError as e:
			self.skipTest('java not found using "which"')

		# Use the open-source program comparevcd (http://sourceforge.net/projects/comparevcd/)
		try:
			subprocess.check_output(["java", "-jar", "testDEVS/comparevcd.jar", "devstrace.vcd", "testDEVS/vcd/" + str(expectedFile) + ".vcd"], stderr=subprocess.STDOUT)
		except subprocess.CalledProcessError as e:
			# NOTE: because the comparison fails, doesn't mean that there is a real problem! Because of some factor of randomness it might be possible for an element to trigger before an other element, ignoring the previous output and recalculating this output. This might erase some output
			newString = str(e.output).replace("\\n", "\n").replace("\\t", "\t")
			print("Test failed with code " + str(e.returncode) + " and output: \n" + newString)
			import shutil
			shutil.copy("devstrace.vcd", "debug/" + str(expectedFile) + ".vcd")
			print("For debugging, the file 'devstrace.vcd' has been saved to folder 'debug'")
			self.fail()

		# Seems the VCD file was OK
		# Now check the XML file
		# A pure file comparison is used, since completely parsing the XML files takes extremely long...
		# TODO: the order is random and doesn't seem canonicalized
		from filecmp import cmp
		if not cmp("devstrace.xml", "testDEVS/xml/" + str(expectedFile) + ".xml"):
			print("Comparing XML files failed")
			import shutil
			shutil.copy("devstrace.xml", "debug/" + str(expectedFile) + ".xml")
			print("For debugging, the file 'devstrace.xml' has been saved to folder 'debug'")
			self.fail()

	def framework(self, inputs, expected, step_time, shouldFail = False):
		# Make the tester with the given inputs and expected outputs
		tester = open(getConfig("main", "destination") + self.dirname + "/tester.py", 'w')
		fileStr = "import sys\n"
		fileStr += "import os.path\n"
		fileStr += "sys.path.append(os.path.expanduser('" + getConfig("main", "importPath") + "'))\n"
		fileStr += "import pydevs\n"
		fileStr += "from pydevs.devs_exceptions import *\n"
		fileStr += "from pydevs.infinity import *\n"
		fileStr += "from pydevs.DEVS import *\n\n"

		fileStr += "class State:\n"
		fileStr += "\tdef toXML(self):\n"
		fileStr += "\t\treturn 'Testing'\n"
		fileStr += "\tdef __str__(self):\n"
		fileStr += "\t\treturn 'Tester'\n"

		fileStr += "class testCouple(CoupledDEVS):\n"
		fileStr += "\tdef __init__(self, name):\n"
		fileStr += "\t\tCoupledDEVS.__init__(self, name=name)\n"
		fileStr += "\t\tfrom " + getConfig("experiment", "circuit") + " import " + getConfig("experiment", "circuit") + "\n"
		fileStr += "\t\tself.state = State()\n"
		fileStr += "\t\tself." + getConfig("experiment", "circuit") + " = self.addSubModel(" + getConfig("experiment", "circuit") + "(name='" + getConfig("experiment", "circuit") + "'))\n"
		fileStr += "\t\tself.tester = self.addSubModel(tester(name='tester'))\n"
		for i in range(len(inputs)):
			fileStr += "\t\tself.connectPorts(self.tester.output" + str(i) + ", self." + getConfig("experiment", "circuit") + ".input" + str(i) + ")\n"
		for i in range(len(expected)):
			fileStr += "\t\tself.connectPorts(self." + getConfig("experiment", "circuit") + ".output" + str(i) + ", self.tester.input" + str(i) + ")\n"
		fileStr += "\tdef select(self, immList):\n"
		fileStr += "\t\treturn immList[0]\n"
		fileStr += "\tdef intTransition(self):\n"
		fileStr += "\t\treturn State()\n"
		fileStr += "\tdef extTransition(self):\n"
		fileStr += "\t\treturn State()\n"
		fileStr += "\n"
		fileStr += "class tester(AtomicDEVS):\n"
		fileStr += "\tdef __init__(self, name):\n"
		fileStr += "\t\tAtomicDEVS.__init__(self,name=name)\n"
		fileStr += "\t\tself.state = State()\n"
		fileStr += "\t\tself.correct = True\n"
		for i in range(len(inputs)):
			fileStr += "\t\tself.output" + str(i) + " = self.addOutPort(name='output" + str(i) + "')\n"
		for i in range(len(inputs)):
			fileStr += "\t\tself.toSend" + str(i) + " = " + str(inputs[i]) + "\n"
		for i in range(len(expected)):
			fileStr += "\t\tself.input" + str(i) + " = self.addInPort(name='input" + str(i) + "')\n"
		for i in range(len(expected)):
			fileStr += "\t\tself.expected" + str(i) + " = " + str(expected[i]) + "\n"
		fileStr += "\t\tself.done = False\n"
		fileStr += "\t\tself.sent = -1\n"
		fileStr += "\tdef extTransition(self):\n"
		for i in range(len(expected)):
			fileStr += "\t\tinp = self.peek(self.input" + str(i) + ")\n"
			fileStr += "\t\tif inp != None:\n"
			fileStr += "\t\t\tself.inp" + str(i) + " = inp\n"
		fileStr += "\t\treturn State()\n"
		fileStr += "\tdef intTransition(self):\n"
		for i in range(len(expected)):
			fileStr += "\t\tif (self.sent >= 0) and (self.inp" + str(i) + " != self.expected" + str(i) + "[self.sent]):\n"
			fileStr += "\t\t\tsys.exit('Got ' + str(self.inp" + str(i) + ") + ' on input " + str(i) + ", expected ' + str(self.expected" + str(i) + "[self.sent]))\n"
			fileStr += "\t\telse:\n"
			fileStr += "\t\t\tself.correct = True\n"
		fileStr += "\t\tif self.correct == False:\n"
		fileStr += '\t\t\tsys.exit("Didn\'t receive any output for input, might be a self.poke missing!")\n'
		fileStr += "\t\tself.correct = False\n"

		if len(expected) > 0:
			fileStr += "\t\tif self.sent == len(self.expected0) - 1:\n"
			fileStr += "\t\t\tself.done = True\n"
			fileStr += "\t\t\treturn State()\n"

		fileStr += "\t\tself.sent += 1\n"
		for i in range(len(inputs)):
			fileStr += "\t\tself.poke(self.output" + str(i) + ", self.toSend" + str(i) + "[self.sent])\n"
		fileStr += "\t\treturn State()\n"
		fileStr += "\tdef timeAdvance(self):\n"
		fileStr += "\t\tif self.done:\n"
		fileStr += "\t\t\treturn INFINITY\n"
		fileStr += "\t\telse:\n"
		# The 0.0001 is needed to prevent the tester from going before the element itself
		fileStr += "\t\t\treturn " + str(step_time) + " - (self.elapsed) + 0.0001\n"
		
		tester.write(fileStr)
		tester.close()
		# And run the experiment...
		# NOTE: can't just use import, since the simulator runs in Python2
		# while this translator/tester runs in Python3
		# IMPORTANT: run python with the -B flag to prevent pyc files from being created
		# this would mess up all the files, since they are changed and run at nearly the same
		# time, preventing python from recompiling these files
		try:
			subprocess.check_output(["which", "python2"], stderr=subprocess.STDOUT)
		except subprocess.CalledProcessError as e:
			self.skipTest('python2 not found using "which"')

		# Python2 found, continue
		try:
			subprocess.check_output(["python2", "-B", getConfig("main", "destination") + self.dirname + "/experiment.py"], stderr=subprocess.STDOUT)
			failed = False
		except subprocess.CalledProcessError as e:
			if not shouldFail:
				newString = str(e.output).replace("\\n", "\n").replace("\\t", "\t")
				print("Test failed with code " + str(e.returncode) + " and output: \n" + newString)
			failed = True

		# To prevent multiple exceptions in 1
		if failed and not shouldFail:
			self.fail()
		if not failed and shouldFail:
			self.fail()

	def test_basicNOT(self):
		self.makeOutput("testDEVS/basic_NOT.circ")
		input0 = [	"b01010101", "b1xxE0E10", "b11111111", "b00000000", "bxxxxxxxx", "bEEEEEEEE"]
		expected0 = [	"b10101010", "b0EEE1E01", "b00000000", "b11111111", "bEEEEEEEE", "bEEEEEEEE"]
		self.framework([input0], [expected0], 5)
		self.compare("basic_NOT")
	
	def test_multiGates(self):
		self.makeOutput("testDEVS/multi_Gates.circ")
		input0 = [	"b01010101", "b1xxE0E10", "b11111111", "b00000000", "bxxxxxxxx", "bEEEEEEEE"]
		expected0 = [	"b10101010", "b0EEE1E01", "b00000000", "b11111111", "bEEEEEEEE", "bEEEEEEEE"]
		self.framework([input0], [expected0], 5)
		self.compare("multi_gates")
	
	def test_multiNOT(self):
		self.makeOutput("testDEVS/multi_NOT.circ")
		input0 = [	"b01010101", "b1xxE0E10", "b11111111", "b00000000", "bxxxxxxxx", "bEEEEEEEE"]
		expected0 = [	"b01010101", "b1EEE0E10", "b11111111", "b00000000", "bEEEEEEEE", "bEEEEEEEE"]
		self.framework([input0], [expected0], 5)
		self.compare("multi_not")
	
	def test_basicNOToutChange0(self):
		self.makeOutput("testDEVS/basic_NOToutchange0.circ")
		input0 = [	"b01010101", "b1xxE0E10", "b11111111", "b00000000", "bxxxxxxxx", "bEEEEEEEE"]
		expected0 = [	"b1x1x1x1x", "bxEEE1Ex1", "bxxxxxxxx", "b11111111", "bEEEEEEEE", "bEEEEEEEE"]
		self.framework([input0], [expected0], 5)
		self.compare("basic_notOutChange0")

	def test_basicNOToutChange1(self):
		self.makeOutput("testDEVS/basic_NOToutchange1.circ")
		input0 = [	"b01010101", "b1xxE0E10", "b11111111", "b00000000", "bxxxxxxxx", "bEEEEEEEE"]
		expected0 = [	"bx0x0x0x0", "b0EEExE0x", "b00000000", "bxxxxxxxx", "bEEEEEEEE", "bEEEEEEEE"]
		self.framework([input0], [expected0], 5)
		self.compare("basic_notOutChange1")
	
	def test_basicControlledInverter(self):
		self.makeOutput("testDEVS/basic_ControlledInverter.circ")
		input0 = [	"b01010101", 	"b1xxE0E10", 	"b11111111", 	"b00000000", 	"bxxxxxxxx", 	"bEEEEEEEE", 	"b0101xxE0",	"b01xE10EE", 	"b01xE10EE"]
		control = [	"b1", 		"b1", 		"b1", 		"b1",		"b1",		"b1",		"b0",		"bx",		"bE"]
		expected0 = [	"b10101010", 	"b0EEE1E01", 	"b00000000", 	"b11111111", 	"bEEEEEEEE", 	"bEEEEEEEE",	"bxxxxxxxx",	"bEEEEEEEE",	"bEEEEEEEE"]
		self.framework([control, input0], [expected0], 5)
		self.compare("basic_controlledInverter")
	
	def test_basicControlledBuffer(self):
		self.makeOutput("testDEVS/basic_ControlledBuffer.circ")
		input0 = [	"b01010101", 	"b1xxE0E10", 	"b11111111", 	"b00000000", 	"bxxxxxxxx", 	"bEEEEEEEE",	"b0101xxE0",	"b01xE10EE",	"b01xE10EE"]
		control = [	"b1", 		"b1", 		"b1", 		"b1",		"b1",		"b1",		"b0",		"bx",		"bE"]
		expected0 = [	"b01010101", 	"b1xxE0E10", 	"b11111111", 	"b00000000", 	"bxxxxxxxx", 	"bEEEEEEEE",	"bxxxxxxxx",	"bEEEEEEEE",	"bEEEEEEEE"]
		self.framework([control, input0], [expected0], 5)
		self.compare("basic_controlledBuffer")
	
	def test_basicBuffer(self):
		self.makeOutput("testDEVS/basic_Buffer.circ")
		input0 = [	"b01010101", "b1xxE0E10", "b11111111", "b00000000", "bxxxxxxxx", "bEEEEEEEE"]
		expected0 = [	"b01010101", "b1xxE0E10", "b11111111", "b00000000", "bxxxxxxxx", "bEEEEEEEE"]
		self.framework([input0], [expected0], 5)
		self.compare("basic_buffer")
	
	def test_basicAND(self):
		self.makeOutput("testDEVS/basic_AND.circ")
		input0 = [	"b01010101", "b1ExE0E10", "b11011011", "bx0E01001", "bxxxxxxxx", "bEEEEEEEE"]
		input1 = [	"b01010101", "b1x1E0E11", "b00110111", "b010010x1", "bxxxxxxxx", "bEEEEEEEE"]
		input2 = [	"b01010101", "b1xxE0E11", "b10111100", "b00E01001", "bxxxxxxxx", "bEEEEEEEE"]
		expected0 = [	"b01010101", "b1EEE0E10", "b00010000", "b00001001", "bEEEEEEEE", "bEEEEEEEE"]
		self.framework([input0, input1, input2], [expected0], 5)
		self.compare("basic_and")
	
	def test_basicOR(self):
		self.makeOutput("testDEVS/basic_OR.circ")
		input0 = [	"b01010101", "b1ExE0E10", "b11011011", "bx0E01001", "bxxxxxxxx", "bEEEEEEEE"]
		input1 = [	"b01010101", "b1x1E0E11", "b00110111", "b010010x1", "bxxxxxxxx", "bEEEEEEEE"]
		input2 = [	"b01010101", "b1xxE0E11", "b10111100", "b00E01001", "bxxxxxxxx", "bEEEEEEEE"]
		expected0 = [	"b01010101", "b1E1E0E11", "b11111111", "bE1E010E1", "bEEEEEEEE", "bEEEEEEEE"]
		self.framework([input0, input1, input2], [expected0], 5)
		self.compare("basic_or")
	
	def test_basicNAND(self):
		self.makeOutput("testDEVS/basic_NAND.circ")
		input0 = [	"b01010101", "b1ExE0E10", "b11011011", "bx0E01001", "bxxxxxxxx", "bEEEEEEEE"]
		input1 = [	"b01010101", "b1x1E0E11", "b00110111", "b010010x1", "bxxxxxxxx", "bEEEEEEEE"]
		input2 = [	"b01010101", "b1xxE0E11", "b10111100", "b00E01001", "bxxxxxxxx", "bEEEEEEEE"]
		expected0 = [	"b10101010", "b0EEE1E01", "b11101111", "b11110110", "bEEEEEEEE", "bEEEEEEEE"]
		self.framework([input0, input1, input2], [expected0], 5)
		self.compare("basic_nand")
	
	def test_basicNOR(self):
		self.makeOutput("testDEVS/basic_NOR.circ")
		input0 = [	"b01010101", "b1ExE0E10", "b11011011", "bx0E01001", "bxxxxxxxx", "bEEEEEEEE"]
		input1 = [	"b01010101", "b1x1E0E11", "b00110111", "b010010x1", "bxxxxxxxx", "bEEEEEEEE"]
		input2 = [	"b01010101", "b1xxE0E11", "b10111100", "b00E01001", "bxxxxxxxx", "bEEEEEEEE"]
		expected0 = [	"b10101010", "b0E0E1E00", "b00000000", "bE0E101E0", "bEEEEEEEE", "bEEEEEEEE"]
		self.framework([input0, input1, input2], [expected0], 5)
		self.compare("basic_nor")

	def test_basicXOR(self):
		self.makeOutput("testDEVS/basic_XOR.circ")
		input0 = [	"b01010101", "b1ExE0E10", "b11011011", "bx0E01001", "bxxxxxxxx", "bEEEEEEEE"]
		input1 = [	"b01010101", "b1x1E0E11", "b00110111", "b010010x1", "bxxxxxxxx", "bEEEEEEEE"]
		input2 = [	"b01010101", "b1xxE0E11", "b10111100", "b00E01001", "bxxxxxxxx", "bEEEEEEEE"]
		expected0 = [	"b00000000", "b0EEE0E00", "b01000000", "bE1E000E0", "bEEEEEEEE", "bEEEEEEEE"]
		self.framework([input0, input1, input2], [expected0], 5)
		self.compare("basic_xor")

	def test_failBitsize(self):
		self.makeOutput("testDEVS/basic_AND.circ")
		input0 = [	"b010101"]
		input1 = [	"b010101"]
		expected0 = [	"FAIL"]
		self.framework([input0, input1], [expected0], 5, shouldFail=True)

	def test_failMultiBitsize(self):
		self.makeOutput("testDEVS/basic_AND.circ")
		input0 = [	"b01010101"]
		input1 = [	"b0101"]
		expected0 = [	"FAIL"]
		self.framework([input0, input1], [expected0], 5, shouldFail=True)

	def test_failNoBitsHeader(self):
		self.makeOutput("testDEVS/basic_AND.circ")
		input0 = [	"01010101"]
		input1 = [	"b01010101"]
		expected0 = [	"FAIL"]
		self.framework([input0, input1], [expected0], 5, shouldFail=True)

	def test_failNoBits(self):
		self.makeOutput("testDEVS/basic_AND.circ")
		input0 = [	"bXFAILX"]
		input1 = [	"b010101"]
		expected0 = [	"FAIL"]
		self.framework([input0, input1], [expected0], 5, shouldFail=True)

	def test_failWrongExpect(self):
		self.makeOutput("testDEVS/basic_AND.circ")
		input0 = [	"b010101"]
		input1 = [	"b010101"]
		expected0 = [	"b000000"]
		self.framework([input0, input1], [expected0], 5, shouldFail=True)
	
	def test_basicXNOR(self):
		self.makeOutput("testDEVS/basic_XNOR.circ")
		input0 = [	"b01010101", "b1ExE0E10", "b11011011", "bx0E01001", "bxxxxxxxx", "bEEEEEEEE"]
		input1 = [	"b01010101", "b1x1E0E11", "b00110111", "b010010x1", "bxxxxxxxx", "bEEEEEEEE"]
		input2 = [	"b01010101", "b1xxE0E11", "b10111100", "b00E01001", "bxxxxxxxx", "bEEEEEEEE"]
		expected0 = [	"b11111111", "b1EEE1E11", "b10111111", "bE0E111E1", "bEEEEEEEE", "bEEEEEEEE"]
		self.framework([input0, input1, input2], [expected0], 5)
		self.compare("basic_xnor")
	
	def test_basicEven(self):
		self.makeOutput("testDEVS/basic_Even.circ")
		input0 = [	"b01010101", "b1ExE0E10", "b11011011", "bx0E01001", "bxxxxxxxx", "bEEEEEEEE"]
		input1 = [	"b01010101", "b1x1E0E11", "b00110111", "b010010x1", "bxxxxxxxx", "bEEEEEEEE"]
		input2 = [	"b01010101", "b1xxE0E11", "b10111100", "b00E01001", "bxxxxxxxx", "bEEEEEEEE"]
		expected0 = [	"b10101010", "b0EEE1E01", "b10101111", "bE0E101E0", "bEEEEEEEE", "bEEEEEEEE"]
		self.framework([input0, input1, input2], [expected0], 5)
		self.compare("basic_even")
	
	def test_basicOdd(self):
		self.makeOutput("testDEVS/basic_Odd.circ")
		input0 = [	"b01010101", "b1ExE0E10", "b11011011", "bx0E01001", "bxxxxxxxx", "bEEEEEEEE"]
		input1 = [	"b01010101", "b1x1E0E11", "b00110111", "b010010x1", "bxxxxxxxx", "bEEEEEEEE"]
		input2 = [	"b01010101", "b1xxE0E11", "b10111100", "b00E01001", "bxxxxxxxx", "bEEEEEEEE"]
		expected0 = [	"b01010101", "b1EEE0E10", "b01010000", "bE1E010E1", "bEEEEEEEE", "bEEEEEEEE"]
		self.framework([input0, input1, input2], [expected0], 5)
		self.compare("basic_odd")
	
	def test_helperPull(self):
		self.makeOutput("testDEVS/helper_Pull.circ")
		input0 = [	"bEx01","bEEEE","b0000","bxxxx","b1111"]
		expected0 = [	"bE001","bEEEE","b0000","b0000","b1111"]
		self.framework([input0], [expected0], 5)
		self.compare("basic_pull")

	def test_helperMerger(self):
		self.makeOutput("testDEVS/helper_Merger.circ")
		input0 = [	"bEx01","bxxx1","bEEEx","b1111","bExx0","b0000","bxxxx","b1010","bEEEE","bxExE"]
		input1 = [	"bEx01","bxxx1","b10Ex","b0101","b110x","b10xE","bE010","bxxxx","b101x","bxE10"]
		input2 = [	"b0x10","b1011","bEEEE","b1111","b1111","b000x","bxxxx","bxxx1","b010E","b1EE0"]
		expected0 = [	"bExEE","b1011","bEEEE","bE1E1","bE1EE","bE00E","bE010","b101E","bEEEE","b1EEE"]
		self.framework([input0, input1, input2], [expected0], 5)
		self.compare("basic_merger")

	def test_helperNull(self):
		self.makeOutput("testDEVS/helper_Null.circ")
		input0 = [	"b0",	"b1",	"bx",	"bE"]
		expected0 = [	"b0",	"b1",	"bx",	"bE"]
		self.framework([input0], [expected0], 5)
		self.compare("basic_null")

	def test_basicConstant(self):
		self.makeOutput("testDEVS/basic_Constant.circ")
		expected0 = ["b11100101"]
		self.framework([], [expected0], 5)
		self.compare("basic_constant")

	def test_basicPower(self):
		self.makeOutput("testDEVS/basic_Power.circ")
		expected0 = ["b11111111"]
		self.framework([], [expected0], 5)
		self.compare("basic_power")

	def test_basicGround(self):
		self.makeOutput("testDEVS/basic_Ground.circ")
		expected0 = ["b00000000"]
		self.framework([], [expected0], 5)
		self.compare("basic_ground")

	def test_basicMultiplex(self):
		self.makeOutput("testDEVS/basic_Multiplexer.circ")
		input0 = [	"b0000",	"b0001",	"b1111",	"bxE01",	"b0101",	"bxxxx",	"bEEEE",	"b0000",	"bxxxx"]
		input1 = [	"b1111",	"b1111",	"b0xxE",	"bEEEE",	"b11xE",	"bxxxE",	"bEEEx",	"b0000",	"bxxxx"]
		enable = [	"b0",		"b1",		"b1",		"bx",		"bx",		"bE",		"b1",		"b1",		"bx"]
		select = [	"b0",		"b0",		"b1",		"b1",		"bx",		"bE",		"b0",		"b0",		"bx"]
		expected0 = [	"bxxxx",	"b0001",	"b0xxE",	"bEEEE",	"bxxxx",	"bEEEE",	"bEEEE",	"b0000",	"bxxxx"]
		self.framework([input0, enable, input1, select], [expected0], 5)
		self.compare("basic_multiplex")

	def test_basic0Extender(self):
		self.makeOutput("testDEVS/basic_0Extender.circ")
		input0 = [	"b0000",	"b1111",	"bxxx1",	"bE01x"]
		expected0 = [	"b00000000",	"b00001111",	"b0000xxx1",	"b0000E01x"]
		self.framework([input0], [expected0], 5)
		self.compare("basic_0extender")

	def test_basicSignExtender(self):
		self.makeOutput("testDEVS/basic_SignExtender.circ")
		input0 = [	"b0000",	"b1111",	"bxxx1",	"bE01x"]
		expected0 = [	"b00000000",	"b11111111",	"bxxxxxxx1",	"bEEEEE01x"]
		self.framework([input0], [expected0], 5)
		self.compare("basic_signextender")

	def test_basicInputExtender(self):
		self.makeOutput("testDEVS/basic_InputExtender.circ")
		input1 = [	"b0000",	"b1111",	"bxxx1",	"bE01x"]
		input0 = [	"b1",		"b0",		"bE",		"bx"]
		expected0 = [	"b11110000",	"b00001111",	"bEEEExxx1",	"bxxxxE01x"]
		self.framework([input0, input1], [expected0], 5)
		self.compare("basic_inputextender")

	def test_basicShortExtender(self):
		self.makeOutput("testDEVS/basic_ShortExtender.circ")
		input0 = [	"b0000",	"b1111",	"bxxx1",	"bE01x"]
		expected0 = [	"b0",		"b1",		"b1",		"bx"]
		self.framework([input0], [expected0], 5)
		self.compare("basic_shortextender")

	def test_basicClock(self):
		self.makeOutput("testDEVS/basic_Clock.circ")
		expected0 = [	"b0",	"b1",	"b0",	"b1",	"b0",	"b1",	"b0",	"b1"]
		self.framework([], [expected0], 10)
		self.compare("basic_clock")

	def test_basicPTransistor(self):
		self.makeOutput("testDEVS/basic_PTransistor.circ")
		input0 = [	"bxE01",	"bxE01",	"bxE01",	"bxE01"]
		input1 = [	"b0",		"b1",		"bx",		"bE"]
		expected0 = [	"bxE01",	"bxxxx",	"bxEEE",	"bxEEE"]
		self.framework([input0, input1], [expected0], 5)
		self.compare("basic_ptransistor")

	def test_basicNTransistor(self):
		self.makeOutput("testDEVS/basic_NTransistor.circ")
		input0 = [	"bxE01",	"bxE01",	"bxE01",	"bxE01"]
		input1 = [	"b0",		"b1",		"bx",		"bE"]
		expected0 = [	"bxxxx",	"bxE01",	"bxEEE",	"bxEEE"]
		self.framework([input0, input1], [expected0], 5)
		self.compare("basic_ntransistor")

	def test_basicTransmission(self):
		self.makeOutput("testDEVS/basic_Transmission.circ")
		input0 = ["bxE01","bxE01","bxE01","bxE01","bxE01","bxE01","bxE01","bxE01","bxE01","bxE01","bxE01","bxE01", "bxE01","bxE01","bxE01","bxE01","bxE01","bxE01","bxE01","bxE01"]
		input1 = ["b0",	"b0", "b0", "b0", "b1", "b1", "b1", "b1", "bE", "bE", "bE", "bE", "bx", "bx", "bx", "bx"]
		input2 = ["b0",	"b1", "bE", "bx", "b0",	"b1", "bE", "bx", "b0",	"b1", "bE", "bx", "b0",	"b1", "bE", "bx"]
		expected0 = [	"bxEEE", "bxE01", "bxEEE", "bxEEE", \
				"bxxxx", "bxEEE", "bxEEE", "bxEEE", \
				"bxEEE", "bxEEE", "bxEEE", "bxEEE", \
				"bxEEE", "bxEEE", "bxEEE", "bxEEE"]
		self.framework([input0, input1, input2], [expected0], 3)
		self.compare("basic_transmission")

	def test_basicDecoder(self):
		self.makeOutput("testDEVS/basic_Decoder.circ")
		select = [	"b0", "b0", "b0", "b0", "b1", "b1", "b1", "b1", "bE", "bE", "bE", "bE", "bx", "bx", "bx", "bx"]
		enable = [	"b0", "b1", "bE", "bx", "b0", "b1", "bE", "bx", "b0", "b1", "bE", "bx", "b0", "b1", "bE", "bx"]
		expected0 = [	"bx", "b1", "bE", "b1", "bx", "b0", "bE", "b0", "bx", "bE", "bE", "bE", "bx", "bx", "bE", "bx"]
		expected1 = [	"bx", "b0", "bE", "b0", "bx", "b1", "bE", "b1", "bx", "bE", "bE", "bE", "bx", "bx", "bE", "bx"]
		self.framework([select, enable], [expected1, expected0], 3)
		self.compare("basic_decoder")

	def test_basicBigDecoder(self):
		self.makeOutput("testDEVS/basic_BigDecoder.circ")
		select = [	"b00","b01","b10","b11","bx1","bEx","b0E"]
		expected0 = [	"b1", "b0", "b0", "b0", "bx", "bE", "bE"]
		expected1 = [	"b0", "b1", "b0", "b0", "bx", "bE", "bE"]
		expected2 = [	"b0", "b0", "b1", "b0", "bx", "bE", "bE"]
		expected3 = [	"b0", "b0", "b0", "b1", "bx", "bE", "bE"]
		self.framework([select], [expected0, expected1, expected2, expected3], 3)
		self.compare("basic_bigDecoder")

	def test_basicDemultiplexer(self):
		self.makeOutput("testDEVS/basic_Demultiplexer.circ")
		input0 = ["bxE01","bxE01","bxE01","bxE01","bxE01","bxE01","bxE01","bxE01","bxE01","bxE01","bxE01","bxE01", "bxE01","bxE01","bxE01","bxE01","bxE01","bxE01","bxE01","bxE01"]
		enable = ["b0",	"b0", "b0", "b0", "b1", "b1", "b1", "b1", "bE", "bE", "bE", "bE", "bx", "bx", "bx", "bx"]
		select = ["b0",	"b1", "bE", "bx", "b0",	"b1", "bE", "bx", "b0",	"b1", "bE", "bx", "b0",	"b1", "bE", "bx"]
		expected0 = [	"bxxxx", "bxxxx", "bxxxx", "bxxxx", \
				"bxE01", "b0000", "bEEEE", "bxxxx", \
				"bEEEE", "bEEEE", "bEEEE", "bEEEE", \
				"bxE01", "b0000", "bEEEE", "bxxxx"]
		expected1 = [	"bxxxx", "bxxxx", "bxxxx", "bxxxx", \
				"b0000", "bxE01", "bEEEE", "bxxxx", \
				"bEEEE", "bEEEE", "bEEEE", "bEEEE", \
				"b0000", "bxE01", "bEEEE", "bxxxx"]
		self.framework([enable, select, input0], [expected0, expected1], 3)
		self.compare("basic_demultiplexer")

	def test_basicBitSelector(self):
		self.makeOutput("testDEVS/basic_BitSelector.circ")
		input0 = [	"bxE01",	"bxE01",	"bxE01",	"bxE01",	"b0000",	"b0000"]
		select = [	"b00",		"b01",		"b10",		"b11",		"b0x",		"bEx"]
		expected0 = [	"b1",		"b0",		"bE",		"bx",		"bx",		"bx"]
		self.framework([input0, select], [expected0], 3)
		self.compare("basic_bitSelector")

	def test_basicBitSelectorPad(self):
		self.makeOutput("testDEVS/basic_BitSelectorPad.circ")
		input0 = [	"bxE01",	"bxE01", "b0000",	"b0000"]
		select = [	"b0",		"b1",	 "bx",		"bE"]
		expected0 = [	"bE01",		"b00x",	 "bxxx",	"bxxx"]
		self.framework([input0, select], [expected0], 3)
		self.compare("basic_bitSelectorPad")

	def test_basicBitSelectorPadExtend(self):
		self.makeOutput("testDEVS/basic_BitSelectorPadExtend.circ")
		input0 = [	"bxE01",	"bxE01",	"b0000",	"b0000"]
		select = [	"b0",		"b1",		"bx",		"bE"]
		expected0 = [	"b0000xE01",	"b00000000",	"bxxxxxxxx",	"bxxxxxxxx"]
		self.framework([input0, select], [expected0], 3)
		self.compare("basic_bitSelectorPadExtend")

	def test_basicPriorityEncoder(self):
		self.makeOutput("testDEVS/basic_PriorityEncoder.circ")
		input0 = [	"b0",	"b0",	"b1",	"b1",	"b0", 	"b1",	"bE"]
		input1 = [	"b0",	"b1",	"b0",	"b1",	"b0", 	"b0",	"bx"]
		enable = [	"b1",	"b1", 	"b1",	"b1",	"b0", 	"bE",	"bx"]
		expected0 = [	"bx",	"b1",	"b0",	"b1",	"bx", 	"b0",	"bx"]
		isEnable = [	"b1",	"b0",	"b0",	"b0",	"b0", 	"b0",	"b1"]
		groupSelect = [	"b0", 	"b1",	"b1",	"b1",	"b0", 	"b1",	"b0"]
		self.framework([input1, enable, input0], [isEnable, expected0, groupSelect], 3)
		self.compare("basic_priorityEncoder")

	def test_basicPriorityEncoderLarge(self):
		self.makeOutput("testDEVS/basic_PriorityEncoderLarge.circ")
		input0 = [	"b0",	"b0",	"b1",	"b1"]
		input1 = [	"b0",	"b1",	"b0",	"b0"]
		input2 = [	"b0",	"b0",	"b1",	"b0"]
		input3 = [	"b0",	"b1",	"b0",	"b0"]
		input4 = [	"b0",	"b0",	"b1",	"b1"]
		input5 = [	"b0",	"b1",	"b0",	"b0"]
		input6 = [	"b0",	"b0",	"b1",	"b0"]
		input7 = [	"b0",	"b1",	"b0",	"b0"]
		input8 = [	"b0",	"b0",	"b1",	"b1"]
		input9 = [	"b0",	"b1",	"b0",	"b0"]
		input10 = [	"b0",	"b0",	"b1",	"b0"]
		input11 = [	"b0",	"b1",	"b0",	"b0"]
		input12 = [	"b0",	"b0",	"b1",	"b0"]
		input13 = [	"b0",	"b1",	"b0",	"b0"]
		input14 = [	"b0",	"b0",	"b1",	"b0"]
		input15 = [	"b0",	"b1",	"b0",	"b0"]
		enable = [	"b1",	"b1", 	"b1",	"b1"]
		isEnable = [	"b1",	"b0",	"b0",	"b0"]
		expected0 = [	"bxxxx","b1111","b1110","b1000"]
		groupSelect = [	"b0", 	"b1",	"b1",	"b1"]
		self.framework([input15, input9, input3, input4, input0, enable, input12, input6, input1, input2, input5, input13, input8, input10, input11, input7, input14], [groupSelect, isEnable, expected0], 3)
		self.compare("basic_priorityEncoderLarge")

	def test_multiSubcircuit(self):
		self.makeOutput("testDEVS/multi_Subcircuit.circ")
		input0 = [	"b01010101", "b1xxE0E10", "b11111111", "b00000000", "bxxxxxxxx", "bEEEEEEEE"]
		expected0 = [	"b10101010", "b0EEE1E01", "b00000000", "b11111111", "bEEEEEEEE", "bEEEEEEEE"]
		self.framework([input0], [expected0], 5)
		self.compare("multi_subCircuit")
	
	def test_basicANDNegate(self):
		self.makeOutput("testDEVS/basic_ANDNegate.circ")
		negated = [	"b10000010", "b0ExE1E01", "b00100100", "bx1E10110", "bxxxxxxxx", "bEEEEEEEE"]
		input1 = [	"b01010101", "b1x1E0E11", "b00110111", "b010010x1", "bxxxxxxxx", "bEEEEEEEE"]
		input0 = [	"b01010101", "b1xxE0E11", "b10111100", "b00E01001", "bxxxxxxxx", "bEEEEEEEE"]
		expected0 = [	"b01010101", "b1EEE0E10", "b00010000", "b00001001", "bEEEEEEEE", "bEEEEEEEE"]
		self.framework([negated, input0, input1], [expected0], 5)
		self.compare("basic_andNegate")

	def test_basicMultiplexFloating0(self):
		self.makeOutput("testDEVS/basic_Multiplexer0.circ")
		input0 = [	"b0000",	"b0001",	"b1111",	"bxE01",	"b0101",	"bxxxx",	"bEEEE",	"b0000",	"bxxxx"]
		input1 = [	"b1111",	"b1111",	"b0xxE",	"bEEEE",	"b11xE",	"bxxxE",	"bEEEx",	"b0000",	"bxxxx"]
		enable = [	"b0",		"b1",		"b1",		"bx",		"bx",		"bE",		"b1",		"b1",		"bx"]
		select = [	"b0",		"b0",		"b1",		"b1",		"bx",		"bE",		"b0",		"b0",		"bx"]
		expected0 = [	"b0000",	"b0001",	"b0xxE",	"bEEEE",	"bxxxx",	"bEEEE",	"bEEEE",	"b0000",	"bxxxx"]
		self.framework([select, enable, input0, input1], [expected0], 5)
		self.compare("basic_multiplexFloating0")

	def test_basicAdder(self):
		self.makeOutput("testDEVS/basic_Adder.circ")
		input0 = [	"b10101010", 	"b11111111", 	"b10101010", 	"bxx101010", 	"b00000000", 	"b10101010",	"b00000000",	"b00000000",	"b0000000x",	"b00000000",	"b00000001", 	"b00000111",	"b0x000011"]
		input1 = [	"b00000000", 	"b11111111", 	"b10101010", 	"b0x011001", 	"b01010x01", 	"b0101E100",	"b0101010E",	"b0101010x",	"b0101010E",	"b00000000",	"b00000001", 	"b00000x11",	"b000x0011"]
		carryIn = [	"b0", 		"b1", 		"b1", 		"b0", 		"b1", 		"b0",		"b0",		"b0",		"b0",		"bE",		"bx",		"b0",		"b0"]
		expected0 = [	"b10101010", 	"b11111111", 	"b01010101", 	"bxx000011", 	"bxxxxxx10", 	"bEEEEE110",	"bEEEEEEEE",	"bxxxxxxxx",	"bEEEEEEEE",	"bEEEEEEEE",	"b00000010",	"bxxxxxx10",	"bxxxx0110"]
		carryOut = [	"b0", 		"b1", 		"b1", 		"bx", 		"bx",		"bE",		"bE",		"bx",		"bE",		"bE",		"b0",		"bx",		"bx"]
		self.framework([input0, input1, carryIn], [carryOut, expected0], 5)
		self.compare("basic_adder")
	
	def test_basicSubtractor(self):
		self.makeOutput("testDEVS/basic_Subtractor.circ")
		input0 = [	"b00000000", 	"b11111111", 	"b10101010", 	"b0x011001", 	"b01010x01", 	"b0101E100",	"b0101010E",	"b0101010x",	"b0101010E",	"b00000000",	"b00000001",	"b00x01x11"]
		input1 = [	"b10101010", 	"b11111111", 	"b10101010", 	"bxx101010", 	"b00000000", 	"b10101010",	"b00000000",	"b00000000",	"b0000000x",	"b00000000",	"b00000001",	"b1x0010x0"]
		carryIn = [	"b0", 		"b1", 		"b1", 		"b0", 		"b1", 		"b0",		"b0",		"b0",		"b0",		"bE",		"bx",		"bx"]
		expected0 = [	"b01010110", 	"b11111111", 	"b11111111", 	"bEE101111", 	"bxxxxxx00", 	"bEEEEE010",	"bEEEEEEEE",	"bxxxxxxxx",	"bEEEEEEEE",	"bEEEEEEEE",	"b00000000",	"bEEEEEEE1"]
		carryOut = [	"b1", 		"b1", 		"b1", 		"bE", 		"bE",		"bE",		"bE",		"bE",		"bE",		"bE",		"b0",		"bE"]
		self.framework([input0, input1, carryIn], [carryOut, expected0], 5)
		self.compare("basic_subtractor")
	
	def test_basicMultiplier(self):
		self.makeOutput("testDEVS/basic_Multiplier.circ")
		input0 = [	"b10101010", 	"b11111111", 	"b10101010", 	"bxx101010", 	"b00000000", 	"b10101010",	"b00000000",	"b00000000",	"b0000000x",	"b00000000",	"b00000001"]
		input1 = [	"b00000000", 	"b11111111", 	"b10101010", 	"b0x011001", 	"b01010x01", 	"b0101E100",	"b0101010E",	"b0101010x",	"b0101010E",	"b00000000",	"b00000001"]
		carryIn = [	"b00101000", 	"b11111111", 	"b01010101", 	"b00000000", 	"bxxxxxxxx", 	"b01x01011",	"bEEEEEEEx",	"b1111E000",	"b1110xxxx",	"bEE000100",	"b000x0010"]
		expected0 = [	"b00101000", 	"b00000000", 	"b00111001", 	"bxx011010", 	"bxxxxxx00", 	"bEEEEE011",	"bEEEEEEEE",	"bEEEEExxx",	"bEEEEEEEE",	"bEE000100",	"bxxxx0011"]
		carryOut = [	"b00000000", 	"b11111111", 	"b01110001", 	"bxxxxxxxx", 	"bxxxxxxxx",	"bEEEEEEEE",	"bEEEEEEEE",	"bEEEEEEEE",	"bEEEEEEEE",	"bEEEEEEEE",	"bxxxxxxxx"]
		self.framework([input0, input1, carryIn], [expected0, carryOut], 5)
		self.compare("basic_multiplier")

	def test_basicDivider(self):
		self.makeOutput("testDEVS/basic_Divider.circ")
		input0 = [	"b10101010", 	"b11111111", 	"b10101010", 	"b10101010",	"bxx101010", 	"b00000000", 	"b10101010",	"b00000000",	"b00000000",	"b0000000x",	"b00000000",	"b00000001"]
		input1 = [	"b00000000", 	"b11111111", 	"b10101010", 	"b10101110",	"b0x011001", 	"b01010x01", 	"b0101E100",	"b0101010E",	"b0101010x",	"b0101010E",	"b00000000",	"b00000001"]
		carryIn = [	"b00101000", 	"b11111111", 	"b01010101", 	"b01010111",	"b00000000", 	"bxxxxxxxx", 	"b01x01011",	"bEEEEEEEx",	"b11111000",	"b1110xxxx",	"bEE000100",	"b000x0010"]
		expected0 = [	"b10101010", 	"b00000001", 	"b10000001", 	"b10000000",	"bxxxxxxxx", 	"bxxxxxxxx", 	"bEEEEEEEE",	"bEEEEEEEE",	"bxxxxxxxx",	"bEEEEEEEE",	"bEEEEEEEE",	"bxxxxxxxx"]
		carryOut = [	"b00000000", 	"b00000000", 	"b00000000", 	"b10101010",	"bxxxxxxxx", 	"bxxxxxxxx",	"bEEEEEEEE",	"bEEEEEEEE",	"bxxxxxxxx",	"bEEEEEEEE",	"bEEEEEEEE",	"bxxxxxxxx"]
		self.framework([input0, carryIn, input1], [expected0, carryOut], 5)
		self.compare("basic_divider")
	
	def test_basicNegator(self):
		self.makeOutput("testDEVS/basic_Negator.circ")
		input0 = [	"b00000000",	"b10101010",	"b11111111",	"b0101x110",	"b00x01E11",	"bxxxxxxxx",	"bEEEEEEEE", 	"b001010x0"]
		expected0 = [	"b00000000",	"b01010110",	"b00000001",	"b1010x010",	"b11x10E01",	"bxxxxxxxx",	"bEEEEEEEE",	"b1101xxx0"]
		self.framework([input0], [expected0], 5)
		self.compare("basic_negator")

	def test_multiCalculator(self):
		self.makeOutput("testDEVS/multi_calculator.circ")
		input0 = [	"b00000000",	"b10101010",	"b11111111",	"b01011110",	"b1x0010x0",	"b00x01x11",	"bx0001000"]
		input1 = [	"b00000000",	"b01010101",	"b01010101",	"b11001101",	"b00x01x11",	"b1x0010x0",	"bxx110011"]
		select = [	"b0",		"b0",		"b0",		"b1",		"b1",		"b0",		"b1"]
		expected0 = [	"b00000000",	"b11111111",	"b01010100",	"b01101111",	"bEEEEEEE1",	"bxxxxxxx1",	"bxx101011"]
		carryMerged = [	"b0",		"bE",		"b1",		"bE",		"bE",		"bE",		"bE"]
		self.framework([select, input0, input1], [carryMerged, expected0], 10)
		self.compare("multi_calculator")

	def test_multiBigCalculator(self):
		self.makeOutput("testDEVS/multi_bigCalculator.circ")
		input0 = [	"b00000000",	"b10101010",	"b11111111",	"b01011110",	"b1x0010x0",	"b00x01x11",	"bx0001000"]
		input1 = [	"b00000000",	"b01010101",	"b01010101",	"b11001101",	"b00x01x11",	"b1x0010x0",	"bxx110011"]
		select = [	"b00",		"b00",		"b00",		"b01",		"b01",		"b00",		"b01"]
		expected0 = [	"b00000000",	"b11111111",	"b01010100",	"b01101111",	"bEEEEEEE1",	"bxxxxxxx1",	"bxx101011"]
		self.framework([select, input1, input0], [expected0], 10)
		self.compare("multi_bigCalculator")
		
	def test_basicMultiplexInverse(self):
		self.makeOutput("testDEVS/basic_MultiplexerInverse.circ")
		input0 = [	"b0000","b0001","b1111","bxE01","b0101","bxxxx","bEEEE","b0000","bxxxx","b0000","b0000"]
		input1 = [	"b1111","b1111","b0xxE","bEEEE","b11xE","bxxxE","bEEEx","b0000","bxxxx","b0001","b0001"]
		input2 = [	"b1111","b1111","b0xxE","bEEEE","b11xE","bxxxE","bEEEx","b0000","bxxxx","b0010","b0010"]
		input3 = [	"b1111","b1111","b0xxE","bEEEE","b11xE","bxxxE","bEEEx","b0000","bxxxx","b0011","b0011"]
		enable = [	"b0",	"b1",	"b1",	"bx",	"bx",	"bE",	"b1",	"b1",	"bx",	"b1",	"b1"]
		select = [	"b00",	"b00",	"b01",	"b01",	"bxx",	"bEE",	"b00",	"b00",	"bxx",	"b0x",	"bEx"]
		expected0 = [	"bxxxx","b0001","b0xxE","bEEEE","bxxxx","bEEEE","bEEEE","b0000","bxxxx","bxxxx","bEEEE"]
		self.framework([input2, input0, input3, input1, enable, select], [expected0], 5)
		self.compare("basic_multiplexInverse")

	def test_basicMultiplexInverseTurn(self):
		self.makeOutput("testDEVS/basic_MultiplexerInverseTurn.circ")
		input0 = [	"b0000",	"b0001",	"b1111",	"bxE01",	"b0101",	"bxxxx",	"bEEEE",	"b0000",	"bxxxx"]
		input1 = [	"b1111",	"b1111",	"b0xxE",	"bEEEE",	"b11xE",	"bxxxE",	"bEEEx",	"b0000",	"bxxxx"]
		input2 = [	"b1111",	"b1111",	"b0xxE",	"bEEEE",	"b11xE",	"bxxxE",	"bEEEx",	"b0000",	"bxxxx"]
		input3 = [	"b1111",	"b1111",	"b0xxE",	"bEEEE",	"b11xE",	"bxxxE",	"bEEEx",	"b0000",	"bxxxx"]
		enable = [	"b0",		"b1",		"b1",		"bx",		"bx",		"bE",		"b1",		"b1",		"bx"]
		select = [	"b00",		"b00",		"b01",		"b01",		"bxx",		"bEE",		"b00",		"b00",		"bxx"]
		expected0 = [	"bxxxx",	"b0001",	"b0xxE",	"bEEEE",	"bxxxx",	"bEEEE",	"bEEEE",	"b0000",	"bxxxx"]
		self.framework([input2, input0, input3, input1, enable, select], [expected0], 5)
		self.compare("basic_multiplexInverseTurn")

	def test_basicMultiplexNumbers(self):
		self.makeOutput("testDEVS/basic_MultiplexerNumbers.circ")
		input1 = [	"b1111",	"b1111",	"b0101",	"b0101"]
		input3 = [	"b1111",	"b1111",	"b1010",	"b1010"]
		select = [	"b00",		"b10",		"b01",		"b11"]
		expected0 = [	"bxxxx",	"bxxxx",	"b0101",	"b1010"]
		self.framework([select, input1, input3], [expected0], 5)
		self.compare("basic_multiplexNumbers")

	def test_multiExotic(self):
		self.makeOutput("testDEVS/multi_exotic.circ")
		input0 = [	"b0","b0","b0","b0","b0","b0","b0","b0","b1","b1","b1","b1","b1","b1","b1","b1"]
		input1 = [	"b0","b0","b0","b0","b1","b1","b1","b1","b0","b0","b0","b0","b1","b1","b1","b1"]
		input2 = [	"b0","b0","b1","b1","b0","b0","b1","b1","b0","b0","b1","b1","b0","b0","b1","b1"]
		input3 = [	"b0","b1","b0","b1","b0","b1","b0","b1","b0","b1","b0","b1","b0","b1","b0","b1"]
		expected = [	"bE","bE","bE","bE","b1","bE","b1","b0","bE","bE","bE","bE","bE","bE","bE","bE"]
		self.framework([input0, input1, input2, input3], [expected], 7)
		self.compare("multi_exotic")

	def test_advTransistor(self):
		self.makeOutput("testDEVS/adv_Transistor.circ")
		input0 = [	"b01xE", "b01xE", "b01xE", "b01xE", "b01xE", "b01xE", "b01xE", "b01xE", "b01xE", "b01xE", "b01xE", "b01xE", "b01xE", "b01xE", "b01xE", "b01xE"]
		input1 = [	"b0","b0","b0","b0","b1","b1","b1","b1","bx","bx","bx","bx","bE","bE","bE","bE"]
		input2 = [	"b0","b1","bx","bE","b0","b1","bx","bE","b0","b1","bx","bE","b0","b1","bx","bE"]
		ptype = [	"b01xE", "b01xE", "b01xE", "b01xE", "bxxxx", "bxxxx", "bxxxx", "bxxxx", "bEExE", "bEExE", "bEExE", "bEExE", "bEExE", "bEExE", "bEExE", "bEExE"]
		ntype = [	"bxxxx", "bxxxx", "bxxxx", "bxxxx", "b01xE", "b01xE", "b01xE", "b01xE", "bEExE", "bEExE", "bEExE", "bEExE", "bEExE", "bEExE", "bEExE", "bEExE"]
		ctype = [	"bEExE", "b01xE", "bEExE", "bEExE", "bxxxx", "bEExE", "bEExE", "bEExE", "bEExE", "bEExE", "bEExE", "bEExE", "bEExE", "bEExE", "bEExE", "bEExE"]
		self.framework([input0, input1, input2], [ptype, ptype, ptype, ptype, ptype, ptype, ptype, ptype, ntype, ntype, ntype, ntype, ntype, ntype, ntype, ntype, ctype, ctype, ctype, ctype, ctype, ctype, ctype, ctype], 7)
		self.compare("adv_transistor")

	def test_basicComparatorUnsigned(self):
		self.makeOutput("testDEVS/basic_ComparatorUnsigned.circ")
		input0 = [	"b0000","b0001","b0000","b1111","bE000","b00x1","b11x0"]
		input1 = [	"b0000","b0001","b0001","b0001","b1111","b0000","b10EE"]
		smaller = [	"b0",	"b0",	"b1",	"b0",	"bE",	"bx",	"b0"]
		equal = [	"b1",	"b1",	"b0",	"b0",	"bE",	"bx",	"b0"]
		greater = [	"b0",	"b0",	"b0",	"b1",	"bE",	"bx",	"b1"]
		self.framework([input0, input1], [greater, equal, smaller], 5)
		self.compare("basic_comparatorUnsigned")

	def test_basicComparatorComplement(self):
		self.makeOutput("testDEVS/basic_ComparatorComplement.circ")
		input0 = [	"b0000","b0001","b0000","b1111","bE000","b00x1","b11x0"]
		input1 = [	"b0000","b0001","b0001","b0001","b1111","b0000","b10EE"]
		smaller = [	"b0",	"b0",	"b1",	"b1",	"bE",	"bx",	"b0"]
		equal = [	"b1",	"b1",	"b0",	"b0",	"bE",	"bx",	"b0"]
		greater = [	"b0",	"b0",	"b0",	"b0",	"bE",	"bx",	"b1"]
		self.framework([input0, input1], [greater, equal, smaller], 5)
		self.compare("basic_comparatorComplement")

	def test_basicShifter(self):
		self.makeOutput("testDEVS/basic_shifter.circ")
		input0 = [	"b0000","b0001","b0100","b1111","bE000","b00x1","b11x0","b010E","b0E01","b0xE1","b0xE1","b0xE1"]
		input1 = [	"b01",  "b01",  "b01",  "b01",  "b01",  "b01",  "b01",  "b10"  ,"b00",  "bx0",  "bxE",  "b0E"]
		ll = 	 [	"b0000","b0010","b1000","b1110","b0000","b0x10","b1x00","b0E00","b0E01","bEEEE","bEEEE","bEEEE"]
		lr = 	 [	"b0000","b0000","b0010","b0111","b0E00","b000x","b011x","b0001","b0E01","bEEEE","bEEEE","bEEEE"]
		ar = 	 [	"b0000","b0000","b0010","b1111","bEE00","b000x","b111x","b0001","b0E01","bEEEE","bEEEE","bEEEE"]
		rl = 	 [	"b0000","b0010","b1000","b1111","b000E","b0x10","b1x01","b0E01","b0E01","bEEEE","bEEEE","bEEEE"]
		rr = 	 [	"b0000","b1000","b0010","b1111","b0E00","b100x","b011x","b0E01","b0E01","bEEEE","bEEEE","bEEEE"]
		self.framework([input0, input1], [ll, lr, ar, rl, rr], 5)
		self.compare("basic_shifter")

	def test_basicBitAdder(self):
		self.makeOutput("testDEVS/basic_BitAdder.circ")
		input0 = [	"b0000","b0001","b0100","b1111","bE000","bx0x0","b1x01"]
		output = [	"b000",	"b001",	"b001",	"b100",	"b00E",	"b0EE",	"b01E"]
		self.framework([input0], [output], 5)
		self.compare("basic_bitAdder")

	def test_basicBitAdderBig(self):
		self.makeOutput("testDEVS/basic_BitAdderBig.circ")
		input0 = [	"b0000", "bE001", "bx011", "b1111"]
		input1 = [	"b1010", "b01E1", "b0101", "b1111"]
		input2 = [	"b0101", "b0x01", "b0001", "b1111"]
		input3 = [	"b0011", "b100x", "b1000", "b1111"]
		output = [	"b00110","b0EEEE","b0011E","b10000"]
		self.framework([input0, input1, input2, input3], [output], 5)
		self.compare("basic_bitAdderBig")

	def test_basicBitFinder(self):
		self.makeOutput("testDEVS/basic_BitFinder.circ")
		input0 = [	"b00000000",	"b00100100",	"bE010101x",	"bE0101010"]
		output0 = [	"b000",		"b010",		"bEEE",		"b001"]
		output1 = [	"b0",		"b1",		"bE",		"b1"]
		output2 = [	"b000",		"b101",		"bEEE",		"bEEE"]
		output3 = [	"b0",		"b1",		"bE",		"bE"]
		output4 = [	"b000",		"b000",		"bEEE",		"b000"]
		output5 = [	"b1",		"b1",		"bE",		"b1"]
		output6 = [	"b111",		"b111",		"bEEE",		"bEEE"]
		output7 = [	"b1",		"b1",		"bE",		"bE"]
		self.framework([input0], [output0, output1, output2, output3, output4, output5, output6, output7], 5)
		self.compare("basic_bitFinder")

	def test_basicSplitter(self):
		self.makeOutput("testDEVS/basic_splitter.circ")
		input0 = [	"b00","b01","b10","b11","bE0","b0x","bEx"]
		output0 = [	"b0", "b1", "b0", "b1",	"b0", "bx", "bx"]
		output1 = [	"b0", "b0", "b1", "b1",	"bE", "b0", "bE"]
		self.framework([input0], [output0, output1], 5)
		self.compare("basic_splitter")

	def test_basicSplitterReverse(self):
		self.makeOutput("testDEVS/basic_splitterReverse2.circ")
		output0 = [	"b0000","b0101","b1011","b1111","bE0E0","b1xxx","bE001"]
		input0 = [	"b0", "b1", "b1", "b1",	"b0", "bx", "b1"]
		input1 = [	"b0", "b0", "b1", "b1",	"bE", "bx", "b0"]
		input2 = [	"b0", "b1", "b0", "b1",	"b0", "bx", "b0"]
		input3 = [	"b0", "b0", "b1", "b1",	"bE", "b1", "bE"]
		self.framework([input0, input1, input2, input3], [output0], 5)
		self.compare("basic_splitterReverse")

	def test_basicDFlipFlop(self):
		self.makeOutput("testDEVS/basic_DFlipFlop.circ")
		clock = [	"b0","b0","b0","b1","b0","b0","b1","b1","bE","b1","b0","b1","b0"]
		data = [	"b0","b1","b0","b1","b1","b1","b0","b1","bE","b1","bE","b1","b1"]
		enable = [	"b1","b1","b1","b1","b1","b1","b1","b1","bE","b1","b1","b0","bx"]
		output0 = [	"b0","b0","b0","b1","b1","b1","b0","b0","b0","b0","b0","b0","b0"]
		output1 = [	"b1","b1","b1","b0","b0","b0","b1","b1","b1","b1","b1","b1","b1"]
		self.framework([clock, data, enable], [output0, output1], 5)
		self.compare("basic_dFlipFlop")

	def test_basicTFlipFlop(self):
		self.makeOutput("testDEVS/basic_TFlipFlop.circ")
		clock = [	"b0","b0","b0","b1","b0","b0","b1","b1","bE","b1","b0","b1","b0","b1"]
		data = [	"b0","b1","b0","b1","b1","b1","b0","b1","bE","b1","bE","b1","b1","b1"]
		enable = [	"b1","b1","b1","b1","b1","b1","b1","b1","bE","b1","b1","b0","bx","bx"]
		output0 = [	"b0","b0","b0","b1","b1","b1","b1","b1","b1","b1","b1","b1","b1","b0"]
		output1 = [	"b1","b1","b1","b0","b0","b0","b0","b0","b0","b0","b0","b0","b0","b1"]
		self.framework([clock, data, enable], [output0, output1], 5)
		self.compare("basic_tFlipFlop")

	def test_basicJKFlipFlop(self):
		self.makeOutput("testDEVS/basic_JKFlipFlop.circ")
		clock = [	"b0","b1","b0","b1","b0","b1","b1","b0","b1","b1","b0","b0","b0","b1","b0","b1"]
		j = [		"b1","b1","b0","b1","b1","b0","b0","b0","b1","b1","bE","b1","b1","b1","bx","b0"]
		k = [		"b0","b1","b1","b1","b1","b0","b0","b1","b0","b1","b1","b0","bx","bx","bx","b1"]
		output0 = [	"b0","b1","b1","b0","b0","b0","b0","b0","b1","b1","b1","b1","b1","b1","b1","b0"]
		output1 = [	"b1","b0","b0","b1","b1","b1","b1","b1","b0","b0","b0","b0","b0","b0","b0","b1"]
		self.framework([clock, k, j], [output0, output1], 5)
		self.compare("basic_jkFlipFlop")


	def test_basicSRFlipFlop(self):
		self.makeOutput("testDEVS/basic_SRFlipFlop.circ")
		clock = [	"b0","b1","b0","b1","b0","b1","b1","b0","b1","b1","b0","b0","b0","b1","b0","b1"]
		r = [		"b1","b1","b0","b0","b1","b0","b0","b0","b1","b1","bE","b1","b1","b0","bx","b0"]
		s = [		"b0","b1","b1","b1","b1","b0","b0","b1","b0","b1","b1","b0","bx","bx","bx","b1"]
		output0 = [	"b0","b0","b0","b1","b1","b1","b1","b1","b0","b0","b0","b0","b0","b0","b0","b1"]
		output1 = [	"b1","b1","b1","b0","b0","b0","b0","b0","b1","b1","b1","b1","b1","b1","b1","b0"]
		self.framework([clock, r, s], [output0, output1], 5)
		self.compare("basic_rsFlipFlop")

	def test_basicRegister(self):
		self.makeOutput("testDEVS/basic_Register.circ")
		clock = [	"b0",   "b1",   "b0",   "b1",   "b0",   "b1",   "b0",   "b1",   "b0",   "b1",   "b0"]
		data = [	"b0011","b1010","b0000","b1x01","b1111","b0001","b110E","bE1EE","b1x1E","b1110","b0000"]
		output0 = [	"b0000","b1010","b1010","b1010","b1010","b0001","b0001","b0001","b0001","b1110","b1110"]
		self.framework([data, clock], [output0], 5)
		self.compare("basic_register")

	def test_basicCounter(self):
		self.makeOutput("testDEVS/basic_Counter.circ")
		clock = [	"b0",   "b1",   "b0",   "b1",   "b0",   "b1",   "b0",   "b1",   "b0",   "b1",   "b0",	"b1"]
		data = [	"b0011","b1010","b0000","b1x01","b1111","b1001","b1001","bE1EE","b1x1E","b1111","b0010","b1100"]
		load = [	"b0",	"b0",	"b1",	"b1",	"b0",	"b1",	"bx",	"b1",	"b0",	"b1",	"b0",	"b0"]
		count = [	"b1",	"b1",	"b1",	"b1",	"b0",	"b0",	"bE",	"b0",	"b0",	"b0",	"b0",	"b1"]
		# Wrap around
		output0 = [	"b0000","b0001","b0001","b0000","b0000","b1001","b1001","b0000","b0000","b1111","b1111","b0000"]
		carry0 = [	"b0",   "b0",   "b0",   "b1",   "b0",   "b0",   "b0",   "b0",   "b0",   "b1",   "b1",	"b0"]
		# Stay the same
		output1 = [	"b0000","b0001","b0001","b0000","b0000","b1001","b1001","b0000","b0000","b1111","b1111","b1111"]
		carry1 = [	"b0",   "b0",   "b0",   "b1",   "b0",   "b0",   "b0",   "b0",   "b0",   "b1",   "b1",	"b1"]
		# Continue counting
		output2 = [	"b0000","b0001","b0001","b0000","b0000","b1001","b1001","b0000","b0000","b1111","b1111","b0000"]
		carry2 = [	"b0",   "b0",   "b0",   "b1",   "b0",   "b0",   "b0",   "b0",   "b0",   "b1",   "b1",	"b0"]
		# Load value
		output3 = [	"b0000","b0001","b0001","b0000","b0000","b1001","b1001","b0000","b0000","b1111","b1111","b1100"]
		carry3 = [	"b0",   "b0",   "b0",   "b1",   "b0",   "b0",   "b0",   "b0",   "b0",   "b1",   "b1",	"b0"]
		self.framework([load, data, count, clock], [output0, carry0, output1, carry1, output2, carry2, output3, carry3], 5)
		self.compare("basic_counter")

	def test_basicShiftRegister(self):
		self.makeOutput("testDEVS/basic_ShiftRegister.circ")
		clock = [	"b0","b1","b0","b1","b0","b1","b1","b0","b1","b1","b0","b0","b0","b1","b0","b1"]
		load = [	"b1","b1","b0","b0","b1","b0","b0","b0","b1","b1","bE","b1","b1","b0","bx","b0"]
		shift = [	"b0","b1","b1","b1","b1","b0","b0","b1","b0","b1","b1","b0","bx","b1","bx","b1"]
		inputnew = [	"b0","b0","b0","b1","b1","b1","b1","b1","b0","b0","b0","b0","b0","b0","b0","b1"]
		input0 = [	"b0","b0","b0","b1","b1","b1","b1","b1","b0","b0","b0","b0","b0","b0","b0","b1"]
		input1 = [	"b1","b1","b1","b0","b0","b0","b0","b0","b1","b1","b1","b1","b1","b1","b1","b0"]
		input2 = [	"b1","b1","b1","b0","b0","b0","b0","b0","b1","b1","b1","b1","b1","b1","b1","b0"]
		output0 = [	"b0","b0","b0","b1","b1","b1","b1","b1","b0","b0","b0","b0","b0","b0","b0","b1"]
		output1 = [	"b0","b1","b1","b0","b0","b0","b0","b0","b1","b1","b1","b1","b1","b0","b0","b0"]
		output2 = [	"b0","b1","b1","b1","b1","b1","b1","b1","b1","b1","b1","b1","b1","b1","b1","b0"]
		outputfin = output2
		self.framework([inputnew, shift, clock, load, input0, input1, input2], [outputfin, output0, output1, output2], 5)
		self.compare("basic_shiftRegister")

	# Hard to test because the name of the file is called depending on the time
	def test_basicROM(self):
		self.makeOutput("testDEVS/basic_ROM.circ")
		input0 = [	"b0000","b0001","b0010","b0011","b0100","b0101","b0110","b0111","b1000","b1001","b1010","b1011","b1100","b1101","b1110","b1111"]
		output0 = [	"b1000","b1010","b0000","b0101","b0001","b0001","b1011","b1111","b0000","b0000","b0100","b1001","b1010","b0000","b0000","b0000"]
		self.framework([input0], [output0], 5)
		self.compare("basic_rom")

	def test_basicRAMseparate(self):
		self.makeOutput("testDEVS/basic_RAMseparate.circ")
		clock = [	"b0",   "b1",   "b0",   "b1",   "b0",   "b1",   "b0",   "b1",   "b0",   "b1",   "b0",	"b1",	"b0",	"b1"]
		store = [	"b0",   "b1",   "b0",   "b1",   "b0",   "b0",   "b0",   "b0",   "b1",   "b1",   "b0",	"b1",	"bx",	"b1"]
		load = [	"b1",   "b0",   "b1",   "b1",   "b1",   "b1",   "b0",   "b0",   "b1",   "b1",   "b1",	"b1",	"bx",	"b1"]
		address = [	"b0000","b0010","b1000","b1010","b0010","b1010","b0001","b0001","b0001","b1010","b1010","b0000","bxxxx","b1111"]
		data = [	"b0111","b1110","b0110","b1101","b1000","b1100","b110E","b0100","b1100","b1111","b1000","b0xx0","bxxxx","b000E"]
		output0 = [	"b0000","bxxxx","b0000","b1101","b1110","b1101","bxxxx","bxxxx","b0000","b1111","b1111","b1111","b1111","b1111"]
		self.framework([address, data, clock, load, store], [output0], 5)
		self.compare("basic_ramseparate")

	def test_basicRAMasynch(self):
		self.makeOutput("testDEVS/basic_RAMasynch.circ")
		load = [	"b1",   "b0",   "b1",   "b1",   "b0",   "b1",   "b0",   "b1",   "b0",   "b1",   "b1",	"b0",	"b1",	"b0",	"b1",	"b1",	"b0"]
		address = [	"b0000","b0010","b1000","b0010","b0010","b0010","b0000","b0001","b0001","b0010","b0001","b0000","b0000","b1111","b1111","bxxx1","bxxx1"]
		data = [	"b0111","b1110","b0110","b1110","b0100","b0100","b1101","b0100","b1111","b1101","b1000","b0xx0","b0000","b101E","b0000","b0101","b1010"]
		output0 = [	"b0000","b1110","b0000","b1110","b0100","b0100","b1101","b0000","b1111","b0100","b1111","b0xx0","b1111","b101E","b1111","b1111","b1E1E"]
		self.framework([address, data, load], [output0], 10)
		self.compare("basic_ramasynch")

	def test_basicRAMsynch(self):
		# It is needed to put the data good already, otherwise a race condition occurs (that will certainly cause problems)
		self.makeOutput("testDEVS/basic_RAMsynch.circ")
		clock = [	"b0",   "b1",   "b0",   "b1",   "b0",   "b1",   "b0",   "b1",   "b0",   "b1",   "b0"]
		load = [	"b1",   "b0",   "b1",   "b1",   "b1",   "b1",   "b0",   "b0",   "b1",   "b0",   "b1"]
		address = [	"b0000","b0010","b1000","b1010","b0010","b1010","b0001","b0001","b0001","b1010","b1010"]
		data = [	"b1110","b1110","b0110","b1101","b1000","b1100","b0100","b0100","b1111","b1111","b1000"]
		output0 = [	"b0000","b1110","b0000","b0000","b1110","b0000","b0100","b0100","b0100","b1111","b1111"]
		self.framework([address, data, load, clock], [output0], 5)
		self.compare("basic_ramsynch")

	def test_basicSplitterBoth(self):
		self.makeOutput("testDEVS/basic_splitterBoth.circ")
		input0 = [	"b00","b01","b10","b11","b01"]
		input1 = [	"b0", "b0", "b1", "b1", "b1"]
		output0 = [	"b00","b01","b10","b11","bE1"]
		output1 = [	"b0", "b1", "b0", "b1", "b1"]
		self.framework([input0, input1], [output0, output1], 5)
		self.compare("basic_splitterBoth")

	def test_advSplitterBoth(self):
		self.makeOutput("testDEVS/adv_splitterBoth.circ")
		input0 = [	"b0", "b1", "bE", "bx"]
		output0 = [	"b0", "b1", "bE", "bx"]
		self.framework([input0], [output0], 5)
		self.compare("adv_splitterBoth")

	def test_basicSplitterMerger(self):
		self.makeOutput("testDEVS/basic_splitterMerge.circ")
		input0 = [	"b00","b01","b10","b11","b01","b11"]
		input1 = [	"b0", "b0", "b1", "b1", "b1", "b0"]
		input2 = [	"b00","b01","b00","b11","b01","b10"]
		output0 = [	"b00","b01","bE0","b11","bE1","bEE"]
		output1 = [	"b0", "b1", "b0", "b1", "b1", "bE"]
		self.framework([input0, input1, input2], [output0, output1], 5)
		self.compare("basic_splitterMerger")

	def test_basicSplitterCross(self):
		self.makeOutput("testDEVS/basic_splitterCross.circ")
		input0 = [	"b0110101010100110","b0101001010010111","b1010010010010010","b1001010010010110"]
		output0 = [	"b0001","b0011","b1011","b1011"]
		output1 = [	"b110","b100","b000","b000"]
		output2 = [	"b10","b11","b10","b10"]
		output3 = [	"b11","b01","b00","b00"]
		output4 = [	"b11","b01","b10","b01"]
		output5 = [	"b000","b010","b100","b110"]
		self.framework([input0], [output0, output1, output2, output3, output4, output5], 5)
		self.compare("basic_splitterCross")

	def test_parseNonDefaultMain(self):
		self.makeOutput("testDEVS/parse_nonDefault.circ")
		setConfigLine("experiment", "circuit", "theRealOne")
		input0 = [	"b0","b1","bE","bx"]
		output0 = [	"b1","b0","bE","bE"]
		self.framework([input0], [output0], 5)
		self.compare("parse_nonDefaultMain")

	def test_basicSplitterPull(self):
		self.makeOutput("testDEVS/basic_SplitterPull.circ")
		input0 = [	"b00","b11","b10","b01","bxx","bx1","bE0","b0x"]
		output0 = [	"b00","b11","b10","b01","b10","b11","bE0","b00"]
		self.framework([input0], [output0], 5)
		self.compare("basic_splitterPull")

	def test_advSplitterLoop(self):
		self.makeOutput("testDEVS/adv_SplitterLoop.circ")
		input0 = [	"b00","b11","b10","b01","bxx","bEx","bE0","b0x"]
		output0 = [	"b00","b11","b10","b01","bxx","bEx","bE0","b0x"]
		self.framework([input0], [output0], 5)
		self.compare("adv_splitterLoop")

	def test_basicSplitterMulti(self):
		self.makeOutput("testDEVS/basic_splitterMulti.circ")
		input0 = [	"b00","b11","b10","b01","bxx","bEx","bE0","b0x"]
		input1 = [	"b00","b11","b10","b01","bxx","bEx","bE0","b0x"]
		output0 = [	"b1", "b0", "b1", "b0", "bE", "bE", "b1", "bE"]
		output1 = [	"b1", "b0", "b0", "b1", "bE", "bE", "bE", "b1"]
		output2 = [	"b0", "b1", "b0", "b1", "bx", "bx", "b0", "bx"]
		output3 = [	"b0", "b1", "b1", "b0", "bx", "bE", "bE", "b0"]
		self.framework([input0, input1], [output0, output1, output2, output3], 5)
		self.compare("basic_splitterMulti")

	def test_basicSplitterBig(self):
		self.makeOutput("testDEVS/basic_splitterBig.circ")
		input0 = [	"b" + "0" * 32, \
				"b" + "1" * 32, \
				"b" + "x" * 32]
		input1 = [	"b" + "0" * 32, \
				"b" + "1" * 32, \
				"b" + "x" * 32]
		input2 = [	"b" + "0" * 32, \
				"b" + "1" * 32, \
				"b" + "x" * 32]
		output0 = [	"b" + "0" * 32, \
				"b" + "1" * 32, \
				"b" + "x" * 32]
		output1 = [	"b" + "0" * 32, \
				"b" + "1" * 32, \
				"b" + "x" * 32]
		output2 = [	"b" + "0" * 32, \
				"b" + "1" * 32, \
				"b" + "x" * 32]
		self.framework([input0, input1, input2], [output0, output1, output2], 5)
		self.compare("basic_splitterBig")

	def test_basicSplitterStrange(self):
		self.makeOutput("testDEVS/basic_splitterStrange.circ")
		input0 = [	"b00000001","b00000010","b00000100","b00001000","b00010000","b00100000","b01000000","b10000000"]
		output0 = [	"b00000001","b00010000","b00000010","b01000000","b00100000","b10000000","b00000100","b00001000"]
		self.framework([input0], [output0], 5)
		self.compare("basic_splitterStrange")

	def test_basicPullFloat(self):
		self.makeOutput("testDEVS/basic_pullFloating.circ")
		output0 = [	"b0"]
		self.framework([], [output0], 5)
		self.compare("basic_pullFloating")

	def test_helperDualNull(self):
		self.makeOutput("testDEVS/helper_dualNull.circ")
		input0 = [	"b0", "b1", "bx", "bE", "b0", "b1", "bx", "bE", "b0", "b1", "bx", "bE", "b0", "b1", "bx", "bE"]
		input1 = [	"b0", "b0", "b0", "b0", "b1", "b1", "b1", "b1", "bx", "bx", "bx", "bx", "bE", "bE", "bE", "bE"]
		output0 = [	"b0", "b1", "bx", "bE", "b0", "b1", "bx", "bE", "b0", "b1", "bx", "bE", "b0", "b1", "bx", "bE"]
		output1 = [	"b0", "b0", "b0", "b0", "b1", "b1", "b1", "b1", "bx", "bx", "bx", "bx", "bE", "bE", "bE", "bE"]
		self.framework([input0, input1], [output0, output1], 5)
		self.compare("helper_dualNull")

	def test_advALU(self):
		self.makeOutput("testDEVS/adv_ALU.circ")
		input0 = [	"b1110x01000101000","b0101000x010E0001","b0101000x010E0001","b010x0E010x0E0100","b0E010x0010101000","b00x0001001001011","b0101001000100100","b1101000100E00xx1","b1101000100E00xx1"]
		input1 = [	"b010100100x000010","b010001010x000E0E","b010001010x000E0E","bE0E0E010x0x00E01","b0100E00100E00x1E","b0000x00001100011","b000x000001010110","b00E00x00010010x0","bx1E00x00100E0E01"]
		opcode = [	"b000000",	    "b000001",		"b000010",	    "b000011",		"b000100",	    "b000101",		"b000110",	    "b000111",		"b001000"]
		output0 = [	"b0000000000000000","b0100000E0E00000E","b01010101010E0E01","b101E1E101E1E1011","b1E101x1101011000","bxxxxx01010101110","bEEEE000111001110","b101000100E00xx10","b01101000100E00xx"]
		output1 = [	"b0", "b0", "b0", "b0", "b0", "bE", "bE", "b0", "b0"]
		input0.extend([	"b1010x01000101000","b1101000x010E0001","b0101000x01000001","b0101000x01000001","b0101000x01000001","b00x0001001001011","b0101001000100100","b1101000100E00xx1","b1101000100E00xx1"])
		input1.extend([	"b010100100x000010","b010001010x000E0E","b010001010x000100","b010001010x001000","b010001010x001000","b0000x00001100011","b000x000001010110","b00E00x00010010x0","bx1E00x00100E0E01"])
		opcode.extend([	"b001001",	    "b001010",		"b001011",	    "b001100",		"b001101",	    "b001110",		"b001111",	    "b010000",		"b100100"])
		output0.extend(["b110x010001010000","b11101000x010E000","b0000000000000000","b0000000000000001","b0000000000000000","b0000000000000000","b0000000000000000","b0000000000000000","b0000000000000000"])
		output1.extend(["b0", "b0", "b0", "b0", "b0", "b1", "b1", "b1", "b1"])
		self.framework([input0, input1, opcode], [output0, output1], 5)
		self.compare("adv_ALU")

	def test_failBitsize(self):
		try:
			self.makeOutput("testDEVS/fail_bitsize.circ")
		except Exception as e:
			return
		self.fail()

	def test_basicLoop(self):
		self.makeOutput("testDEVS/basic_loop.circ")
		input0 = [	"b0", "b1", "bx", "bE"]
		output0 = [	"b0", "b0", "b0", "b0"]
		self.framework([input0], [output0], 5)
		self.compare("basic_loop")

	def test_advTrafficLights(self):
		self.makeOutput("testDEVS/adv_trafficLights.circ")
		output0 = [	"b1", "b1", "b1", "b1", "b1", "b1", "b1", "b1", "b1", "b1", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b1", "b1", "b1", "b1", "b1", "b1", "b1", "b1", "b1", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b1"]
		output1 = [	"b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b1", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b1", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0"]
		output2 = [	"b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b1", "b1", "b1", "b1", "b1", "b1", "b1", "b1", "b1", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b0", "b1", "b1", "b1", "b1", "b1", "b1", "b1", "b1", "b1", "b0"]
		self.framework([], [output0, output1, output2], 10)
		self.compare("adv_trafficLights")

	def test_advSplitterFest(self):
		self.makeOutput("testDEVS/adv_splitterFest.circ")
		input0 = [	"b0000000000000001","b0000000000000010","b0000000000000100","b0000000000001000","b0000000000010000","b0000000000100000","b0000000001000000","b0000000010000000"]
		output0 = [	"b0000000000001000","b0000000000000001","b0000000000010000","b0000100000000000","b0000010000000000","b0010000000000000","b0000000000000010","b0000000010000000"]
		input0.extend([	"b0000000100000000","b0000001000000000","b0000010000000000","b0000100000000000","b0001000000000000","b0010000000000000","b0100000000000000","b1000000000000000"])
		output0.extend(["b0001000000000000","b1000000000000000","b0000000000000100","b0000000100000000","b0100000000000000","b0000001000000000","b0000000000100000","b0000000001000000"])
		self.framework([input0], [output0], 5)
		self.compare("adv_splitterFest")

	def test_basicTunnel(self):
		self.makeOutput("testDEVS/basic_tunnel.circ")
		input0 = [	"b0", "b1", "b0", "b1"]
		input1 = [	"b0", "b1", "b1", "b0"]
		input2 = [	"b0", "b1", "b1", "b1"]
		output0 = [	"b0", "b1", "b0", "b1"]
		output1 = [	"b0", "b1", "b1", "bE"]
		self.framework([input0, input1, input2], [output0, output1], 5)
		self.compare("basic_tunnel")

	def test_basicProbe(self):
		self.makeOutput("testDEVS/basic_probe.circ")
		input0 = [	"b0", "b1", "b0", "b1"]
		output0 = [	"b0", "b1", "b0", "b1"]
		self.framework([input0], [output0], 5)
		self.compare("basic_probe")

	def test_parseAppearance(self):
		self.makeOutput("testDEVS/parse_appearance.circ")
		input0 = [	"b0", "b1", "bx", "bE"]
		output0 = [	"b1", "b0", "bE", "bE"]
		self.framework([input0], [output0], 5)
		self.compare("parse_appearance")
