#!/usr/bin/env python3

import unittest      

from translator.intermediateBuilder.intermediateBuilder import IntermediateBuilder
from translator.visitor.intermediateVisitor.intermediateNamingVisitor import IntermediateNamingVisitor
from translator.visitor.intermediateVisitor.intermediateConnectivityVisitor import IntermediateConnectivityVisitor

class TestConnectivityVisitor(unittest.TestCase):
	""" Tests the connectivity tester of the intermediate tree"""
	def setUp(self):
		self.namingVisitor = IntermediateNamingVisitor()
		self.connectivityVisitor = IntermediateConnectivityVisitor()
		
	def tearDown(self):
		pass
	
	def makeTree(self, filename, bool= False):
		xml = IntermediateBuilder(filename, bool)
		intermediateTree = xml.makeIntermediateTree()[0]
		return intermediateTree
		
	def test_connectivityVisitor(self):		
		intermediateTree = self.makeTree('testTestingVisitor/simpleConnectivityVisitor.circ', True)	
		#We want to see whether there is a not gate that has output wires that lead to something
		dictNot = dict()
		dictNot["class"] = "NotGate"
		dictElement = dict()
		dictElement["class"] = "Node"
		self.connectivityVisitor.setChecks(dictNot, dictElement, False, True)
		intermediateTree.traverse(self.connectivityVisitor, False)
		result = self.connectivityVisitor.result()
		#There are 5 not gates present
		self.assertEqual(len(result), 5)
		#Not2 is connected with a controlled buffer, not has no nodes attached to it, ...
		self.assertEqual(result, [1, 0, 6, 1, 1])
		#We want to see whether the not gate with name "not" has output wires 
		dictNot = dict()
		dictNot["class"] = "NotGate"
		dictNot["name"] = "not"
		dictElement = dict()
		dictElement["class"] = "Element"
		self.connectivityVisitor.setChecks(dictNot, dictElement, False, True)
		intermediateTree.traverse(self.connectivityVisitor, False)
		result = self.connectivityVisitor.result()
		self.assertEqual(result, [3])
		#We want to see whether the in1 base has wires connected to it
		dictBase = dict()
		dictBase["class"] = "InputBase"
		dictBase["name"] = "in1"
		dictElement = dict()
		dictElement["class"] = "Element"
		self.connectivityVisitor.setChecks(dictBase, dictElement, False, True)
		intermediateTree.traverse(self.connectivityVisitor, False)
		result = self.connectivityVisitor.result()
		self.assertEqual(result, [0])				
		#We want to see whether there is a node with 3 bits that is connected to another node
		dictBase = dict()
		dictBase["class"] = "Node"
		dictBase["bits"] = 3
		dictElement = dict()
		dictElement["class"] = "Element"
		self.connectivityVisitor.setChecks(dictBase, dictElement)
		intermediateTree.traverse(self.connectivityVisitor, False)
		result = self.connectivityVisitor.result()
		self.assertEqual(result, [False, False])			
		#We want to see whether in2 is connected with out3 (is via a long chain connected)
		dictBase = dict()
		dictBase["class"] = "InputBase"
		dictBase["name"] = "in2"
		dictBase["bits"] = 1
		dictBase2 = dict()
		dictBase2["class"] = "OutputBase"
		dictBase2["name"] = "out3"
		dictBase2["bits"] = 1
		self.connectivityVisitor.setChecks(dictBase, dictBase2)
		intermediateTree.traverse(self.connectivityVisitor, False)
		result = self.connectivityVisitor.result()
		self.assertEqual(result, [True, False])		
		#We want to see whether out3 is connected with in2 (and the direction matters)
		dictBase = dict()
		dictBase["class"] = "InputBase"
		dictBase["name"] = "in2"
		dictBase["bits"] = 1
		dictBase2 = dict()
		dictBase2["class"] = "OutputBase"
		dictBase2["name"] = "out3"
		dictBase2["bits"] = 1
		self.connectivityVisitor.setChecks(dictBase2, dictBase, False)
		intermediateTree.traverse(self.connectivityVisitor, False)
		result = self.connectivityVisitor.result()
		self.assertEqual(result, False)		
		#See if the out1 base has an input and direction matters
		dictBase = dict()
		dictBase["class"] = "OutputBase"
		dictBase["name"] = "out1"
		dictBase["bits"] = 1
		dictElement = dict()
		dictElement["class"] = "BaseNode"
		self.connectivityVisitor.setChecks(dictElement, dictBase, False)
		intermediateTree.traverse(self.connectivityVisitor, False)
		result = self.connectivityVisitor.result()
		self.assertEqual(result, True)		
		#See if in5 is connected with out6 (collides with col)
		dictBase = dict()
		dictBase["class"] = "OutputBase"
		dictBase["name"] = "out6"
		dictBase["bits"] = 1
		dictBase2 = dict()
		dictBase2["class"] = "InputBase"
		dictBase2["name"] = "col"
		dictBase2["bits"] = 1
		self.connectivityVisitor.setChecks(dictBase, dictBase2)
		intermediateTree.traverse(self.connectivityVisitor, False)
		result = self.connectivityVisitor.result()
		self.assertEqual(result, [False, True])
		#See if inS is connected with a splitter
		dictBase = dict()
		dictBase["class"] = "InputBase"
		dictBase["name"] = "inS"
		dictBase["bits"] = 2
		dictBase2 = dict()
		dictBase2["class"] = "OutputBase"
		dictBase2["name"] = "outS"
		dictBase2["bits"] = 2
		self.connectivityVisitor.setChecks(dictBase, dictBase2)
		intermediateTree.traverse(self.connectivityVisitor, False)
		result = self.connectivityVisitor.result()
		self.assertEqual(result, [True, False])	
		#See if inS is connected with itself
		dictBase = dict()
		dictBase["class"] = "InputBase"
		dictBase["name"] = "inS"
		dictBase["bits"] = 2
		dictBase2 = dict()
		dictBase2["class"] = "InputBase"
		dictBase2["name"] = "inS"
		dictBase2["bits"] = 2
		self.connectivityVisitor.setChecks(dictBase, dictBase2)
		intermediateTree.traverse(self.connectivityVisitor, False)
		result = self.connectivityVisitor.result()
		self.assertEqual(result, [False, False])	
		#See how many clocks are connected with a flipflop
		dictBase = dict()
		dictBase["class"] = "ClockBase"
		dictBase["highDuration"] = 10
		dictBase["lowDuration"] = 1
		dictMemory = dict()
		dictMemory["class"] = "FlipFlop"
		dictMemory["trigger"] = "rising"
		self.connectivityVisitor.setChecks(dictBase, dictMemory, True, True)
		intermediateTree.traverse(self.connectivityVisitor, False)
		result = self.connectivityVisitor.result()
		self.assertEqual(result, [[0, 3], [0, 0, 0]])
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		