#!/usr/bin/env python3

# First check if it is invoked with Python3
import sys
if sys.version_info[0] != 3:
	print("Python 3 is required!")
	sys.exit(2)

from translator.intermediateBuilder.intermediateBuilder import IntermediateBuilder
from translator.intermediateBuilder.circuit import Circuit
from translator.visitor.verilogVisitor.verilogWriteVisitor import VerilogWriteVisitor
from translator.visitor.intermediateVisitor.intermediateTranslatorVisitor import IntermediateTranslatorVisitor
from translator.visitor.intermediateVisitor.intermediateNamingVisitor import IntermediateNamingVisitor
from translator.visitor.intermediateVisitor.intermediateConnectivityVisitor import IntermediateConnectivityVisitor
from translator.visitor.intermediateVisitor.intermediateIndependentVisitor import IntermediateIndependentCheckVisitor
from translator.visitor.DEVSlibrary.devsTranslatorVisitor import DEVSTranslatorVisitor
from translator.visitor.DEVSlibrary.helper import *
from translator.visitor.DEVSlibrary.devsConnect import DEVSConnect

from gui import Gui

import os
import argparse

# Parse from command line arguments

if __name__ == '__main__':
	good = False
	if len(sys.argv) == 1:
		print("Launching gui!")
		gui = Gui()
		conf = gui.launch(None)
	elif (len(sys.argv) == 2) and ((sys.argv[1] != "--help") or (sys.argv[1] != "-h")):
		print("Launching gui!")
		gui = Gui()
		conf = gui.launch(sys.argv[1])
	else:
		parser = argparse.ArgumentParser(description='Translate Logisim circ-files to DEVS py-files or Verilog files')
		parser.add_argument('filename', metavar='FILE', action='store', help='Defines the circ-file to be translated')
		parser.add_argument('--config', metavar='CONF', action='store', default="config.ini", help='Defines the config file to be used (default: config.ini)')
		parser.add_argument('--devs', action='store_true', help='Should generate DEVS files')
		parser.add_argument('--verilog', action='store_true', help='Should generate Verilog files')
		conf = parser.parse_args()
		good = True
	
	if not good and conf == None:
		print("Config parsing failed, exiting")
		sys.exit(3)

	# Don't let this be user passed, since it would just cause trouble
	conf.testing = False
	
	setConfig(conf.config)
	
	if not conf.devs and not conf.verilog:
		warn("Not generating anything, since no output specified")
		exit(4)
	
	if conf.devs and not conf.verilog:
		setProgress(13)
	elif conf.devs and conf.verilog:
		setProgress(17)
	elif not conf.devs and conf.verilog:
		setProgress(8)
	
	setConfigLine("main", "filename", conf.filename)
	# Just to be sure, we will make a new directory for the files and mark this as the destination
	filename = conf.filename.split("/")[-1].split(".")[0]
	try:
		os.mkdir(getConfig("main", "destination") + filename + "/")
	except OSError:
		# Seems the folder already exists, just overwrite what is there
		pass
	setConfigLine("main", "destination", getConfig("main", "destination") + filename + "/")
	
	debug("Argument parsing done")
	info("Starting parsing of file... " + progress())
	builder = IntermediateBuilder(getConfig("main", "filename"), getConfig("main", "allowCollision", "bool"))
	info("Starting intermediate tree construction... " + progress())
	returnValue = builder.makeIntermediateTree()
	intermediateTree = returnValue[0]
	
	info("Naming all intermediate elements... " + progress())
	#We traverse the intermediateTree with a naming visitor, so that all the components get a valid name
	namingVisitor = IntermediateNamingVisitor()
	intermediateTree.traverse(namingVisitor)	
	
	info("Connecting all intermediate elements... " + progress())
	#With this visitor we can check whether 2 elements are connected
	connectivityVisitor = IntermediateConnectivityVisitor()
	#Check whether there is a flipflop that has an output element
	dictFlipFlop = dict()
	dictFlipFlop["class"] = "FlipFlop"
	dictElement = dict()
	dictElement["class"] = "Element"
	connectivityVisitor.setChecks(dictFlipFlop, dictElement)
	intermediateTree.traverse(connectivityVisitor, False)
	result = connectivityVisitor.result()
	
	#With this visitor we can check the presence of elements
	independentCheckVisitor = IntermediateIndependentCheckVisitor()
	#Checks how many memory objects there are and if there is a clock present
	dictMem = dict()
	dictMem["class"] = "MemoryNode"
	dictClock = dict()
	dictClock["class"] = "ClockBase"
	listOfAttributeMaps = [dictMem, dictClock]
	independentCheckVisitor.setChecks(listOfAttributeMaps)
	intermediateTree.traverse(independentCheckVisitor, False)
	results = independentCheckVisitor.results()
	
	if conf.devs and conf.verilog:
		# Backup the tree to prevent interference from the other process
		from copy import deepcopy
		debug("Backing up intermediate tree")
		copied = deepcopy(intermediateTree)
	
	if conf.devs:
		info("Generating connection blocks... " + progress())
		translatorVisitor = DEVSTranslatorVisitor()
		intermediateTree.traverse(translatorVisitor)
	
		info("Grouping colliding wires... " + progress())
		translatorVisitor.makeCollide()
	
		info("Translating elements to DEVS... " + progress())
		connectVisitor = DEVSConnect()
		connectVisitor.setConnections(translatorVisitor.connections, translatorVisitor.colliding, translatorVisitor.wires)
		intermediateTree.traverse(connectVisitor)
	
		info("Fixing anomalies... " + progress())
		connectVisitor.afterTranslation()
	
		debug(connectVisitor.printConnections())
	
		if getConfig("speed", "staticChecks", "bool"):
			debug("Checking bitsizes...")
			connectVisitor.checkBitsizes()
	
		info("Writing elements to files... " + progress())
		connectVisitor.write()
	
		info("DEVS is done! " + progress())
	if conf.verilog:
		if conf.devs:
			debug("Restoring intermediate tree from backup")
			intermediateTree = copied
		info("Translating elements to Verilog... " + progress())
		translateVisitor = IntermediateTranslatorVisitor()
		intermediateTree.traverse(translateVisitor)
	
		info("Generating Verilog tree... " + progress())
		verilogTree = translateVisitor.makeVerilogTree()
	
		info("Writing Verilog data to files..." + progress())
		writeVisitor = VerilogWriteVisitor()
		verilogTree.traverse(writeVisitor)
		writeVisitor.writeToFiles(getConfig("main", "destination"))
		info("Verilog is done! " + progress())
	info("The files are written to the folder " + str(getConfig("main", "destination")))
