#!/usr/bin/env python3

from translator.intermediateBuilder.annotated.annotatedNode import AnnotatedNode
from translator.intermediateBuilder.annotated.annotatedInputBase import AnnotatedInputBase
from translator.intermediateBuilder.referencedOutput import ReferencedOutput
from translator.intermediateBuilder.referencedInput import ReferencedInput
from translator.intermediateBuilder.helperFunctions import sub
from translator.intermediateBuilder.helperFunctions import add
from itertools import groupby
from operator import attrgetter

class Facing():
	""" Enumeration class for facings """
	west=0
	south=1
	east=2
	north=3

class AnnotatedImportedCircuit(AnnotatedNode):	
	""" AnnotatedImportedCircuit is the annotated version of the imported circuit """
	
	def __init__(self, circuit, coordinates, facing, pinList):
		""" Constructor """
		AnnotatedNode.__init__(self, circuit, coordinates, facing)	
		#The list of tuples with input coordinates and their respective bases
		#When we initialize the annotated imported circuit we place all the pins that belong to this circuit in the inputCoordinateTuples 
		self.inputCoordinateTuples = pinList
		self.outputCoordinateTuples = []

	def findExtraCoordinates(self):
		""" Overwrites the parent function (but has the same meaning) """
		pinList = self.inputCoordinateTuples
		#We clean the temporary storage room, so that it can store the tuples that we are going to make
		self.inputCoordinateTuples = []
		
		def sortPinList(annotatedPinList):
			""" We need to group and sort the pinList first
			We sort the pin list according to their facings and then according to their position """
			groupedFacingList = []
			annotatedPinList = sorted(annotatedPinList, key = attrgetter('facing'))
			#We group the tunnels according to their facing ((facing1, (pin1 with facing1, pin2 with facing1, ...)), (facing2, (pin1 with facing2), ...))
			for key, group in groupby(annotatedPinList, key=attrgetter('facing')):
				samegroup = []
				for pin in group:
					samegroup.append(pin)
				groupedFacingList.append((key, samegroup))			

			#We sort the groups according to their location
			sortedDict = dict()
			#The first, we sort according to the least important coordinate and then according to the most important coordinate
			for facing in groupedFacingList:
				listToSort = facing[1]
				if facing[0] == "west" or facing[0] == "east":
					#Sort from left to right (so on x coordinates)
					listToSort = sorted(listToSort, key=lambda tup: tup.coordinates[0])
					#From above to below (so on y coordinates)
					listToSort = sorted(listToSort, key=lambda tup: tup.coordinates[1])
				else:
					#From above to below (so on y coordinates)
					listToSort = sorted(listToSort, key=lambda tup: tup.coordinates[1])
					#Sort from left to right (so on x coordinates)
					listToSort = sorted(listToSort, key=lambda tup: tup.coordinates[0])
				sortedDict[facing[0]] = listToSort
			return sortedDict	
		
		sortedDict = dict()
		sortedDict = sortPinList(pinList)
		#We need to know with which pin the saved coordinates match with
		#So we need to know which facings are present in the sortedDict	
		numberOfFacings = self.getNumberOfPinsPerFacing(sortedDict)
		
		#For the further coordinate calculations we distinguish between the 4 facing the imported circuit can have
		if self.facing == "east":
			startingPoints = self.calculateStartingPointsEastFacing(numberOfFacings)
			self.findOtherCoordinates(sortedDict, numberOfFacings, startingPoints, (0, 10), (10, 0))
		elif self.facing == "south":
			startingPoints = self.calculateStartingPointsSouthFacing(numberOfFacings)	 
			self.findOtherCoordinates(sortedDict, numberOfFacings, startingPoints, (-10, 0), (0, 10))
		elif self.facing == "west":
			startingPoints = self.calculateStartingPointsWestFacing(numberOfFacings)	
			self.findOtherCoordinates(sortedDict, numberOfFacings, startingPoints, (0, -10), (-10, 0))		
		elif self.facing == "north":
			startingPoints = self.calculateStartingPointsNorthFacing(numberOfFacings)	
			self.findOtherCoordinates(sortedDict, numberOfFacings, startingPoints, (10, 0), (0, -10))	

	def findOtherCoordinates(self, sortedDict, numberOfFacings, startingPoints, changeWestEast, changeNorthSouth): 			
		""" Calculates the other coordinates starting from the starting points """
		#Calculate the other coordinates and set the matching base
		numberOfFacings[Facing.west] > 0 and self.setInOutputCoordinateTuples(numberOfFacings[Facing.west], sortedDict["west"], startingPoints[Facing.west], changeWestEast[0], changeWestEast[1])
		numberOfFacings[Facing.east] > 0 and self.setInOutputCoordinateTuples(numberOfFacings[Facing.east], sortedDict["east"], startingPoints[Facing.east], changeWestEast[0], changeWestEast[1])
		numberOfFacings[Facing.south] > 0 and self.setInOutputCoordinateTuples(numberOfFacings[Facing.south], sortedDict["south"], startingPoints[Facing.south], changeNorthSouth[0], changeNorthSouth[1])
		numberOfFacings[Facing.north] > 0 and self.setInOutputCoordinateTuples(numberOfFacings[Facing.north], sortedDict["north"], startingPoints[Facing.north], changeNorthSouth[0], changeNorthSouth[1])

	def setCoordinatesToStartingPoint(self, numberOfFacings):	
		""" The starting point of one of the faced pins can be set equal to the saved coordinates """
		startingWest = startingEast = startingSouth = startingNorth = None
		#If there is a west faced pin present, the other starting points should be derived from this pin
		if numberOfFacings[Facing.west] > 0:
			startingWest = self.coordinates		
		#If there is a south faced pin present (and no west), the other starting points should be derived from this pin
		elif numberOfFacings[Facing.south] > 0:
			startingSouth = self.coordinates
		#If there is a east faced pin present (and no west or south), the other starting points should be derived from this pin
		elif numberOfFacings[Facing.east] > 0:
			startingEast = self.coordinates
		#If there were no west or south or north	
		else:
			startingNorth = self.coordinates	
		return (startingWest, startingEast, startingSouth, startingNorth)

	def setInOutputCoordinateTuples(self, numberPins, sortedList, startingPoint, changeX, changeY):		
		""" Sets the tuples of the input and outputs right (a tuple consists of the coordinate and it's matching base) """
		if numberPins > 0:
			if self.mapping != []:
				count = 0
				for pin in sortedList:
					coordinate = None
					for i in self.mapping:
						if i[0] == pin.coordinates:
							coordinate = (i[1], i[2])
							break
					if coordinate == None:
						raise Exception("Coordinate not found: " + str(pin.coordinates))
					if isinstance(pin, AnnotatedInputBase):
						self.inputCoordinateTuples.append((coordinate, pin.node))
					else:	
						self.outputCoordinateTuples.append((coordinate, pin.node))
					count += 1

			else:
				count = 0
				for pin in sortedList:
					coordinate = (startingPoint[0] + changeX*count, startingPoint[1] + changeY*count)
					if isinstance(pin, AnnotatedInputBase):
						self.inputCoordinateTuples.append((coordinate, pin.node))
					else:	
						self.outputCoordinateTuples.append((coordinate, pin.node))
					count += 1

	def getNumberOfPinsPerFacing(self, sortedDict):			
		""" Returns for each facing how many pins are present """
		numberOfWest = numberOfSouth = numberOfEast = numberOfNorth = 0
		if "west" in sortedDict:
			numberOfWest = len(sortedDict["west"])
		if "east" in sortedDict:
			numberOfEast = len(sortedDict["east"])
		if "south" in sortedDict:
			numberOfSouth = len(sortedDict["south"])
		if "north" in sortedDict:
			numberOfNorth = len(sortedDict["north"])
				
		return (numberOfWest, numberOfSouth, numberOfEast, numberOfNorth)		

	def	getDifference(self, facing1, facing2):	
		""" Calculates the difference between facings and how much pixels they differ at on one side """
		differInNumber = int((facing1 - facing2) / 2)
		return differInNumber*10

	def calculateStartingPointSouth(self, numberOfFacings, normalPosition, firstIndex, secondIndex, operators, startingWest, startingSouth):			
		""" Calculates the south starting point """
		(firstOperator, secondOperator, thirdOperator) = operators
		if numberOfFacings[Facing.south] > 0 and startingSouth == None:
			if max(numberOfFacings[Facing.north], numberOfFacings[Facing.south]) < 3:
				firstCoordinate = firstOperator(startingWest[firstIndex], 20)
			elif numberOfFacings[Facing.south] >= numberOfFacings[Facing.north]:
				firstCoordinate = firstOperator(startingWest[firstIndex], numberOfFacings[Facing.south]*10)
			else: 
				firstCoordinate = firstOperator(startingWest[firstIndex], numberOfFacings[Facing.north]*10) 
				firstCoordinate = secondOperator(self.getDifference(numberOfFacings[Facing.north], numberOfFacings[Facing.south]))
			if numberOfFacings[Facing.west] >= numberOfFacings[Facing.east] or max(numberOfFacings[Facing.west], numberOfFacings[Facing.east]) < 3:
				secondCoordinate = startingWest[secondIndex]
			else:
				secondCoordinate = thirdOperator(startingWest[secondIndex], self.getDifference(numberOfFacings[Facing.east], numberOfFacings[Facing.west]))
			startingSouth = (firstCoordinate, secondCoordinate) if normalPosition else (secondCoordinate, firstCoordinate)
			
		return (startingSouth)		

	def calculateStartingPointEast(self, numberOfFacings, normalPosition, firstIndex, secondIndex, operators, startingWest, startingSouth, startingEast):		
		""" Calculates the east starting point """
		(firstOperator, secondOperator, thirdOperator, fourthOperator) = operators
		if numberOfFacings[Facing.east] > 0 and startingEast == None:
			if startingWest != None: 
				#We just need to add or subtract the max{number of south, number of north} to calculate the coordinate for the east starting point
				firstCoordinate = firstOperator(startingWest[firstIndex], (30 if max(numberOfFacings[Facing.north], numberOfFacings[Facing.south]) < 3 else (max(numberOfFacings[Facing.north], numberOfFacings[Facing.south]))*10 + 10))
				if numberOfFacings[Facing.east] == numberOfFacings[Facing.west] or max(numberOfFacings[Facing.west], numberOfFacings[Facing.east]) < 3:		
					secondCoordinate = startingWest[secondIndex]
				elif numberOfFacings[Facing.east] > numberOfFacings[Facing.west]:		
					secondCoordinate = secondOperator(startingWest[secondIndex], self.getDifference(numberOfFacings[Facing.east], numberOfFacings[Facing.west]))
				else:
					secondCoordinate = thirdOperator(startingWest[secondIndex], self.getDifference(numberOfFacings[Facing.west], numberOfFacings[Facing.east]))
				startingEast = (firstCoordinate, secondCoordinate) if normalPosition else (secondCoordinate, firstCoordinate)
			#We know that south has pins	
			else:	
				if numberOfFacings[Facing.south] >= numberOfFacings[Facing.north]:
					firstCoordinate = startingSouth[firstIndex] 
				else:
					firstCoordinate = fourthOperator(startingSouth[firstIndex], self.getDifference(numberOfFacings[Facing.north], numberOfFacings[Facing.south]))
				startingEast = (firstCoordinate, startingSouth[secondIndex]) if normalPosition else (startingSouth[secondIndex], firstCoordinate)
				
		return (startingEast)	

	def calculateStartingPointNorth(self, numberOfFacings, normalPosition, firstIndex, secondIndex, operators, startingWest, startingSouth, startingEast, startingNorth):	
		""" Calculates the north starting point """	
		(firstOperator, secondOperator, thirdOperator, fourthOperator, fifthOperator, sixthOperator, seventhOperator) = operators
		if numberOfFacings[Facing.north] > 0 and startingNorth == None:
			if startingEast != None:
				if max(numberOfFacings[Facing.north], numberOfFacings[Facing.south]) < 3 or numberOfFacings[Facing.north] >= numberOfFacings[Facing.south]:
					firstCoordinate = firstOperator(startingEast[firstIndex], 10)
				else: 
					firstCoordinate = firstOperator(startingSouth[firstIndex], self.getDifference(numberOfFacings[Facing.south], numberOfFacings[Facing.north]))
				if numberOfFacings[Facing.west] > numberOfFacings[Facing.east]:
					value = startingWest[secondIndex] 
				else:	
					value = startingEast[secondIndex]
				value = secondOperator(value, max(numberOfFacings[Facing.west], numberOfFacings[Facing.east])*10)
				secondCoordinate = secondOperator(value, 10 if max(numberOfFacings[Facing.west], numberOfFacings[Facing.east]) == 1 else 0)
			elif startingSouth != None:
				if numberOfFacings[Facing.north] == numberOfFacings[Facing.south] or max(numberOfFacings[Facing.north], numberOfFacings[Facing.south]) < 3:
					firstCoordinate = startingSouth[firstIndex] 
				elif numberOfFacings[Facing.north] > numberOfFacings[Facing.south]:
					firstCoordinate = thirdOperator(startingSouth[firstIndex], self.getDifference(numberOfFacings[Facing.north], numberOfFacings[Facing.south]))
				else:
					firstCoordinate = fourthOperator(startingSouth[firstIndex], self.getDifference(numberOfFacings[Facing.south], numberOfFacings[Facing.north]))
				secondCoordinate = fifthOperator(startingSouth[secondIndex], (30 if numberOfFacings[Facing.west] < 3 else numberOfFacings[Facing.west]*10 + 10))
			else:
				firstCoordinate = sixthOperator(startingWest[firstIndex], (20 if numberOfFacings[Facing.north] < 3 else numberOfFacings[Facing.north]*10))
				secondCoordinate = seventhOperator(startingWest[secondIndex], (20 if numberOfFacings[Facing.west] < 3 else numberOfFacings[Facing.west]*10))
				
			startingNorth = (firstCoordinate, secondCoordinate) if normalPosition else (secondCoordinate, firstCoordinate)
		return startingNorth

	def calculateStartingPointsEastFacing(self, numberOfFacings):			
		""" Calculates the starting points when the facing of the circuit is east """
		(startingWest, startingEast, startingSouth, startingNorth) = self.setCoordinatesToStartingPoint(numberOfFacings)
		#Calculate the other starting Points
		startingSouth = self.calculateStartingPointSouth(numberOfFacings, True, 0, 1, (sub, add, sub), startingWest, startingSouth)
		startingSouth = (startingSouth[0], startingSouth[1]-10) if startingSouth != None else startingSouth
		startingEast = self.calculateStartingPointEast(numberOfFacings, True, 0, 1, (sub, sub, add, sub), startingWest, startingSouth, startingEast)
		startingEast = (startingEast[0]-10, startingEast[1]+10) if startingEast != None and startingWest == None else startingEast
		startingNorth = self.calculateStartingPointNorth(numberOfFacings, True, 0, 1, (add, add, sub, add, add, sub, add), startingWest, startingSouth, startingEast, startingNorth)	
		return (startingWest, startingSouth, startingEast, startingNorth)	

	def calculateStartingPointsSouthFacing(self, numberOfFacings):			
		""" Calculates the starting points when the facing of the circuit is south """
		(startingWest, startingEast, startingSouth, startingNorth) = self.setCoordinatesToStartingPoint(numberOfFacings)
		startingSouth = self.calculateStartingPointSouth(numberOfFacings, False, 1, 0, (sub, add, add), startingWest, startingSouth)
		startingSouth = (startingSouth[0]+10, startingSouth[1]) if startingSouth != None else startingSouth
		startingEast = self.calculateStartingPointEast(numberOfFacings, False, 1, 0, (sub, add, sub, sub), startingWest, startingSouth, startingEast)
		startingEast = (startingEast[0]-10, startingEast[1]-10) if startingEast != None and startingWest == None else startingEast
		startingNorth = self.calculateStartingPointNorth(numberOfFacings, False, 1, 0, (add, sub, sub, add, sub, sub, sub), startingWest, startingSouth, startingEast, startingNorth)				
		return (startingWest, startingSouth, startingEast, startingNorth)	

	def calculateStartingPointsWestFacing(self, numberOfFacings):			
		""" Calculates the starting points when the facing of the circuit is west """
		(startingWest, startingEast, startingSouth, startingNorth) = self.setCoordinatesToStartingPoint(numberOfFacings)
		startingSouth = self.calculateStartingPointSouth(numberOfFacings, True, 0, 1, (add, sub, add), startingWest, startingSouth)
		startingSouth = (startingSouth[0], startingSouth[1]+10) if startingSouth != None else startingSouth
		startingEast = self.calculateStartingPointEast(numberOfFacings, True, 0, 1, (add, add, sub, add), startingWest, startingSouth, startingEast)
		startingEast = (startingEast[0]+10, startingEast[1]-10) if startingEast != None and startingWest == None else startingEast
		startingNorth = self.calculateStartingPointNorth(numberOfFacings, True, 0, 1, (sub, sub, add, sub, sub, add, sub), startingWest, startingSouth, startingEast, startingNorth)	
		return (startingWest, startingSouth, startingEast, startingNorth)		

	def calculateStartingPointsNorthFacing(self, numberOfFacings):	
		""" Calculates the starting points when the facing of the circuit is north """
		(startingWest, startingEast, startingSouth, startingNorth) = self.setCoordinatesToStartingPoint(numberOfFacings)
		startingSouth = self.calculateStartingPointSouth(numberOfFacings, False, 1, 0, (add, add, sub), startingWest, startingSouth)
		startingSouth = (startingSouth[0]-10, startingSouth[1]) if startingSouth != None else startingSouth
		startingEast = self.calculateStartingPointEast(numberOfFacings, False, 1, 0, (add, sub, add, add), startingWest, startingSouth, startingEast)
		startingEast = (startingEast[0]+10, startingEast[1]+10) if startingEast != None and startingWest == None else startingEast
		startingNorth = self.calculateStartingPointNorth(numberOfFacings, False, 1, 0, (sub, add, add, sub, add, add, add), startingWest, startingSouth, startingEast, startingNorth)			
		return (startingWest, startingSouth, startingEast, startingNorth)	

	def getInputCoordinates(self, list):	
		""" Overwrites the parent function (but has the same meaning) """
		for inputTuple in self.inputCoordinateTuples:
			list.append(inputTuple[0])

	def getOutputCoordinates(self, list):		
		""" Overwrites the parent function (but has the same meaning) """
		for outputTuple in self.outputCoordinateTuples:
			list.append(outputTuple[0])

	def bindOutput(self, annotatedConnectionList, annotatedComponentList, inputList, wiresToDelete, wiresToInsert, collide, faultList, circuitName=""):				
		""" Overwrites the parent function (but has the same meaning) """
		compOutputList = ReferencedOutput()	
		for outputTuple in self.outputCoordinateTuples:
			self.bindOutputWtihExtraInfoOfCoordinate(outputTuple[0], annotatedConnectionList, annotatedComponentList, compOutputList, inputList, wiresToDelete, wiresToInsert, outputTuple[1], collide, faultList, circuitName)	
		self.node.outputWire = compOutputList			

	def setInputRight(self, coordinate, annotatedConnection, inputsFound, extraInfo):				
		""" Overwrites the parent function (but has the same meaning) """				
		input = ReferencedInput(annotatedConnection.connection, extraInfo)
		inputsFound.append(input)	

	def bindWithInputWiresOrComp(self, annotatedComponentList, annotatedConnectionList):	
		""" Overwrites the parent function (but has the same meaning) """
		connectedInputs = []
		inputsFound = []
		self.bindCircuitWithComp(annotatedComponentList, annotatedConnectionList, inputsFound, connectedInputs)
		self.bindCircuitWithInputWires(annotatedConnectionList, inputsFound, connectedInputs)
		self.node.inputWires = inputsFound

	def bindCompWithComp(self, coordinate, inputComp, annotatedConnectionList, inputsFound, connectedInputs=[], extraInfo= None):			
		""" Overwrites the parent function (but has the same meaning) """			
		coordinatesToCheck = []
		for tupleCircuit in self.outputCoordinateTuples:
			coordinatesToCheck.append((self.node, tupleCircuit[0], tupleCircuit[1]))
		inputComp.bindWithComp(coordinate, self, annotatedConnectionList, inputsFound, coordinatesToCheck, connectedInputs, extraInfo)	

	def bindCircuitWithComp(self, annotatedComponentList, annotatedConnectionList, inputsFound, connectedInputs):		
		""" Binds the imported circuit with other nodes that lie immediately next to the imported circuit  """
		for annotatedList in annotatedComponentList:
			for annotatedComp in annotatedList:
				if annotatedComp == self:
					continue
				for tupleCircuit in self.inputCoordinateTuples:
					annotatedComp.bindCompWithComp(tupleCircuit[0], self, annotatedConnectionList, inputsFound, connectedInputs, tupleCircuit[1])

	def bindCircuitWithInputWires(self, annotatedConnectionList, inputsFound, connectedInputs):				
		""" Binds the imported circuit with its connected input wires  """	
		#Search for wires that are attached to the comp	
		for annotatedConnection in annotatedConnectionList:	
			for tupleCircuit in self.inputCoordinateTuples:
				self.bindWithInputWires(tupleCircuit[0], annotatedConnection, annotatedConnectionList, inputsFound, connectedInputs, tupleCircuit[1])	
	
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		
