#!/usr/bin/env python3

from translator.intermediateBuilder.annotated.annotatedNode import AnnotatedNode
from translator.intermediateBuilder.numberedInput import NumberedInput
from translator.intermediateBuilder.input import Input
from translator.intermediateBuilder.output import Output

class AnnotatedPriorityEncoder(AnnotatedNode):
	""" AnnotatedPriorityEncoder is the annotated version of a priority encoder """
	
	def __init__(self, priority, coordinates, facing):
		""" Constructor """
		AnnotatedNode.__init__(self, priority, coordinates, facing)	
		self.inputCoordinates = []
		self.enableInputCoordinate = None
		self.enableOutputCoordinate = None
		#If the facing is east or west, only the y coordinate changes southwards, otherwise the x-coordinate changes eastwards
		self.groupOutputCoordinate = ((facing == "east" or facing  == "west") and (coordinates[0], coordinates[1]+10)) or (coordinates[0]+10, coordinates[1])

	def findExtraCoordinates(self):
		""" Overwrites the parent function (but has the same meaning) """
		size = 40		
		(stableXCoordinate, stableYCoordinate) = self.coordinates
		#The coordinates that stay the same, no matter what input or output number we are at
		stableCoordinateOfVarying = ((self.facing == "east") and (stableXCoordinate-size)) or ((self.facing == "west") and (stableXCoordinate+size)) or ((self.facing == "north") and (stableYCoordinate+size)) or ((self.facing == "south") and (stableYCoordinate-size))				
		#The stable coordinate of the enable in and output
		stableCoordinateOfEnable = ((self.facing == "east") and (stableXCoordinate-size/2)) or ((self.facing == "west") and (stableXCoordinate+size/2)) or ((self.facing == "north") and (stableYCoordinate+size/2)) or ((self.facing == "south") and (stableYCoordinate-size/2))
		#If the stableCoordinate = 0, than x is the stable coordinate, otherwise y is the stable coordinate
		stableCoordinateNumber = ((self.facing == "east" or self.facing == "west") and 0) or ((self.facing == "south" or self.facing == "north") and 1)
					
		#The middle point of the varying input coordinates
		middlePoint = (stableCoordinateNumber and stableXCoordinate) or (not stableCoordinateNumber and stableYCoordinate)
		#The list with all the varying coordinates (inputs or outputs)
		varyingList = self.calculateInputCoordinates(middlePoint)	
				
		#We also have a enable output (which has 10 pixels distance to the lowest index number)		
		self.enableOutputCoordinate = (stableCoordinateNumber == 0 and (stableCoordinateOfEnable, varyingList[0][0]-10)) or (varyingList[0][0]-10, stableCoordinateOfEnable)		
		#We also have a enable input (which has 10 pixels distance to the highest index number)		
		self.enableInputCoordinate = (stableCoordinateNumber == 0 and (stableCoordinateOfEnable, varyingList[-1][0]+10)) or (varyingList[-1][0]+10, stableCoordinateOfEnable)		
			
		for varCoordinate in varyingList:
			if stableCoordinateNumber == 0:
				#The tuple contains the coordinates and the number of the coordinate
				tuple = ((stableCoordinateOfVarying, varCoordinate[0]), varCoordinate[1])
			else:
				tuple = ((varCoordinate[0], stableCoordinateOfVarying), varCoordinate[1])

			self.inputCoordinates.append(tuple)	

	def calculateInputCoordinates(self, middlePoint):	
		""" Calculate the input coordinates (the coordinates that can change in number as the width of the select bit changes) 	""" 
		coordinateList = []
		nrOfInputs = 2**self.node.bits
		startingPoint = middlePoint - 10*(nrOfInputs/2)+10
		for nr in range(nrOfInputs):
			coordinateList.append((startingPoint+10*nr, nr))
		return coordinateList

	def getInputCoordinates(self, list):				
		""" Overwrites the parent function (but has the same meaning) """
		list.extend(self.inputCoordinates)
		list.append(self.enableInputCoordinate)

	def getOutputCoordinates(self, list):		
		""" Overwrites the parent function (but has the same meaning) """
		AnnotatedNode.getOutputCoordinates(self, list)	
		list.append(self.enableOutputCoordinate)		
		list.append(self.groupOutputCoordinate)		

	def bindWithInputWiresOrComp(self, annotatedComponentList, annotatedConnectionList):			
		""" Overwrites the parent function (but has the same meaning) """
		connectedInputs = []
		inputsFound = []
		self.bindPriorityEncoderWithComp(annotatedComponentList, annotatedConnectionList, inputsFound, connectedInputs)
		self.bindPriorityEncoderWithInputWires(annotatedConnectionList, inputsFound, connectedInputs)
		self.node.inputWires = inputsFound		

	def setInputRight(self, coordinate, annotatedConnection, inputsFound, extraInfo):			
		""" Overwrites the parent function (but has the same meaning) """					
		if coordinate == self.enableInputCoordinate:
			input = Input(annotatedConnection.connection)
			self.node.enableInputWire = input			
		else:
			#Because numbers are important in the priority encoder, we need to make a NumberedInput
			input = NumberedInput(annotatedConnection.connection, extraInfo)
			inputsFound.append(input)			
	
	def bindPriorityEncoderWithComp(self, annotatedComponentList, annotatedConnectionList, inputsFound, connectedInputs):		
		""" Binds the priority encoder with other nodes that lie immediately next to the priority encoder  """
		for annotatedList in annotatedComponentList:
			for annotatedComp in annotatedList:
				if annotatedComp == self:
					continue
				for coordinate in self.inputCoordinates:
					annotatedComp.bindCompWithComp(coordinate[0], self, annotatedConnectionList, inputsFound, connectedInputs, coordinate[1])

				annotatedComp.bindCompWithComp(self.enableInputCoordinate, self, annotatedConnectionList, inputsFound, connectedInputs)		

	def bindPriorityEncoderWithInputWires(self, annotatedConnectionList, inputsFound, connectedInputs):				
		""" Binds the priority encoder with its connected input wires  """		
		#Search for wires that are attached to the comp	
		for annotatedConnection in annotatedConnectionList:	
			for coordinate in self.inputCoordinates:
				self.bindWithInputWires(coordinate[0], annotatedConnection, annotatedConnectionList, inputsFound, connectedInputs, coordinate[1])	
			self.bindWithInputWires(self.enableInputCoordinate, annotatedConnection, annotatedConnectionList, inputsFound, connectedInputs)								

	def bindOutput(self, annotatedConnectionList, annotatedComponentList, inputList, wiresToDelete, wiresToInsert, collide, faultList, circuitName=""):				
		""" Overwrites the parent function (but has the same meaning) """	
		#Binds the normal output (coordinates)
		compOutputList = Output()	
		self.bindOutputOfCoordinate(self.coordinates, annotatedConnectionList, annotatedComponentList, compOutputList, inputList, wiresToDelete, wiresToInsert, collide, faultList, circuitName)
		self.node.outputWire = compOutputList	
		
		#Binds the enable output (the greater output)
		compOutputList = Output()	
		self.bindOutputOfCoordinate(self.enableOutputCoordinate, annotatedConnectionList, annotatedComponentList, compOutputList, inputList, wiresToDelete, wiresToInsert, collide, faultList, circuitName)
		self.node.enableOutputWire = compOutputList		
		
		#Binds the group output (the less output)
		compOutputList = Output()	
		self.bindOutputOfCoordinate(self.groupOutputCoordinate, annotatedConnectionList, annotatedComponentList, compOutputList, inputList, wiresToDelete, wiresToInsert, collide, faultList, circuitName)
		self.node.groupOutputWire = compOutputList				

	def bindCompWithComp(self, coordinate, inputComp, annotatedConnectionList, inputsFound, connectedInputs=[], extraInfo= None):				
		""" Overwrites the parent function (but has the same meaning) """		
		coordinatesToCheck = []
		#We need to remember to which output we need to assign the comp to (if it turns out to be directly connected)
		coordinatesToCheck.append((self.node, self.coordinates, "normal"))
		coordinatesToCheck.append((self.node, self.enableOutputCoordinate, "enable"))
		coordinatesToCheck.append((self.node, self.groupOutputCoordinate, "group"))
		inputComp.bindWithComp(coordinate, self, annotatedConnectionList, inputsFound, coordinatesToCheck, connectedInputs, extraInfo)	

		
			
			
			
			
			
			