#!/usr/bin/env python3

from math import ceil
from translator.intermediateBuilder.annotated.annotatedElement import AnnotatedElement
from translator.intermediateBuilder.output import Output
from translator.intermediateBuilder.input import Input
from translator.intermediateBuilder.wire import Wire
from translator.intermediateBuilder.annotated.annotatedConnection import AnnotatedConnection	
from translator.intermediateBuilder.helperFunctions import sub
from translator.intermediateBuilder.helperFunctions import add
from translator.intermediateBuilder.helperFunctions import returnA

class AnnotatedSplitter(AnnotatedElement):
	""" AnnotatedSplitter is the annotated version of the splitter """
	
	def __init__(self, splitter, coordinatesFrom, facing, fanOut, extraInfo, appearance):
		""" Constructor """
		self.splitter = splitter
		self.coordinatesFrom = coordinatesFrom
		self.coordinatesTo = [] 
		self.facing = facing 
		self.fanOut = fanOut 
		self.appearance = appearance
		self.setExtraInfo(extraInfo)

	def setExtraInfo(self, extraInfo):		
		""" Set the information needed to split the incoming bits properly 	""" 	
		coordinates = self.getExtraCoordinates()
		#First we need to initialize all the coordinatesTo to None
		for i in range(self.fanOut):
			self.coordinatesTo.append([coordinates[i], None])
		#We need to check whether there is something special for this bits
		for i in range(self.splitter.bits):
			#If we can find some information for this bit, we know that we need to handle this bit in a special way
			if "bit" + str(i) in extraInfo:
				coordinateNr = extraInfo["bit"+str(i)]
				#If this coordinate doesn't already gets a signal
				if coordinateNr != "none" and self.coordinatesTo[int(coordinateNr)][1] == None:
					self.coordinatesTo[int(coordinateNr)][1] = [i]
				elif coordinateNr != "none" and self.coordinatesTo[int(coordinateNr)][1] != None:
					self.coordinatesTo[int(coordinateNr)][1].append(i)
			else:
				if self.coordinatesTo[i][1] == None:
					self.coordinatesTo[i][1] = [i]
				else:
					self.coordinatesTo[i][1].append(i)


	def getExtraCoordinates(self):		
		""" Gets the To coordinates of the splitter """ 
		coordinates = []
		(xFromCoordinate, yFromCoordinate) = self.coordinatesFrom
		
		#The coordinates that stay the same, no matter what input number we are at
		stableToCoordinate = ((self.facing == "east") and (xFromCoordinate+20)) or ((self.facing == "west") and (xFromCoordinate-20)) or ((self.facing == "north") and (yFromCoordinate-20)) or ((self.facing == "south") and (yFromCoordinate+20))
		#If the stableCoordinate = 0, than x is the stable coordinate, otherwise y is the stable coordinate
		stableCoordinate = ((self.facing == "east" or self.facing == "west") and 0) or ((self.facing == "south" or self.facing == "north") and 1)
		
		#The middle point of the varying input coordinates
		middlePoint = (stableCoordinate and xFromCoordinate) or (not stableCoordinate and yFromCoordinate)

		# If the number of fanOut's is odd, then there would be as many wires to the left as to the right
		# For the new left/right-handed splitter, this would mean that the formula is OK
		# If the number of fanOut's is even, then there would be a wire more on one side
		# For the new left/right-handed splitter, this would mean that the middlePoint has an extra offset of 10, depending on the facing
		oddNumber = self.fanOut % 2
		if self.appearance == "center":
			pass
		elif self.appearance == "legacy":
			pass
		elif self.appearance == "left":
			if self.facing == "east":
				middlePoint = yFromCoordinate - 10*((self.fanOut - 1)//2+1)
			elif self.facing == "west":
				middlePoint = yFromCoordinate + 10*((self.fanOut - 1)//2+1)
				if not oddNumber:
					middlePoint += 10
			elif self.facing == "north":
				middlePoint = xFromCoordinate - 10*((self.fanOut - 1)//2+1)
			elif self.facing == "south":
				middlePoint = xFromCoordinate + 10*((self.fanOut - 1)//2+1)
				if not oddNumber:
					middlePoint += 10
		elif self.appearance == "right":
			if self.facing == "east":
				middlePoint = yFromCoordinate + 10*((self.fanOut - 1)//2+1)
				if not oddNumber:
					middlePoint += 10
			elif self.facing == "west":
				middlePoint = yFromCoordinate - 10*((self.fanOut - 1)//2+1)
			elif self.facing == "north":
				middlePoint = xFromCoordinate + 10*((self.fanOut - 1)//2+1)
				if not oddNumber:
					middlePoint += 10
			elif self.facing == "south":
				middlePoint = xFromCoordinate - 10*((self.fanOut - 1)//2+1)

		#The list with all the varying coordinates
		return self.calculateExtraCoordinates(middlePoint, stableToCoordinate, stableCoordinate==0)	

	def calculateExtraCoordinates(self, middlePoint, otherCoordinate, EastOrWest):			
		""" Helps to calculate the To coordinates of the splitter (if we have east or west facing, the higher bits lie on a lower coordinate) """ 
		toVarList = [(otherCoordinate, middlePoint)] if EastOrWest else [(middlePoint, otherCoordinate)]
		nrOfInputs = self.fanOut-1
		#We push the coordinates onto the list according to their position
		for i in range(1, nrOfInputs//2+1):
			if EastOrWest:
				toVarList.insert(0, (otherCoordinate, middlePoint-i*10))
				toVarList.append((otherCoordinate, middlePoint+i*10))
			else:
				toVarList.insert(0, (middlePoint+i*10, otherCoordinate))
				toVarList.append((middlePoint-i*10, otherCoordinate))					
		odd = nrOfInputs%2				
		#If the number of fanOuts is even (so after the -1 odd), we need to calculate one extra coordinate 
		if odd != 0:
			if EastOrWest:
				toVarList.insert(0, (otherCoordinate, middlePoint-(self.fanOut//2)*10))
			else:
				toVarList.append((middlePoint-(self.fanOut//2)*10, otherCoordinate))

		return toVarList					

	def getWireCoordinates(self):	
		""" Gives the coordinates of a wire back """ 
		if self.facing == "east":
			return self.getWireCoordinatesHelpFunction(add, returnA, sub, returnA)
		elif self.facing == "west":
			return self.getWireCoordinatesHelpFunction(sub, returnA, add, returnA)
		elif self.facing == "north":
			return self.getWireCoordinatesHelpFunction(returnA, sub, returnA, add)
		else:
			return self.getWireCoordinatesHelpFunction(returnA, add, returnA, sub)

	def getWireCoordinatesHelpFunction(self, operation1, operation2, operation3, operation4):			
		""" Helps to give the coordinates of a wire back and adjusts the coordinates of the splitter  """ 	
		coordinateList = []
		#The wire for the starting point of the splitter
		coordinateList.append((self.coordinatesFrom, (operation1(self.coordinatesFrom[0], 5), operation2(self.coordinatesFrom[1], 5))))
		self.coordinatesFrom = (operation1(self.coordinatesFrom[0], 5), operation2(self.coordinatesFrom[1], 5))
		#The wires for the ending points of the splitter
		for wire in self.coordinatesTo:
			coordinateList.append(((operation3(wire[0][0], 5), operation4(wire[0][1], 5)), (wire[0][0], wire[0][1])))
			wire[0] = (operation3(wire[0][0], 5), operation4(wire[0][1], 5))
		return coordinateList

	def bindOutputOfSplitter(self, annotatedComponentList, annotatedConnectionList, inputList, wiresToDelete, wiresToInsert, collide, faultList, circuitName):	
		""" Binds the coordinates as if they were all sending a signal (so that they were all output coordinates) """ 
		compOutputList = Output()	
		self.bindOutputOfCoordinate(self.coordinatesFrom, annotatedConnectionList, annotatedComponentList, compOutputList, inputList, wiresToDelete, wiresToInsert, collide, faultList, circuitName, self.splitter)
		#We don't know any inputs yet, so we push none for the inputs onto the start and for the outputs, they are equal to compOutputList
		self.splitter.start = [None, compOutputList]
	
		outputList = []
		for coordinate in self.coordinatesTo:
			compOutputList = Output()	
			self.bindOutputOfCoordinate(coordinate[0], annotatedConnectionList, annotatedComponentList, compOutputList, inputList, wiresToDelete, wiresToInsert, collide, faultList, circuitName, self.splitter)
			outputList.append((coordinate[1], compOutputList))
		self.splitter.end = [[], outputList]

	def bindWithInputWires(self, annotatedConnectionList):		
		""" Binds the coordinates of the splitter with input wires """ 
		connectedInputs = []
		#Search for wires that are attached to the splitter	
		for annotatedConnection in annotatedConnectionList:	
			foundWire = self.bindWithInputWire(self.coordinatesFrom, annotatedConnection, annotatedConnectionList, connectedInputs)
			self.splitter.start[0] = foundWire if foundWire != None else self.splitter.start[0]
			for coordinate in self.coordinatesTo:
				foundWire = self.bindWithInputWire(coordinate[0], annotatedConnection, annotatedConnectionList, connectedInputs)	
				if foundWire != None:
					self.splitter.end[0].append((coordinate[1], foundWire))

	def bindWithInputWire(self, coordinate, annotatedConnection, annotatedConnectionList, connectedInputs):		
		""" Binds the given coordinate of the splitter with the input wire 	""" 
		#If the input coordinates of the comp are equal to the output coordinates of a wire
		if annotatedConnection.coordinatesTo == coordinate:
			#If there is already an input connected to this wire, we need to make a new wire
			if annotatedConnection.connection.end != None:
				wire = Wire(annotatedConnection.connection.direction)
				annotatedWire = AnnotatedConnection(wire, annotatedConnection.coordinatesFrom, None)
				annotatedConnectionList.append(annotatedWire)
				connectedInputs.append(coordinate)
				annotatedWire.connection.start = annotatedConnection.connection.start
				annotatedWire.connection.end = self.splitter	
				#If this output already had a wire connected to it, we need to add the new wire in its output list
				if annotatedWire.connection.start != None:
					annotatedWire.connection.start.setExtraOutput(annotatedConnection.connection, annotatedWire.connection)
				return Input(annotatedWire.connection)					
			else:	
				connectedInputs.append(coordinate)
				annotatedConnection.connection.end = self.splitter	
				return Input(annotatedConnection.connection)
			#If there is a wire, that isn't connected to another output gate or input component or ... (something that can give it a signal)
			#then it is not sure that the coordinates (from and to) of the wire are set right
			if annotatedConnection.coordinatesFrom == coordinate and coordinate not in connectedInputs:
				#If there is already an input connected to this wire, we need to make a new wire
				if annotatedConnection.connection.end != None:
					wire = Wire(annotatedConnection.connection.direction)
					annotatedWire = AnnotatedConnection(wire, annotatedConnection.coordinatesTo, annotatedConnection.coordinatesFrom)
					annotatedConnectionList.append(annotatedWire)
					annotatedWire.connection.end = self.splitter	
					return Input(annotatedWire.connection)
				#Else we connect this wire with the splitter	
				else:	
					annotatedConnection.connection.end = self.splitter	
					temp = annotatedConnection.coordinatesFrom
					annotatedConnection.coordinatesFrom = annotatedConnection.coordinatesTo
					annotatedConnection.coordinatesTo = temp		
					return Input(annotatedConnection.connection)
