#!/usr/bin/env python3
# -*- coding: iso-8859-15 -*-
from xml.dom import minidom            
from operator import attrgetter
from itertools import groupby
from copy import deepcopy

from translator.intermediateBuilder.wire import Wire
from translator.intermediateBuilder.tunnel import Tunnel
from translator.intermediateBuilder.transistor import Transistor
from translator.intermediateBuilder.andGate import AndGate
from translator.intermediateBuilder.nandGate import NandGate
from translator.intermediateBuilder.orGate import OrGate
from translator.intermediateBuilder.norGate import NorGate
from translator.intermediateBuilder.notGate import NotGate
from translator.intermediateBuilder.bufferGate import BufferGate
from translator.intermediateBuilder.xorGate import XorGate
from translator.intermediateBuilder.xnorGate import XnorGate
from translator.intermediateBuilder.evenParityGate import EvenParityGate
from translator.intermediateBuilder.oddParityGate import OddParityGate
from translator.intermediateBuilder.controlledBufferGate import ControlledBufferGate
from translator.intermediateBuilder.controlledInverterGate import ControlledInverterGate
from translator.intermediateBuilder.standardInputBase import StandardInputBase
from translator.intermediateBuilder.standardOutputBase import StandardOutputBase
from translator.intermediateBuilder.probeBase import ProbeBase
from translator.intermediateBuilder.clockBase import ClockBase
from translator.intermediateBuilder.constantBase import ConstantBase
from translator.intermediateBuilder.buttonBase import ButtonBase
from translator.intermediateBuilder.ledBase import LedBase
from translator.intermediateBuilder.joystickBase import JoystickBase
from translator.intermediateBuilder.segment7Display import Segment7Display
from translator.intermediateBuilder.hexDigitDisplay import HexDigitDisplay
from translator.intermediateBuilder.ledMatrix import LedMatrix
from translator.intermediateBuilder.demultiplexer import Demultiplexer
from translator.intermediateBuilder.multiplexer import Multiplexer
from translator.intermediateBuilder.decoder import Decoder
from translator.intermediateBuilder.bitSelector import BitSelector
from translator.intermediateBuilder.priorityEncoderNode import PriorityEncoderNode
from translator.intermediateBuilder.bitExtenderNode import BitExtenderNode
from translator.intermediateBuilder.multiplier import Multiplier
from translator.intermediateBuilder.divider import Divider
from translator.intermediateBuilder.adder import Adder
from translator.intermediateBuilder.subtractor import Subtractor
from translator.intermediateBuilder.bitAdder import BitAdder
from translator.intermediateBuilder.bitFinder import BitFinder
from translator.intermediateBuilder.comparator import Comparator
from translator.intermediateBuilder.shifter import Shifter
from translator.intermediateBuilder.negator import Negator
from translator.intermediateBuilder.dFlipFlop import DFlipFlop
from translator.intermediateBuilder.tFlipFlop import TFlipFlop
from translator.intermediateBuilder.srFlipFlop import SRFlipFlop
from translator.intermediateBuilder.jkFlipFlop import JKFlipFlop
from translator.intermediateBuilder.random import Random
from translator.intermediateBuilder.counter import Counter
from translator.intermediateBuilder.register import Register
from translator.intermediateBuilder.standardShiftRegister import StandardShiftRegister
from translator.intermediateBuilder.parallelLoadShiftRegister import ParallelLoadShiftRegister
from translator.intermediateBuilder.ram import Ram
from translator.intermediateBuilder.rom import Rom
from translator.intermediateBuilder.importedCircuitNode import ImportedCircuitNode
from translator.intermediateBuilder.splitter import Splitter
from translator.intermediateBuilder.annotated.annotatedStandardConnection import AnnotatedStandardConnection
from translator.intermediateBuilder.annotated.annotatedTunnel import AnnotatedTunnel
from translator.intermediateBuilder.annotatedElementFactory import AnnotatedElementFactory
from translator.intermediateBuilder.parseError import ParseError
from translator.intermediateBuilder.circuit import Circuit
from translator.intermediateBuilder.intermediateTree import IntermediateTree
from translator.intermediateBuilder.helperFunctions import deleteElementsFromElementList
from translator.intermediateBuilder.helperFunctions import getInputList
from translator.intermediateBuilder.helperFunctions import getOutputList

#Sizes of the components
narrow = 30
medium = 50
wide = 70
version = 0

def getCoordinationTuple(string):
	""" Returns a tuple of coordinates from a given string """
	comma = string.find(',')	
	length = len(string)
	x = int(string[1:comma])
	y = int(string[comma+1:length-1])
	coordinates = x, y
	return coordinates	

class IntermediateBuilder():
	""" If we give an xml logisim file to this class, it will make an intermediate tree (makeIntermediateTree) """
	
	def __init__(self, filename, collide= False):
		""" Constructor """
		self.xmldoc = minidom.parse(filename)
		#We keep a list of serious errors
		self.faultList = []
		#The bool specifies if you want to allow outputs wires to collide 
		self.collide = collide
		self.annotatedFactory = AnnotatedElementFactory()
		self.remapping = {}

	def makeIntermediateTree(self):		
		""" Transforms logisim xml file into an intermediate abstract tree """
		xmlVersionList = self.xmldoc.getElementsByTagName("project")
		global version
		version = xmlVersionList[0].attributes["source"].value
		supportedVersions = ["2.6.0", "2.6.1", "2.7.0", "2.7.1", "2.7.1 - MOD"]
		if version not in supportedVersions:
			raise ParseError('Unsupported logisim save version')
		xmlMainCircuitList = self.xmldoc.getElementsByTagName("main")
		mainCircuit = xmlMainCircuitList[0].attributes["name"].value

		xmlCircuitList = self.xmldoc.getElementsByTagName("circuit")
		#When there is no circuit, we have nothing to translate
		if len(xmlCircuitList) == 0:
			raise ParseError('Nothing in the circuitList')
		else:
			#In the tuple list we store the intermediateCircuit object and the xmlCircuit object and the list with all the in- and outputs
			#(we need this list of bases for when we resolve included circuits)
			circuitTupleList = []
			
			for xmlCircuit in xmlCircuitList:
				if xmlCircuit.attributes["name"].value == mainCircuit:
					main = True
				else:
					main = False
				circuit = Circuit((xmlCircuit.attributes["name"].value.replace(" ", "")).replace("/",""), main)
				annotatedPinlist = self.getAnnotatedPinList(xmlCircuit)
				circuitTupleList.append((circuit, xmlCircuit, annotatedPinlist))
				self.getMapping(xmlCircuit)
			circuitList = []
			for circuitTuple in circuitTupleList:
				circuit = circuitTuple[0]
				xmlCircuit = circuitTuple[1]
				
				annotatedComponentList = []
		
				annotatedTunnelList = self.getAnnotatedTunnelList(xmlCircuit)	
				annotatedResistorList = self.getAnnotatedResistorList(xmlCircuit)
				annotatedConnectionList = self.getAnnotatedWireList(xmlCircuit)
				annotatedSplitterList = self.getAnnotatedSplitterList(xmlCircuit)
				annotatedGateList = self.getAnnotatedGateList(xmlCircuit)
				annotatedBaseList = self.getAnnotatedBaseList(xmlCircuit)
				annotatedPlexerList = self.getAnnotatedPlexerList(xmlCircuit)
				annotatedArithmeticList = self.getAnnotatedArithmeticList(xmlCircuit)
				annotatedInOutList = self.getAnnotatedInOutList(xmlCircuit) 
				annotatedPriorityList = self.getPriorityEncoderList(xmlCircuit)
				annotatedExtenderList = self.getBitExtenderList(xmlCircuit)
				annotatedMemoryList = self.getMemoryList(xmlCircuit)
				annotatedImportedCircuitList = self.getImportedList(xmlCircuit, circuitTupleList) 

				annotatedComponentList.append(annotatedGateList)
				annotatedComponentList.append(circuitTuple[2])
				annotatedComponentList.append(annotatedBaseList)
				annotatedComponentList.append(annotatedPlexerList)
				annotatedComponentList.append(annotatedArithmeticList)
				annotatedComponentList.append(annotatedPriorityList)
				annotatedComponentList.append(annotatedExtenderList)
				annotatedComponentList.append(annotatedInOutList)
				annotatedComponentList.append(annotatedMemoryList)
				annotatedComponentList.append(annotatedImportedCircuitList)
				
				for annotatedList in annotatedComponentList:
					for annotatedComp in annotatedList:
						annotatedComp.findExtraCoordinates()
				
				self.addTunnelsToConnectionList(annotatedTunnelList, annotatedComponentList, annotatedConnectionList)
				annotatedConnectionList.extend(annotatedResistorList)
				#We make for each coordinate of the splitter a wire and add this wireList to the connectionList
				annotatedConnectionList.extend(self.makeWiresForSplitters(annotatedSplitterList))
				
				#Binds the components with their connections
				self.bindCompAndSplitterWithWires(annotatedComponentList, annotatedConnectionList, annotatedSplitterList, circuit.name)

				for annotatedConnection in annotatedConnectionList:
					circuit.connectionList.append(annotatedConnection.connection)
				for annotatedCompList in annotatedComponentList:
					for annotatedComp in annotatedCompList:
						circuit.nodeList.append(annotatedComp.node)					
				for annotatedSplitter in annotatedSplitterList:
					circuit.nodeList.append(annotatedSplitter.splitter)

				circuitList.append(circuit)		
					
			return [IntermediateTree(circuitList), self.faultList]

	def getAnnotatedTunnelList(self, xmlCircuit):			
		""" Returns a list with al the tunnels, grouped by their label """
		annotatedTunnelList = []		
		xmlCompList = xmlCircuit.getElementsByTagName("comp")
		for xmlComp in xmlCompList:
			if xmlComp.attributes["name"].value == "Tunnel":
				name = ''
				bits = 1
				descriptions = xmlComp.getElementsByTagName("a")
				for description in descriptions:
					if description.attributes["name"].value == "label": 
						name = description.attributes["val"].value
					elif description.attributes["name"].value == "width":
						bits = int(description.attributes["val"].value)
				
				tunnel = Tunnel(name, bits)
				annotatedTunnel = AnnotatedTunnel(getCoordinationTuple(xmlComp.attributes["loc"].value), tunnel)
				annotatedTunnelList.append(annotatedTunnel)
				#We remove the xmlComp from the xmlCircuit, because we analysed these already, so we don't need them anymore
				xmlCircuit.removeChild(xmlComp)	
				
		#We sort the tunnels (according to their labels)
		annotatedTunnelList = sorted(annotatedTunnelList, key = attrgetter('connection.name'))
		groupedList = []
		#We group the tunnels according to their label ((label1, (tunnel1 with label1, tunnel2 with label2, ...)), (label2, (tunel1 with label2), ...))
		for key, group in groupby(annotatedTunnelList, key=attrgetter('connection.name')):
			samegroup = []
			for tunnel in group:
				samegroup.append(tunnel)
			groupedList.append((key, samegroup))
			
		return groupedList	

	def getAnnotatedWireList(self, circuit):			
		""" Returns a list with all the wires and its coordinates """
		annotatedWireList = []		
		xmlWireList = circuit.getElementsByTagName("wire")
		for xmlWire in xmlWireList:
			wire = Wire()
			coordinatesFrom = getCoordinationTuple(xmlWire.attributes["from"].value)
			coordinatesTo = getCoordinationTuple(xmlWire.attributes["to"].value)
			annotatedWire = AnnotatedStandardConnection(wire, coordinatesFrom, coordinatesTo)
			annotatedWireList.append(annotatedWire)
			circuit.removeChild(xmlWire)
		return annotatedWireList

	def getAnnotatedResistorList(self, circuit):
		""" Returns a list with all the resistors and its coordinates """
		annotatedResistorList = []		
		xmlWireList = circuit.getElementsByTagName("comp")
		for xmlWire in xmlWireList:
			if xmlWire.attributes["name"].value == "Pull Resistor":
				wire = Wire("0")
			else:
				continue	
			#Check whethet the comp has extra special properties	
			descriptions = xmlWire.getElementsByTagName("a")	
			for description in descriptions:
				if description.attributes["name"].value == "pull":
					wire.direction = description.attributes["val"].value
		
			coordinatesFrom = coordinatesTo = getCoordinationTuple(xmlWire.attributes["loc"].value)
			annotatedWire = AnnotatedStandardConnection(wire, coordinatesFrom, coordinatesTo)
			annotatedResistorList.append(annotatedWire)
			circuit.removeChild(xmlWire)
		return annotatedResistorList

	def getAnnotatedSplitterList(self, circuit):		
		""" Returns a list with all the splitters and its special information """
		annotatedSplitterList = []		
		xmlCompList = circuit.getElementsByTagName("comp")
		for xmlComp in xmlCompList:
			if xmlComp.attributes["name"].value == "Splitter":
				splitter = Splitter()
			else:
				continue	
			#Default	
			facing = "east"	
			fanOut = 2
			extraInfo =  dict()

			if version == "2.6.0" or version == "2.6.1":
				appearance = "legacy"
			elif version == "2.7.0" or version == "2.7.1":
				appearance = "left"

			descriptions = xmlComp.getElementsByTagName("a")	
			for description in descriptions:
				if description.attributes["name"].value == "facing":
					facing = description.attributes["val"].value
				elif description.attributes["name"].value == "fanout":
					fanOut = int(description.attributes["val"].value)
				elif description.attributes["name"].value == "appear":
					appearance = description.attributes["val"].value
				elif description.attributes["name"].value == "incoming":
					splitter.bits = int(description.attributes["val"].value)
				else:
					extraInfo[description.attributes["name"].value] = description.attributes["val"].value
						
			coordinates = getCoordinationTuple(xmlComp.attributes["loc"].value)
			#The factory will return the annotatedSplitter
			annotatedSplitter = self.annotatedFactory.giveAnnotatedSplitter(splitter, coordinates, facing, fanOut, extraInfo, appearance)
			annotatedSplitterList.append(annotatedSplitter)	
			circuit.removeChild(xmlComp)				
		return annotatedSplitterList		

	def getAnnotatedGateList(self, circuit):		
		""" Returns a list with all the gates and its important attributes """
		annotatedGateList = []				
		xmlCompList = circuit.getElementsByTagName("comp")
		for xmlComp in xmlCompList:
			#Default
			nrOfInputs = 5
			size = medium
			facing = "east"
			out = "01"
			controlSide = None
			transistorType = None
			if xmlComp.attributes["name"].value == "AND Gate":
				gate = AndGate()
			elif xmlComp.attributes["name"].value == "OR Gate":
				gate = OrGate()
			elif xmlComp.attributes["name"].value == "NAND Gate":
				gate = NandGate()
			elif xmlComp.attributes["name"].value == "NOR Gate":
				gate = NorGate()	
			elif xmlComp.attributes["name"].value == "Odd Parity":
				gate = OddParityGate()
			elif xmlComp.attributes["name"].value == "Even Parity":
				gate = EvenParityGate()	
			elif xmlComp.attributes["name"].value == "XOR Gate":
				gate = XorGate()	
			elif xmlComp.attributes["name"].value == "XNOR Gate":
				gate = XnorGate()	
			elif xmlComp.attributes["name"].value == "NOT Gate":
				gate = NotGate()
				nrOfInputs = 1
				size = narrow
			elif xmlComp.attributes["name"].value == "Controlled Inverter":
				gate = ControlledInverterGate()
				nrOfInputs = 1
				size = narrow	
				controlSide = "right"
			elif xmlComp.attributes["name"].value == "Transistor":
				gate = Transistor()
				nrOfInputs = 1
				size = 40
				controlSide = "tl"
				gate.transistorType = "p"
				transistorType = "p"
			elif xmlComp.attributes["name"].value == "Transmission Gate":
				gate = Transistor()
				nrOfInputs = 1
				size = 40
				controlSide = "tl"
				gate.transistorType = "c"
				transistorType = "c"
			elif xmlComp.attributes["name"].value == "Buffer":
				gate = BufferGate()
				nrOfInputs = 1
				size = 20			
			elif xmlComp.attributes["name"].value == "Controlled Buffer":
				gate = ControlledBufferGate()
				nrOfInputs = 1
				size = 20	
				controlSide = "right"
			#If it wasn't a gate, we have to use another function	
			else:
				continue
	
			descriptions = xmlComp.getElementsByTagName("a")
			negatedInputIndexes = []
			for description in descriptions:
				if description.attributes["name"].value == "inputs":
					nrOfInputs = int(description.attributes["val"].value)
				elif "negate" in description.attributes["name"].value:
					value = description.attributes["name"].value[len('negate'):]
					index = int(value)
					negatedInputIndexes.append(index)
				elif description.attributes["name"].value == "size":
					size = int(description.attributes["val"].value)
				elif description.attributes["name"].value == "facing":
					facing = description.attributes["val"].value
				elif description.attributes["name"].value == "label":
					gate.name = description.attributes["val"].value	
				elif description.attributes["name"].value == "width":
					gate.bits = int(description.attributes["val"].value)
				elif description.attributes["name"].value == "control":
					controlSide = description.attributes["val"].value
				elif description.attributes["name"].value == "gate":
					controlSide = description.attributes["val"].value
				elif description.attributes["name"].value == "out":
					out = description.attributes["val"].value
				elif description.attributes["name"].value == "type":
					gate.transistorType = description.attributes["val"].value
					transistorType = gate.transistorType
				elif description.attributes["name"].value == "xor":
					gate.OneInputMultipleInputBehavior = False
		
			gate.out = out
			coordinates = getCoordinationTuple(xmlComp.attributes["loc"].value)
			annotatedGate = self.annotatedFactory.giveAnnotatedGate(gate, nrOfInputs, coordinates, negatedInputIndexes, size, facing, controlSide, transistorType)
			annotatedGateList.append(annotatedGate)	
			circuit.removeChild(xmlComp)	
			
		return annotatedGateList	

	def getAnnotatedPinList(self, circuit):		
		""" Returns a list with all the pins and its important attributes (will be invoked when we parse a circuit)
		(We need to know this before all the rest if we want to include this circuit in another circuit) """
		annotatedPinList = []				
		xmlCompList = circuit.getElementsByTagName("comp")
		for xmlComp in xmlCompList:
			input = True			
			#Default 
			facing = "east"
			if xmlComp.attributes["name"].value == "Pin":
				pin = StandardInputBase()
			else:
				continue
			
			descriptions = xmlComp.getElementsByTagName("a")
			for description in descriptions:
				if description.attributes["name"].value == "facing":
					facing = description.attributes["val"].value
				elif description.attributes["name"].value == "label":
					pin.name = description.attributes["val"].value	
				elif description.attributes["name"].value == "width":
					pin.bits = int(description.attributes["val"].value)
				elif description.attributes["name"].value == "output" and description.attributes["val"].value != "false":
					pin = StandardOutputBase()
					input = False
				#Only an input pin can have pull behavior (is important if this circuit is imported into another circuit)
				elif description.attributes["name"].value == "pull" and isinstance(pin, StandardInputBase):
					pin.pullBehavior = "1" if description.attributes["val"].value == 'up' else "0"
							
			coordinates = getCoordinationTuple(xmlComp.attributes["loc"].value)
			annotatedBase = self.annotatedFactory.giveAnnotatedBase(pin, coordinates, facing, input) 
			annotatedPinList.append(annotatedBase)	
			circuit.removeChild(xmlComp)	
			
		return annotatedPinList	

	def getAnnotatedBaseList(self, circuit):		
		""" Returns a list with all the bases and its important attributes """
		annotatedBaseList = []				
		xmlCompList = circuit.getElementsByTagName("comp")
		for xmlComp in xmlCompList:
			input = True			
			#Default 
			facing = "east"
			power = False
			if xmlComp.attributes["name"].value == "Constant":
				base = ConstantBase()
			elif xmlComp.attributes["name"].value == "Probe":
				base = ProbeBase()
				base.bits = None
				input = False
			elif xmlComp.attributes["name"].value == "Clock":
				base = ClockBase()
			# Add support for Power and Ground
			# these are actually just constants with a fixed value
			elif xmlComp.attributes["name"].value == "Power":
				base = ConstantBase()
				base.value = 1
				base.facing = "north"
				power = True
			elif xmlComp.attributes["name"].value == "Ground":
				base = ConstantBase()
				base.value = 0
				base.facing = "south"
			else:
				continue

			descriptions = xmlComp.getElementsByTagName("a")
			for description in descriptions:
				if description.attributes["name"].value == "facing":
					facing = description.attributes["val"].value
				elif description.attributes["name"].value == "label":
					base.name = description.attributes["val"].value	
				elif description.attributes["name"].value == "width":
					base.bits = int(description.attributes["val"].value)
					if power:
						# A Power does not save its value in the circ file because it is constant
						# So we have to change the value ourselves
						base.value = hex(pow(2, base.bits) - 1)[2:]
				#If there is a value specified, then we know this base is a constant	
				elif description.attributes["name"].value == "value":
					#We don't save "0x", because we know it is stored as a hexadecimal
					base.value = description.attributes["val"].value[2:]
				#If there is a duration specified, then we know this base is a clock	
				elif description.attributes["name"].value == "highDuration":
					base.highDuration = int(description.attributes["val"].value)
				elif description.attributes["name"].value == "lowDuration":
					base.lowDuration = int(description.attributes["val"].value)	
				#If there is a radix specified we know this is a probe base	
				elif description.attributes["name"].value == "radix":
					base.radix = description.attributes["val"].value	
				elif description.attributes["name"].value == "label":
					base.name = description.attributes["val"].value
		
			coordinates = getCoordinationTuple(xmlComp.attributes["loc"].value)
			annotatedBase = self.annotatedFactory.giveAnnotatedBase(base, coordinates, facing, input) 
			annotatedBaseList.append(annotatedBase)	
			circuit.removeChild(xmlComp)	
		return annotatedBaseList	

	def getAnnotatedPlexerList(self, circuit):		
		""" Returns a list with all the plexers and its important attributes """
		annotatedPlexerList = []				

		xmlCompList = circuit.getElementsByTagName("comp")
		for xmlComp in xmlCompList:
			#Default 
			facing = "east"
			# The enable line
			disabled = "floating"
			direction = "bl"
			if version == "2.7.0" or version == "2.7.1":
				enable = True
			elif version == "2.6.1" or version == "2.6.0":
				enable = False	

			#When we have a decoder or demultiplexer, this will be transformed into a bool
			hasInput = None
			if xmlComp.attributes["name"].value == "Multiplexer":
				plexer = Multiplexer()
			elif xmlComp.attributes["name"].value == "Demultiplexer":
				plexer = Demultiplexer()
				hasInput = True
			elif xmlComp.attributes["name"].value == "Decoder":
				plexer = Decoder()
				hasInput = False 
			elif xmlComp.attributes["name"].value == "BitSelector":
				plexer = BitSelector()	
			else:
				continue
		
			descriptions = xmlComp.getElementsByTagName("a")
			for description in descriptions:
				#If tristate is on, we know this has to be a demultiplexer or decorder
				if description.attributes["name"].value == "tristate":
					plexer.threeState = 'true' == description.attributes["val"].value
				elif description.attributes["name"].value == "facing":
					facing = description.attributes["val"].value
				elif description.attributes["name"].value == "select":
					plexer.selectBits = int(description.attributes["val"].value)
				elif description.attributes["name"].value == "width":
					plexer.bits = int(description.attributes["val"].value)
				elif description.attributes["name"].value == "group":
					plexer.outputBit = int(description.attributes["val"].value)	
				elif description.attributes["name"].value == "disabled":
					disabled = description.attributes["val"].value	
				elif description.attributes["name"].value == "selloc":
					direction = description.attributes["val"].value	
				elif description.attributes["name"].value == "enable":
					if description.attributes["val"].value == "false":
						enable = False
			plexer.disabled = disabled
			coordinates = getCoordinationTuple(xmlComp.attributes["loc"].value)
			annotatedPlexer = self.annotatedFactory.giveAnnotatedPlexer(plexer, coordinates, facing, hasInput, enable, direction)		
			annotatedPlexerList.append(annotatedPlexer)	
			circuit.removeChild(xmlComp)	
		return annotatedPlexerList

	def getPriorityEncoderList(self, circuit):	
		""" Returns a list with all the priorityEncoders """
		annotatedPriorityList = []				

		xmlCompList = circuit.getElementsByTagName("comp")
		for xmlComp in xmlCompList:
			#Default 
			facing = "east"
			disabled = "floating"
			if xmlComp.attributes["name"].value == "Priority Encoder":
				priority = PriorityEncoderNode()
			else:
				continue

			descriptions = xmlComp.getElementsByTagName("a")
			for description in descriptions:
				if description.attributes["name"].value == "facing":
					facing = description.attributes["val"].value
				elif description.attributes["name"].value == "select":
					priority.bits = int(description.attributes["val"].value)
				elif description.attributes["name"].value == "disabled":
					disabled = description.attributes["val"].value	
			priority.disabled = disabled
			coordinates = getCoordinationTuple(xmlComp.attributes["loc"].value)
			annotatedPriority = self.annotatedFactory.giveAnnotatedPriority(priority, coordinates, facing)		
			annotatedPriorityList.append(annotatedPriority)	
			circuit.removeChild(xmlComp)	
		return annotatedPriorityList	

	def getBitExtenderList(self, circuit):		
		""" Returns a list with all the bitExtenders """
		annotatedExtenderList = []				

		xmlCompList = circuit.getElementsByTagName("comp")
		for xmlComp in xmlCompList:
			if xmlComp.attributes["name"].value == "Bit Extender":
				extender = BitExtenderNode()
			else:
				continue
			descriptions = xmlComp.getElementsByTagName("a")
			for description in descriptions:
				if description.attributes["name"].value == "in_width":
					extender.widthIn = int(description.attributes["val"].value)
				elif description.attributes["name"].value == "out_width":
					extender.bits = int(description.attributes["val"].value)	
				elif description.attributes["name"].value == "type":
					extender.extensionType = description.attributes["val"].value

			coordinates = getCoordinationTuple(xmlComp.attributes["loc"].value)
			annotatedExtender = self.annotatedFactory.giveAnnotatedExtender(extender, coordinates)		
			annotatedExtenderList.append(annotatedExtender)	
			circuit.removeChild(xmlComp)	
		return annotatedExtenderList	

	def getAnnotatedArithmeticList(self, circuit): 		
		""" Returns a list with all the arithmetics and its important attributes """
		annotatedArithmeticList = []				

		xmlCompList = circuit.getElementsByTagName("comp")
		for xmlComp in xmlCompList:
			if xmlComp.attributes["name"].value == "Divider":
				arithmetic = Divider()
			elif xmlComp.attributes["name"].value == "Multiplier":
				arithmetic = Multiplier()
			elif xmlComp.attributes["name"].value == "Subtractor":
				arithmetic = Subtractor()
			elif xmlComp.attributes["name"].value == "Adder":
				arithmetic = Adder()
			elif xmlComp.attributes["name"].value == "BitSelector":
				arithmetic = BitSelector()
			elif xmlComp.attributes["name"].value == "Shifter":
				arithmetic = Shifter()	
			elif xmlComp.attributes["name"].value == "BitAdder":
				arithmetic = BitAdder()		
			elif xmlComp.attributes["name"].value == "Negator":
				arithmetic = Negator()	
			elif xmlComp.attributes["name"].value == "Comparator":
				arithmetic = Comparator()		
			elif xmlComp.attributes["name"].value == "BitFinder":
				arithmetic = BitFinder()			
			else:
				continue
			descriptions = xmlComp.getElementsByTagName("a")
			for description in descriptions:
				if description.attributes["name"].value == "width":
					arithmetic.bits = int(description.attributes["val"].value)
				elif description.attributes["name"].value == "inputs":
					arithmetic.numberOfInputs = int(description.attributes["val"].value)
				elif description.attributes["name"].value == "shift":
					arithmetic.shiftType = (description.attributes["val"].value[0], description.attributes["val"].value[1]) 
				elif description.attributes["name"].value == "type":
					arithmetic.type = (description.attributes["val"].value[0:-1], int(description.attributes["val"].value[-1]))
				elif description.attributes["name"].value == "mode":
					arithmetic.numericType = description.attributes["val"].value			
			
			coordinates = getCoordinationTuple(xmlComp.attributes["loc"].value)
			annotatedArithmetic = self.annotatedFactory.giveAnnotatedArithmetic(arithmetic, coordinates)		
			annotatedArithmeticList.append(annotatedArithmetic)	
			circuit.removeChild(xmlComp)	
		return annotatedArithmeticList	

	def getAnnotatedInOutList(self, circuit):	
		""" Returns a list with all the in- and outputs """
		annotatedInOutList = []				

		xmlCompList = circuit.getElementsByTagName("comp")
		for xmlComp in xmlCompList:
			facing = None
			if xmlComp.attributes["name"].value == "Button":
				inOut = ButtonBase()
				facing = "east"
			elif xmlComp.attributes["name"].value == "LED":
				inOut = LedBase()	
				facing = "east"
			elif xmlComp.attributes["name"].value == "Joystick":
				inOut = JoystickBase()	
			elif xmlComp.attributes["name"].value == "7-Segment Display":
				inOut = Segment7Display()		
			elif xmlComp.attributes["name"].value == "Hex Digit Display":
				inOut = HexDigitDisplay()		
			elif xmlComp.attributes["name"].value == "DotMatrix":
				inOut = LedMatrix()					
			else:
				continue
			
			descriptions = xmlComp.getElementsByTagName("a")
			for description in descriptions:
				if description.attributes["name"].value == "facing":
					facing = description.attributes["val"].value
				elif description.attributes["name"].value == "color":
					inOut.color = description.attributes["val"].value
				elif description.attributes["name"].value == "offcolor":
					inOut.offColor = description.attributes["val"].value	
				elif description.attributes["name"].value == "label":
					inOut.name = description.attributes["val"].value	
				elif description.attributes["name"].value == "bits":
					inOut.bits = int(description.attributes["val"].value)
				elif description.attributes["name"].value == "bg":
					inOut.backGround = description.attributes["val"].value
				elif description.attributes["name"].value == "matrixcols":
					inOut.numberOfInputs = int(description.attributes["val"].value)
				elif description.attributes["name"].value == "matrixrows":
					inOut.bits = int(description.attributes["val"].value)	
				elif description.attributes["name"].value == "persist":
					inOut.persist = int(description.attributes["val"].value)		
				elif description.attributes["name"].value == "dotshape":
					inOut.dotShape = description.attributes["val"].value
			
			coordinates = getCoordinationTuple(xmlComp.attributes["loc"].value)
			annotatedInOut = self.annotatedFactory.giveAnnotatedInOut(inOut, coordinates, facing)		
			annotatedInOutList.append(annotatedInOut)	
			circuit.removeChild(xmlComp)	
		return annotatedInOutList

	def getMemoryList(self, circuit):	
		""" Returns a list with all the memories """
		annotatedMemoryList = []	
		annotatedMemoryList.extend(self.getFlipFlopList(circuit))
		annotatedMemoryList.extend(self.getRandomList(circuit))
		annotatedMemoryList.extend(self.getAbstractRegisterList(circuit))
		annotatedMemoryList.extend(self.getShiftRegister(circuit))
		annotatedMemoryList.extend(self.getLargeMemory(circuit))
		return annotatedMemoryList	

	def getFlipFlopList(self, circuit):			
		""" Returns a list with all the flipflops """
		annotatedFlipFlopList = []				

		xmlCompList = circuit.getElementsByTagName("comp")
		for xmlComp in xmlCompList:
			oneInput = True
			if xmlComp.attributes["name"].value == "D Flip-Flop":
				flipflop = DFlipFlop()
			elif xmlComp.attributes["name"].value == "T Flip-Flop":
				flipflop = TFlipFlop()	
			elif xmlComp.attributes["name"].value == "J-K Flip-Flop":
				flipflop = JKFlipFlop()	
				oneInput = False
			elif xmlComp.attributes["name"].value == "S-R Flip-Flop":
				flipflop = SRFlipFlop()					
				oneInput = False
			else:
				continue
			descriptions = xmlComp.getElementsByTagName("a")
			for description in descriptions:
				if description.attributes["name"].value == "trigger":
					flipflop.trigger = description.attributes["val"].value
				elif description.attributes["name"].value == "label":
					flipflop.name = description.attributes["val"].value	
			
			coordinates = getCoordinationTuple(xmlComp.attributes["loc"].value)
			annotatedFlipFlop = self.annotatedFactory.giveAnnotatedFlipFlop(flipflop, coordinates, oneInput)		
			annotatedFlipFlopList.append(annotatedFlipFlop)	
			circuit.removeChild(xmlComp)	
		return annotatedFlipFlopList	

	def getRandomList(self, circuit):	
		""" Returns a list with all the random memory elements """
		annotatedRandomList = []				

		xmlCompList = circuit.getElementsByTagName("comp")
		for xmlComp in xmlCompList:
			if xmlComp.attributes["name"].value == "Random":
				rand = Random()
			else:
				continue
			descriptions = xmlComp.getElementsByTagName("a")
			for description in descriptions:
				if description.attributes["name"].value == "trigger":
					rand.trigger = description.attributes["val"].value
				elif description.attributes["name"].value == "label":
					rand.name = description.attributes["val"].value	
				elif description.attributes["name"].value =="width":
					rand.bits = int(description.attributes["val"].value)
				elif description.attributes["name"].value =="seed":
					rand.seed = int(description.attributes["val"].value)	
			
			coordinates = getCoordinationTuple(xmlComp.attributes["loc"].value)
			annotatedRand = self.annotatedFactory.giveAnnotatedRandom(rand, coordinates)		
			annotatedRandomList.append(annotatedRand)	
			circuit.removeChild(xmlComp)	
		return annotatedRandomList	

	def getAbstractRegisterList(self, circuit):		
		""" Returns a list with all the abstract register elements	"""
		annotatedAbstrRegList = []				

		xmlCompList = circuit.getElementsByTagName("comp")
		for xmlComp in xmlCompList:
			counter = False
			if xmlComp.attributes["name"].value == "Register":
				abstReg = Register()
			elif xmlComp.attributes["name"].value == "Counter":
				abstReg = Counter()	
				counter = True
			else:
				continue
			descriptions = xmlComp.getElementsByTagName("a")
			for description in descriptions:
				if description.attributes["name"].value == "trigger":
					abstReg.trigger = description.attributes["val"].value
				elif description.attributes["name"].value == "label":
					abstReg.name = description.attributes["val"].value	
				elif description.attributes["name"].value =="width":
					abstReg.bits = int(description.attributes["val"].value)
				elif description.attributes["name"].value =="max":
					abstReg.maxValue = description.attributes["val"].value
				elif description.attributes["name"].value =="ongoal":
					abstReg.actionOverflow = description.attributes["val"].value	
		
			coordinates = getCoordinationTuple(xmlComp.attributes["loc"].value)			
			annotatedAbstrReg = self.annotatedFactory.giveAnnotatedAbstractRegister(abstReg, coordinates, counter)		
			annotatedAbstrRegList.append(annotatedAbstrReg)	
			circuit.removeChild(xmlComp)	
		return annotatedAbstrRegList

	def getShiftRegister(self, circuit):		
		""" Returns a list with all the shift register elements	"""
		annotatedRegList = []				

		xmlCompList = circuit.getElementsByTagName("comp")
		for xmlComp in xmlCompList:
			parallelLoad = True
			if xmlComp.attributes["name"].value == "Shift Register":
				#We don't know yet which kind of register it will be
				reg = StandardShiftRegister()
				parReg = ParallelLoadShiftRegister()
			else:
				continue
			descriptions = xmlComp.getElementsByTagName("a")
			for description in descriptions:
				if description.attributes["name"].value == "trigger":
					reg.trigger = parReg.trigger = description.attributes["val"].value
				elif description.attributes["name"].value == "parallel":
					parallelLoad = False
				elif description.attributes["name"].value == "label":
					reg.name = parReg.name = description.attributes["val"].value	
				elif description.attributes["name"].value =="width":
					reg.bits = parReg.bits = int(description.attributes["val"].value)
				elif description.attributes["name"].value =="length":
					reg.numberOfStages = parReg.numberOfStages = int(description.attributes["val"].value)
					
			coordinates = getCoordinationTuple(xmlComp.attributes["loc"].value)
			annotatedReg = self.annotatedFactory.giveAnnotatedShiftRegister(reg, parReg, coordinates, parallelLoad)		
			annotatedRegList.append(annotatedReg)	
			circuit.removeChild(xmlComp)	
		return annotatedRegList	

	def convertROM(self, string):
		data = []
		temp = ""
		for character in string:
			if character == ' ' or character == '\n':
				data.append(temp)
				temp = ""
			else:
				temp += character
		# First 3 are the headers and such, so cut these of
		data = data[3:]
		return data

	def getLargeMemory(self, circuit):			
		""" Returns a list with all the large memories """	
		annotatedMemList = []				

		xmlCompList = circuit.getElementsByTagName("comp")
		for xmlComp in xmlCompList:
			rom = True
			if xmlComp.attributes["name"].value == "ROM":
				mem = Rom()
			elif xmlComp.attributes["name"].value == "RAM":	
				mem = Ram()
				rom = False
			else:
				continue
			descriptions = xmlComp.getElementsByTagName("a")
			# Initialize empty memory
			content = ['0']
			for description in descriptions:
				if description.attributes["name"].value =="addrWidth":
					mem.bits = int(description.attributes["val"].value)
				elif description.attributes["name"].value =="dataWidth":
					 mem.dataBit = int(description.attributes["val"].value)
				elif description.attributes["name"].value =="bus":
					 mem.dataInterface = description.attributes["val"].value
				elif description.attributes["name"].value =="contents":
					 # Fetch the data
					 data = description.childNodes[0].data
					 content = self.convertROM(data)
				
			coordinates = getCoordinationTuple(xmlComp.attributes["loc"].value)
			annotatedMem = self.annotatedFactory.giveAnnotatedLargeMemory(mem, coordinates, rom, content)		
			annotatedMemList.append(annotatedMem)	
			circuit.removeChild(xmlComp)	
		return annotatedMemList	

	def makeInt(self, value):
		""" Make an int, use rounding to nearest addressable coordinate
		
		Addressable coordinates are 0, 10, 20, ..."""
		ret = int((int(value) + 5) / 10) * 10
		return ret

	def getMapping(self, circuit):
		""" Search all the mappings for this circuit """
		circuitName = circuit.attributes["name"].value.replace(" ", "").replace("/", "")
		appearance = circuit.getElementsByTagName("appear")
		mapping = []
		if appearance != []:
			appearance = appearance[0]

			# Search for the anchor
			anchor = appearance.getElementsByTagName("circ-anchor")[0]
			anchorX = self.makeInt(anchor.attributes["x"].value)
			anchorY = self.makeInt(anchor.attributes["y"].value)

			# Search for the ports
			ports = appearance.getElementsByTagName("circ-port")
			for port in ports:
				x = self.makeInt(port.attributes["x"].value)
				y = self.makeInt(port.attributes["y"].value)
				pinx = int(port.attributes["pin"].value.partition(",")[0])
				piny = int(port.attributes["pin"].value.partition(",")[2])
				pin = pinx, piny
				remapX = x - anchorX
				remapY = y - anchorY
				mapping.append([pin, remapX, remapY])
		self.remapping[circuitName] = mapping

	def getImportedList(self, circuit, circuitTupleList):		
		""" Returns a list with all the imported circuits """	
		xmlCompList = circuit.getElementsByTagName("comp")

		#If the len is equal to zero, we know that all the components were resolved and so we can stop here
		if len(xmlCompList) == 0:
			return []
		#First we need to know all the names of the circuits, so we know which circuit we are importing
		circuitDic = dict()
		for circuitTuple in circuitTupleList:
			circuitDic[circuitTuple[0].name] = circuitTuple
			
		#We can start the actual resolving	
		annotatedImportedCircuitList = []
		for xmlComp in xmlCompList:
			#If the component has a library, we know that this isn't an imported circuit
			if "lib" in xmlComp.attributes:
				continue
			facing = "east"
			impCircuit = ImportedCircuitNode(circuitDic[xmlComp.attributes["name"].value.replace(" ", "").replace("/", "")][0])
			descriptions = xmlComp.getElementsByTagName("a")
			for description in descriptions:
				if description.attributes["name"].value =="facing":
					facing = description.attributes["val"].value
			coordinates = getCoordinationTuple(xmlComp.attributes["loc"].value)
			annotatedImportedCircuit = self.annotatedFactory.giveAnnotatedImportedCircuit(impCircuit, coordinates, facing, circuitDic[xmlComp.attributes["name"].value.replace(" ", "").replace("/", "")][2])
			annotatedImportedCircuitList.append(annotatedImportedCircuit)
			circuit.removeChild(xmlComp)	
			circuitname = xmlComp.attributes["name"].value.replace(" ", "").replace("/", "")
			try:
				mapping = self.remapping[circuitname]
				uniquemap = []
				for i in mapping:
					x, y = coordinates
					newx = i[1] + x
					newy = i[2] + y
					uniquemap.append([i[0], newx, newy])
			except Exception:
				uniquemap = []
			annotatedImportedCircuit.mapping = uniquemap

		return annotatedImportedCircuitList	
	
	def makeWiresForSplitters(self, annotatedSplitterList):			
		""" For each coordinate of the splitter we make a wire """	
		wireList = []
		for annotatedSplitter in annotatedSplitterList:
			coordinateList = annotatedSplitter.getWireCoordinates()
			for coordinates in coordinateList:
				wire = Wire()
				(coordinatesFrom, coordinatesTo) = coordinates
				wireList.append(AnnotatedStandardConnection(wire, coordinatesFrom, coordinatesTo))		
		return wireList	

	def addTunnelsToConnectionList(self, annotatedTunnelList, annotatedComponentList, annotatedConnectionList):			
		""" List with all the tunnels is added to the connectionslist, before we can add the tunnel we need to find out which tunnel object gets the input signal """
		foundNew = True
		outputList = getOutputList(annotatedComponentList)
		#If we find a new tunnel with an input signal, we have to keep looking (because for example tunnels can be positioned in a queue)
		while foundNew:
			foundNew = False
			#The tunnels which we found an input signal for
			labelsWithFrom = []
			for label in annotatedTunnelList:
				#If there is only one tunnel with this label, we don't need to look for an input signal, this tunnel has no meaning
				#It cannot send it's signal to another tunnel with the same label, because there is only one tunnel with this label
				if len(label[1]) == 1:
					label[1][0].coordinatesFrom = label[1][0].coordinatesTo	
					annotatedConnectionList.append(label[1][0])		
					labelsWithFrom.append(label)
				else:
					#Gives the index nummer of the tunnel which will get the input signal
					nr = self.searchInputTunnel(label[1], annotatedComponentList, annotatedConnectionList, outputList)
					if nr != None:
						foundNew = True
						for tunnel in label[1]:
							#The tunnel which will get the input signal, will be the starting point of the connections, all the other tunnels will be the end points
							if tunnel != label[1][nr]:
								tunnel.coordinatesFrom = label[1][nr].coordinatesFrom
								annotatedConnectionList.append(tunnel)		
						labelsWithFrom.append(label)
			#If we found a tunnel with an input signal, we don't need to look anymore for these tunnels			
			for label in labelsWithFrom:	
				annotatedTunnelList.remove(label)
		#Tunnels who don't have an input signal (they are meaningless, but still we keep track of them)
		for label in annotatedTunnelList:
			annotatedConnectionList.append(tunnel)	

	def searchInputTunnel(self, tunnelList, annotatedComponentList, annotatedConnectionList, outputList):	
		""" Returns the index nr of the tunnel who gets the input signal """
		for tunnel in tunnelList:
			#When the tunnel isn't directly connected with a component, we have to traverse the wires
			#We use a copy of the wirelist, so that we don't mess things up 
			copyOfWireList = deepcopy(annotatedConnectionList)
			tmpList = []
			#We bind the wires (the other connections) with the tunnels
			tunnel.bindDirectlyConnectedOutputWires(tunnel.coordinatesTo, tmpList, copyOfWireList, self.collide)
			tunnel.bindIndirectlyConnectedOutputWires(tmpList, copyOfWireList, outputList, self.collide)

			for annotatedList in annotatedComponentList:
				for annotatedElement in annotatedList:
					found = False
					#If the tunnel is directly connected to an elements output, we get the input signal from that element
					if annotatedElement.coordinates == tunnel.coordinatesTo:
						found = True
						tunnel.coordinatesFrom = tunnel.coordinatesTo
					#If the tunnel is connected through other connections, we get the input signal from that element	
					for tmpWire in tmpList:	
						if annotatedElement.coordinates == tmpWire.coordinatesTo:
							found = True
							tunnel.coordinatesFrom = tunnel.coordinatesTo
					if found:
						return tunnelList.index(tunnel)
		#If there isn't an input signal for these tunnels
		return None

	def bindCompAndSplitterWithWires(self, annotatedComponentList, annotatedConnectionList, annotatedSplitterList, circuitName = ""):	
		""" We bind all the comps with their wires """
		#We need to bind first all the output wires
		self.bindCompAndSplitterWithOutputs(annotatedComponentList, annotatedConnectionList, annotatedSplitterList, circuitName)	
		#After all the output wires are set right, we need to set the input wires right
		self.bindCompAndSplitterWithInputs(annotatedComponentList, annotatedConnectionList, annotatedSplitterList, circuitName)					

	def bindCompAndSplitterWithInputs(self, annotatedComponentList, annotatedConnectionList, annotatedSplitterList, circuitName=""):	 	
		""" We bind all the comps with their input wires """	
		#After all the output wires are set right, we need to set the input wires right 
		for annotatedList in annotatedComponentList:
			for annotatedComponent in annotatedList:
				annotatedComponent.bindWithInputWiresOrComp(annotatedComponentList, annotatedConnectionList)	
				
		for annotatedSplitter in annotatedSplitterList:
			annotatedSplitter.bindWithInputWires(annotatedConnectionList)

	def bindCompAndSplitterWithOutputs(self, annotatedComponentList, annotatedConnectionList, annotatedSplitterList, circuitName = ""):		
		""" Bind the comps with their output wires """
		#If we tolerate colliding wires, we need to delete wires at the end of the whole cycle 
		wiresToDelete = []
		#Wires that need to be inserted in the connectionList
		wiresToInsert = []
		#List with all the components their input coordinates
		inputList = getInputList(annotatedComponentList)	
		
		for annotatedList in annotatedComponentList:
			for annotatedComponent in annotatedList:	
				annotatedComponent.bindOutput(annotatedConnectionList, annotatedComponentList, inputList, wiresToDelete, wiresToInsert, self.collide, self.faultList, circuitName)
		
		#Bind the splitter with it's assumed output wires
		for annotatedSplitter in annotatedSplitterList:
			annotatedSplitter.bindOutputOfSplitter(annotatedComponentList, annotatedConnectionList, inputList, wiresToDelete, wiresToInsert, self.collide, self.faultList, circuitName)
		
		#If the wires are allowed to collide, we need to delete the duplicated wires 
		if self.collide:
			deleteElementsFromElementList(set(wiresToDelete), annotatedConnectionList)
			self.bindCollidingWires(wiresToInsert)
			annotatedConnectionList.extend(wiresToInsert)		
	
		# More optimized version of the following block of code
		lookupTable = {}
		for connection in annotatedConnectionList:
			coordinatesFrom = connection.coordinatesFrom
			if coordinatesFrom not in lookupTable.keys():
				lookupTable[coordinatesFrom] = []
			lookupTable[coordinatesFrom].append(connection)

		for i in annotatedConnectionList:
			key = i.coordinatesFrom
			########
			# Sometimes, key None is used, which will cause conflicts
			if key == None:
				continue
			########
			for j in lookupTable[key]:
				if i.connection.direction != j.connection.direction:
					#The directions are contradictory
					if i.connection.direction != None and j.connection.direction != None:
						i.connection.direction = j.connection.direction = "X"
					#One of the wires didn't know there was a pull resistor attached to it, so we need to change the direction	
					elif i.connection.direction != None:
						j.connection.direction = i.connection.direction
					else:
						i.connection.direction = j.connection.direction						
					
			
		#If wires have the same starting point and one of the wires has another pull direction, we know we need to adapt the pull direction	
		#for i in annotatedConnectionList:
			#for j in annotatedConnectionList:
			#	if i.coordinatesFrom == j.coordinatesFrom and i.connection.direction != j.connection.direction:
			#		#The directions are contradictory
			#		if i.connection.direction != None and j.connection.direction != None:
			#			i.connection.direction = j.connection.direction = "X"
			#		#One of the wires didn't know there was a pull resistor attached to it, so we need to change the direction	
			#		elif i.connection.direction != None:
			#			j.connection.direction = i.connection.direction
			#		else:
			#			i.connection.direction = j.connection.direction						

	def bindCollidingWires(self, wiresToInsert):						
		""" If there are wires, that end in the same point and colliding is allowed, we need to make a new wire, that binds those wires into 1 wire --- Optimized version"""	
		collidingWires = []
		lookupTable = {}
		for wire in wiresToInsert:
			coordinatesTo = wire.coordinatesTo
			if coordinatesTo not in lookupTable.keys():
				lookupTable[coordinatesTo] = []
			lookupTable[coordinatesTo].append(wire)

		for firstWire in wiresToInsert:
			key = firstWire.coordinatesTo
			if firstWire.coordinatesTo != None:
				collectingWire = None
				key = firstWire.coordinatesTo
				for secondWire in lookupTable[key]:
					if firstWire.coordinatesTo == secondWire.coordinatesTo and not firstWire.wireIsEqual(secondWire):
						#The original colliding wires their TO are set to None, 
						#so that we always have to use the wire that binds these colliding wires
						secondWire.coordinatesTo = None
						if collectingWire == None:
							collectingWire = AnnotatedStandardConnection(Wire(firstWire.connection.direction), None, firstWire.coordinatesTo)
						secondWire.connection.end = collectingWire.connection					
				if collectingWire != None:
					firstWire.coordinatesTo = None
					firstWire.connection.end = collectingWire.connection
					collidingWires.append(collectingWire)
		wiresToInsert.extend(collidingWires)			   	
		   	
	def bindCollidingWires2(self, wiresToInsert):						
		""" If there are wires, that end in the same point and colliding is allowed, we need to make a new wire, that binds those wires into 1 wire --- Original Version"""	
		collidingWires = []
		for firstWire in wiresToInsert:
			#If this wire isn't already bound in another wire
			if firstWire.coordinatesTo != None:
				#The wire that will bind the colliding wires
				collectingWire = None
				for secondWire in wiresToInsert:
					if firstWire.coordinatesTo == secondWire.coordinatesTo and not firstWire.wireIsEqual(secondWire):
						#The original colliding wires their TO are set to None, 
						#so that we always have to use the wire that binds these colliding wires
						secondWire.coordinatesTo = None
						if collectingWire == None:
							collectingWire = AnnotatedStandardConnection(Wire(firstWire.connection.direction), None, firstWire.coordinatesTo)
						secondWire.connection.end = collectingWire.connection					
				if collectingWire != None:
					firstWire.coordinatesTo = None
					firstWire.connection.end = collectingWire.connection
					collidingWires.append(collectingWire)
		wiresToInsert.extend(collidingWires)			   	

		
	
	
	
	
	
	
