#!/usr/bin/env python3

import os
from math import ceil

from translator.visitor.verilogVisitor.verilogVisitor import VerilogVisitor
from translator.intermediateBuilder.wire import Wire
from translator.intermediateBuilder.input import Input
from translator.intermediateBuilder.numberedInput import NumberedInput
from translator.intermediateBuilder.standardInputBase import InputBase
from translator.intermediateBuilder.output import Output
from translator.intermediateBuilder.joystickBase import JoystickBase as Joystick
from translator.intermediateBuilder.circuit import Circuit as Module
from translator.intermediateBuilder.importedCircuitNode import ImportedCircuitNode as ImportedModule

class connectionElement:
	""" A connection element is an element of a connection 
	
	It contains the element, the port name and whether it is an input or not
	"""
	def __init__(self, name, element, isInput):
		""" The constructor of the connection element """
		self.element = element
		self.name = name
		# Needed when connecting multiple lines together
		# Need to make sure that only one input, otherwise DEVSgenerator should make a merger element
		self.input = isInput

class wireCollection:
	""" A wire collection is the connection itself

	Additionally, it keeps track of all wires that are responsible for this connection
	"""
	def __init__(self, leadingWire):
		""" The constructor of the wire collection """
		self.elements = []
		if isinstance(leadingWire, list):
			self.wires = leadingWire
		else:
			self.wires = [leadingWire]

	def add(self, name, element, isInput):
		""" Adds an element, name, input to the wire collection """
		self.elements.append(connectionElement(name, element, isInput))

# VerilogVisitor is a rather sad name for this, however, it had this name
class DEVSTranslatorVisitor(VerilogVisitor):
	def __init__(self):
		""" Constructor """
		self.connectionNr = 0
		# Used to pair up (or even more than a pair...)
		self.connections = []
		self.colliding = []
		self.wires = []
		# Maps all wires and their index
		self.wireIndexes = {}
		# Maps all wires to the index that is searching for this wire
		self.searchingWires = {}

	def merge(self, first, second):
		""" Merge the different wire collections """
		elements = []
		wires = []
		for i in second.wires:
			if i not in wires:
				wires.append(i)
				self.wireIndexes[i] = len(self.connections)
		for i in first.wires:
			if i not in wires:
				wires.append(i)
				self.wireIndexes[i] = len(self.connections)
		newConnection = wireCollection(wires)
		for i in second.elements:
			if i not in elements:
				elements.append(i)
		for i in first.elements:
			if i not in elements:
				elements.append(i)
		newConnection.elements = elements
		first.elements = []
		first.wires = []
		second.elements = []
		second.wires = []
		self.connections.append(newConnection)

	def makeCollide(self):
		""" CollidingWires in the intermediate trees are merged """
		removeList = []
		for i in self.connections:
			for wire in i.wires:
				if ((isinstance(wire.start, Wire)) or (isinstance(wire.end, Wire))):
					# If this is a wire with a wire as start/end
					if isinstance(wire.start, Wire):
						if wire.start not in self.wireIndexes.keys():
							wirecollection = wireCollection(wire.start)
							self.wireIndexes[wire.start] = len(self.connections)
							self.connections.append(wirecollection)
						index = self.wireIndexes[wire.start]
						mergeWire = self.connections[self.wireIndexes[wire.start]]
						if(index == self.connections.index(i)):
							# Would merge with self
							continue
						self.merge(i, mergeWire)
						removeList.append(self.connections.index(i))
						removeList.append(index)
					elif isinstance(wire.end, Wire):
						if wire.end not in self.wireIndexes.keys():
							wirecollection = wireCollection(wire.end)
							self.wireIndexes[wire.end] = len(self.connections)
							self.connections.append(wirecollection)
						index = self.wireIndexes[wire.end]
						mergeWire = self.connections[self.wireIndexes[wire.end]]
						if(index == self.connections.index(i)):
							# Would merge with self
							continue
						self.merge(i, mergeWire)
						removeList.append(self.connections.index(i))
						removeList.append(index)
		removeList.sort()
		removeList.reverse()
		prevPop = -1
		for i in removeList:
			if i == prevPop:
				continue
			self.connections.pop(i)
			prevPop = i

	def changeElementName(self, element, wire, name, newName):
		""" Changes the name of an element """
		for i in self.connections:
			if wire in i.wires:
				for j in i.elements:
					if j.name == name:
						j.name = newName

	def addConnection(self, name, element, wire, isInput):
		""" Recursive function to add a connection """
		if isinstance(wire, tuple):
			wire1, wire2 = wire
			if isinstance(wire1, Wire):
				self.addConnection(name, element, wire1, isInput)
			elif "name" in dir(wire1):
				self.changeElementName(element, wire2, name, wire1.name)
			else:
				# No name detected!
				self.changeElementName(element, wire2, name, "unnamedWire" + str(wire1))

			if isinstance(wire2, Wire):
				self.addConnection(name, element, wire2, isInput)
			elif "name" in dir(wire2):
				self.changeElementName(element, wire1, name, wire2.name)
			else:
				# No name detected!
				self.changeElementName(element, wire1, name, "unnamedWire" + str(wire2))
			return
		done = False
		for i in self.connections:
			# i is a wireCollection
			if wire in i.wires:
				# The wire that we connect to already exists
				i.add(name, element, isInput)
				self.wireIndexes[wire] = self.connections.index(i)
				done = True
				continue
			# Look if there is a similar element
			# Should do something about a inputBase element only...
			for j in i.elements:
				if j.name == name and j.element == element and j.input == isInput:
					# The same element already exists on another wire
					i.wires.append(wire)
					self.wireIndexes[wire] = self.connections.index(i)
					done = True
					continue
		if done:
			return

		# It seems the wire doesn't already exist
		self.connections.append(wireCollection(wire))
		# Recursive call to self
		self.addConnection(name, element, wire, isInput)

	def recurse(self, element, attribute, name):
		""" Recurse through an element and search for all wires """
		if isinstance(attribute, Input):
			self.addConnection(name, element, attribute.connection, True)
		elif isinstance(attribute, Output):
			for i in attribute.connections:
				self.addConnection(name, element, i, False)
		#Recursive 
		elif isinstance(attribute, list):
			listcounter = 0
			for listelem in attribute:
				if isinstance(element, ImportedModule):
					self.recurse(element, listelem, element.inputWires[listcounter].reference.name)
				elif isinstance(listelem, NumberedInput):
					self.recurse(element, listelem, name + str(listelem.number))
				else:	
					self.recurse(element, listelem, name + str(listcounter))
				listcounter += 1
		elif isinstance(attribute, tuple):
			# Used in standard gates, where an input can be negated
			negate, input = attribute
			self.recurse(element, input, name)
		else:
			pass

	def visit(self, element):
		""" The visit function """
		if isinstance(element, Module):
			return
		# Check all attributes of a file for wires
		if isinstance(element, Wire):
			# Don't do anything
			return
		for attr in dir(element):
			if not attr.startswith("__"):
				attribute = getattr(element, attr)
				self.recurse(element, attribute, attr)
