#!/usr/bin/env python3
from inspect import getmro

from translator.visitor.intermediateVisitor.intermediateVisitor import IntermediateVisitor

class IntermediateConnectivityVisitor(IntermediateVisitor):
	""" Intermediate Connectivity Visitor takes an intermediate tree element and checks whether it is transitively connected to another element  """
	
	def __init__(self):
		""" Constructor 	"""
		self.attributesOfElement1 = None
		self.attributesOfElement2 = None
		self.found = False

	def setChecks(self, attributesOfElement1, attributesOfElement2, noDirection= True, getCounters= False):		
		""" With this function we set the conditions for element1 and element2	(who must be connected) """
		self.attributesOfElement1 = attributesOfElement1
		self.attributesOfElement2 = attributesOfElement2
		self.noDirection = noDirection
		self.getCounters = getCounters
		self.found = [[], []] if self.getCounters else [False, False]

	def visit(self, element):		
		""" The visit function """
		if self.isRightElement(element, self.attributesOfElement1):
			isConnected = self.isConnected(element, self.attributesOfElement2)
			if self.getCounters:
				self.found[0].append(isConnected) 
			else:
				self.found[0] = isConnected if not self.found[0] else self.found[0]
		if self.noDirection and self.isRightElement(element, self.attributesOfElement2):
			isConnected = self.isConnected(element, self.attributesOfElement1)
			if self.getCounters:
				self.found[1].append(isConnected)
			else:
				self.found[1] = isConnected if not self.found[1] else self.found[1]	

	def isRightElement(self, element, compareElement):			
		""" Test whether the element meets the conditions of the compareElement """
		for attributeName, attributeValue in compareElement.items():
			#When one of the conditions isn't met, we should start looking in the following map
			if attributeName == 'class':
				#We get the elements class and parent classes
				classes = getmro(element.__class__)
				rightClass = False
				for cl in classes:
					#If one of it's parents classes is the right class, we are on the right track
					if cl.__name__ == attributeValue:
						rightClass = True
				#If no class was the right one, we know this element can't be the right one (and we should look in the following map)
				if not rightClass:
					return False
			elif attributeName not in element.__dict__ or element.__dict__[attributeName] != attributeValue:
				return False
		return True

	def result(self):			
		""" Give the results back """
		if self.noDirection:
			return self.found
		else:
			return self.found[0]

	def isConnected(self, startingPoint, endingInAttributes):		
		""" Searches depth first (correct direction only) """
		found = False
		alreadyVisitedNodes = []
		fringe = [startingPoint]
		count = 0
		while len(fringe) > 0 and not found:
			candidateSuccessors = fringe[0].getAllElementsViaOuputConnected()
			successors = []
			for candidate in candidateSuccessors:
				#Only look at the non-visited child states
				if candidate not in alreadyVisitedNodes and candidate not in fringe:
					successors.append(candidate) 
			alreadyVisitedNodes.append(fringe.pop(0))		
			#If there are no new child states, we should pop the fringe	
			if len(successors) > 0:    	
				fringe = successors + fringe
			if fringe != [] and self.isRightElement(fringe[0], endingInAttributes):
				if self.getCounters:
					count += 1	
				else:
					return True
		if self.getCounters:				
			return count  
		else:
			return False
		
		
		
		
		