#!/usr/bin/env python3
from translator.visitor.intermediateVisitor.intermediateVisitor import IntermediateVisitor
from translator.intermediateBuilder.intermediateBuilder import IntermediateBuilder
from translator.intermediateBuilder.negatedInput import NegatedInput 
from translator.intermediateBuilder.gateNode import GateNode
from translator.intermediateBuilder.node import Node 
from translator.intermediateBuilder.plexerNode import PlexerNode
from translator.intermediateBuilder.arithmeticNode import ArithmeticNode 
from translator.intermediateBuilder.carryArithmetic import CarryArithmetic 
from translator.intermediateBuilder.adder import Adder 
from translator.intermediateBuilder.multiplier import Multiplier 
from translator.intermediateBuilder.divider import Divider 
from translator.intermediateBuilder.subtractor import Subtractor 
from translator.intermediateBuilder.bitAdder import BitAdder 
from translator.intermediateBuilder.shifter import Shifter 
from translator.intermediateBuilder.negator import Negator 
from translator.intermediateBuilder.comparator import Comparator 
from translator.intermediateBuilder.bitFinder import BitFinder 
from translator.intermediateBuilder.andGate import AndGate 
from translator.intermediateBuilder.nandGate import NandGate 
from translator.intermediateBuilder.orGate import OrGate 
from translator.intermediateBuilder.norGate import NorGate 
from translator.intermediateBuilder.notGate import NotGate 
from translator.intermediateBuilder.xorGate import XorGate
from translator.intermediateBuilder.xnorGate import XnorGate 
from translator.intermediateBuilder.demultiplexer import Demultiplexer 
from translator.intermediateBuilder.multiplexer import Multiplexer 
from translator.intermediateBuilder.decoder import Decoder
from translator.intermediateBuilder.bitSelector import BitSelector
from translator.intermediateBuilder.priorityEncoderNode import PriorityEncoderNode
from translator.intermediateBuilder.bitExtenderNode import BitExtenderNode 
from translator.intermediateBuilder.evenParityGate import EvenParityGate
from translator.intermediateBuilder.oddParityGate import OddParityGate
from translator.intermediateBuilder.controlledBufferGate import ControlledBufferGate
from translator.intermediateBuilder.controlledInverterGate import ControlledInverterGate 
from translator.intermediateBuilder.bufferGate import BufferGate 
from translator.intermediateBuilder.inputBase import InputBase 
from translator.intermediateBuilder.joystickBase import JoystickBase
from translator.intermediateBuilder.outputBase import OutputBase 
from translator.intermediateBuilder.probeBase import ProbeBase
from translator.intermediateBuilder.clockBase import ClockBase
from translator.intermediateBuilder.constantBase import ConstantBase 
from translator.intermediateBuilder.baseNode import BaseNode
from translator.intermediateBuilder.memoryNode import MemoryNode 
from translator.intermediateBuilder.flipFlop import FlipFlop 
from translator.intermediateBuilder.dFlipFlop import DFlipFlop 
from translator.intermediateBuilder.tFlipFlop import TFlipFlop 
from translator.intermediateBuilder.srFlipFlop import SRFlipFlop 
from translator.intermediateBuilder.jkFlipFlop import JKFlipFlop 
from translator.intermediateBuilder.abstractRegister import AbstractRegister
from translator.intermediateBuilder.register import Register 
from translator.intermediateBuilder.counter import Counter 
from translator.intermediateBuilder.random import Random 
from translator.intermediateBuilder.largeMemory import LargeMemory 
from translator.intermediateBuilder.rom import Rom 
from translator.intermediateBuilder.ram import Ram 
from translator.intermediateBuilder.shiftRegister import ShiftRegister 
from translator.intermediateBuilder.standardShiftRegister import StandardShiftRegister 
from translator.intermediateBuilder.parallelLoadShiftRegister import ParallelLoadShiftRegister 
from translator.intermediateBuilder.importedCircuitNode import ImportedCircuitNode 
from translator.intermediateBuilder.splitter import Splitter 
from translator.intermediateBuilder.circuit import Circuit
from translator.intermediateBuilder.tunnel import Tunnel
from translator.intermediateBuilder.connection import Connection 
from translator.intermediateBuilder.helperFunctions import validName 

class IntermediateNamingVisitor(IntermediateVisitor):
	""" Intermediate Naming Visitor takes an intermediate tree node and gives all the nodes and splitters a (unique) name 	"""
	
	def __init__(self):
		""" Constructor """
		self.inputCount = 0
		self.outputCount = 0
		self.gateCount = 0
		self.netCount = 0
		self.parameterCount = 0
		self.statementCount = 0
		self.circuitCount = 0
		self.nameList = None

	def visit(self, element):	
		""" The visit function """
		if isinstance(element, Circuit):
			self.visitCircuit(element)
			return
			
		if isinstance(element, GateNode):
			element.name = self.visitGate(element)
			return
			
		#The connections will be named when we translate the intermediate tree to verilog	
		if isinstance(element, Connection):
			return
			
		if isinstance(element, BaseNode):
			if isinstance(element, JoystickBase):
				element.name = self.visitJoystick(element)
				return	
			element.name = self.visitBase(element)
			return
		
		if isinstance(element, PlexerNode):
			element.name = self.visitPlexer(element)
			return
			
		if isinstance(element, ArithmeticNode):
			element.name = self.visitArithmetic(element)
			return	
			
		if isinstance(element, PriorityEncoderNode):
			element.name = self.visitPriorityEncoder(element)
			return
			
		if isinstance(element, BitExtenderNode):
			element.name = self.visitBitExtender(element)
			return	
			
		if isinstance(element, MemoryNode):		
			element.name = self.visitMemory(element)
			return 
			
		if isinstance(element, ImportedCircuitNode):
			element.name = self.visitImportedCircuit(element)
			return 			
			
		if isinstance(element, Splitter):
			element.name = self.visitSplitter(element)
			return

	def visitCircuit(self, circuit):			
		""" If the element was a circuit, we don't need to make a name, because logisim requires that a circuit has a unique name """
		#We need to start counting again
		self.inputCount = 0
		self.outputCount = 0
		self.gateCount = 0
		self.netCount = 0
		self.parameterCount = 0
		self.statementCount = 0
		self.probeCount = 0
		#We clear the self.nameList
		self.nameList = []

	def visitGate(self, node):		
		""" If the element is an intermediate gate, we can use this function """
		self.gateCount += 1
		return validName(node.name, "gate", self.gateCount-1, self.nameList)

	def visitBase(self, node):	
		""" If the element is an intermediate base, we can use this function """
		name = node.name
		if isinstance(node, ProbeBase):
			name = validName(node.name, "probe", self.probeCount, self.nameList)
			self.probeCount += 1
		elif isinstance(node, OutputBase):
			name = validName(node.name, "output", self.outputCount, self.nameList)
			self.outputCount += 1
		elif isinstance(node, ClockBase):
			name = validName(node.name, "clock", self.parameterCount, self.nameList)
			self.parameterCount += 1
		elif isinstance(node, InputBase):
			name = validName(node.name, "input", self.inputCount, self.nameList)
			self.inputCount += 1
		elif isinstance(node, ConstantBase):
			name = validName(node.name, "constant", self.inputCount, self.nameList)
			self.inputCount += 1
	
		return name

	def visitJoystick(self, node):
		""" If the element is an intermediate joystick, we can use this function """
		#The first pin and the second pin will be translated into 2 inputs registers 
		#(joystick1 will equal the upper pin and joystick 2 will equal the lower pin)
		name1 = validName("", "in1Joystick", self.outputCount, self.nameList)
		name2 = validName("", "in2Joystick", self.outputCount, self.nameList)
		self.inputCount += 2
	
		return	(name1, name2)

	def visitPlexer(self, node):	
		""" If the element is an intermediate plexer, we can use this function """	
		self.statementCount +=1
		#Logisim doesn't allow plexers to be named
		return validName("", "plexer",self.statementCount-1, self.nameList)

	def visitArithmetic(self, node):		
		""" If the element is an intermediate arithmetic, we can use this function """	
		self.statementCount += 1
		if isinstance(node, CarryArithmetic):
			name = validName("", "carryArithmetic", self.statementCount-1, self.nameList)
		else:
			name = validName("", "arithmetic", self.statementCount-1, self.nameList)
		return name	
	
	def visitPriorityEncoder(self, node):			
		""" If the element is an intermediate priority encoder, we can use this function """	
		self.statementCount += 1	
		return validName("", "priorityEnc", self.statementCount-1, self.nameList)
		
	def visitBitExtender(self, node):	
		""" If the element is an intermediate bit extender, we can use this function """	
		self.statementCount += 1
		return validName("", "bitExtender", self.statementCount-1, self.nameList) 

	def visitMemory(self, node):		
		""" If the element is an intermediate memory, we can use this function """
		self.statementCount += 1
		if isinstance(node, FlipFlop):
			name = validName(node.name, "flipflop", self.statementCount-1, self.nameList)	 
		elif isinstance(node, AbstractRegister):
			name = validName(node.name, "abstrReg", self.statementCount-1, self.nameList)	 
		elif isinstance(node, Random):
			name = validName(node.name, "random", self.statementCount-1, self.nameList)	 
		elif isinstance(node, LargeMemory):
			name = validName(node.name, "largeMem", self.statementCount-1, self.nameList)	 
			if isinstance(node, Ram):
				node.extraNames = (validName("", name+"File", 0, self.nameList), validName("", name+"TmpReg", 0, self.nameList))	
		elif isinstance(node, ShiftRegister):
			name = validName(node.name, "shiftReg", self.statementCount-1, self.nameList) 
			
		return name

	def visitImportedCircuit(self, node):		
		""" If the element is an intermediate imported circuit, we can use this function """
		#name = validName("", "importedCircuit", self.statementCount, self.nameList)	 
		name = validName("", "importedCircuit", self.circuitCount, self.nameList)	 
		#self.statementCount += 1
		self.circuitCount += 1
		return name
	
	def visitSplitter(self, node):		
		""" If the element is an intermediate splitter, we can use this function """	
		name1 = validName("", "part1Splitter", self.statementCount, self.nameList)	 
		name2 = validName("", "part2Splitter", self.statementCount, self.nameList)	 
		self.statementCount += 2
		return (name1, name2)	
