/**
 * @author R. Diez
 */

package vcdEngine;

import java.io.BufferedReader;
import java.io.Closeable;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;

import vcdEngine.VcdKeywords.TokenTypeEnum;

public class VcdFileParser implements Closeable
{
    private enum StateEnum
    {
        STATE_NORMAL,

        STATE_SKIPPING_UNTIL_END_KEYWORD,
        STATE_EXPECTING_END_KEYWORD,
        
        // These keywords are only valid in the header section.
        
        STATE_READING_TIMESCALE,
        STATE_EXPECTING_TIMESCALE_UNIT,

        STATE_EXPECTING_SCOPE_TYPE,
        STATE_EXPECTING_SCOPE_ID,

        STATE_EXPECTING_VAR_TYPE,
        STATE_EXPECTING_VAR_SIZE,
        STATE_EXPECTING_VAR_ID,
        STATE_EXPECTING_VAR_REF,
        STATE_EXPECTING_VAR_INDEX_RANGE,

        STATE_EXPECTING_END_OF_ENDDEFINITIONS,

        
        // These keywords are only valid in the value change section.
        
        STATE_READING_DATA_INSIDE_DUMP_SECTION,

        STATE_EXPECTING_VECTOR_ID,
    }

    private final String STDIN_FILENAME = "-";
    private final String filenameForUserMessages;
    private final BufferedReader br;

    private StateEnum state = StateEnum.STATE_NORMAL;
    private boolean isInHeaderSection = true;  // The other section is called the value change section.

    private String currentLine       = "";
    private long   currentLineNumber = 0;
    private int    currentLineColumnNumber; // See comment below for more information on the column position.

    private VcdToken lastSectionStartToken;  // Helps generate error messages that mention where the failed section started.

    private ArrayList< String > scopeTypeStack = new ArrayList<String>();
    private ArrayList< String > scopeIdStack   = new ArrayList<String>();

    private VcdVariableManager.VcdVariable vcdVarDefinition;
    private String vcdVarDefinitionId;

    // The variable manager could be handed over by the caller.
    private VcdVariableManager vcdVarManager = new VcdVariableManager();

    private VcdTimescale timescale;  // Available only when the header section is complete.
    

    private long lastTimestampFound = VcdTimePeriod.TIMESTAMP_INITIAL_VALUES;

    private boolean isInsideReadingDataVarDumpSection = false;
    
    private String newValueForVcdVectorVar;


    public VcdFileParser ( String filename )
    {
        InputStream is;

        if ( filename.equals( STDIN_FILENAME ) )
        {
            is = System.in;
            filenameForUserMessages = "<stdin>";
        }
        else
        {
            filenameForUserMessages = filename;

            File f = new File( filename );

            try
            {
                is = new FileInputStream( f );
            }
            catch ( Throwable ex )
            {
                throw myutils.ErrUtils.asRuntimeExceptionWithFilename( ex, f, myutils.ErrUtils.ErrFileOp.FE_OPENING );
            }
        }

        // About the column position:
        //
        // Note that ASCII codes get translated to Java's internal Unicode representation
        // when reading with BufferedReader's readLine(). VCD files should only contain
        // ASCII characters, therefore there should be a one-to-one mapping between
        // ASCII and Java characters. Otherwise, the column position in the error messages
        // will not match.
        //
        // An alternative would have been to write a byte-level parser that reads
        // the ASCII codes without Unicode conversion. Such a parser is always accurate
        // in the column positions, but it's harder to write.
        //
        // Note also that tab characters count as 1 position, and not 4 or 8,
        // like displayed in a text editor.

        try
        {
            InputStreamReader isr = new InputStreamReader( is, "US-ASCII" );

            br = new BufferedReader( isr );
        }
        catch ( Throwable ex )
        {
            throw myutils.ErrUtils.asRuntimeException( ex, String.format( "Error opening file %s: ", filenameForUserMessages ) );
        }
    }

    
    // The first time this routine is called, it reads the VCD header and stops
    // when the first timestamp (or the end of the file) is reached.
    // VCD Variables will only be discovered and added to the variable manager
    // on this first call. If no valid timescale definition is found during this
    // first call, an error will be raised.
    //
    // Subsequent calls will only update the VCD variable values.
    
    public void parseUntilNextTimestamp ( VcdTimePeriod tp  // Reused in the hope of more performance.
                                        )
    {
        tp.reset();  // Helps find bugs, see ts.isValid() below.
        
        VcdToken currentToken = new VcdToken();  // Reuse in the hope of more performance.
        
        for ( ;; )
        {
            getNextToken( currentToken );

            if ( stepToken( currentToken, tp ) )
                break;
            
            // If stepToken() did not signal that a timestamp was found,
            // it should not have touched 'tp'.
            assert tp.isReset() : myutils.ErrUtils.assertionFailed();
        }
        
        assert tp.isValid() : myutils.ErrUtils.assertionFailed();
        
        boolean shouldTraceTimestampInfo = false;
        if ( shouldTraceTimestampInfo )
        {
            System.out.printf( "Timestamp %s found in file \"%s\"\n",
                               tp.format( timescale.getUnit() ),
                                          filenameForUserMessages );
        }
    }

    
    private void getNextToken ( VcdToken nextToken )
    {
        for ( ;; )
        {
            // Read the next line if necessary.
            if ( currentLine.length() == 0 || currentLineColumnNumber > currentLine.length() )
            {
                try
                {
                    currentLine = br.readLine();
                }
                catch ( Throwable ex )
                {
                    throw myutils.ErrUtils.asRuntimeException( ex, "Error reading from file: " );
                }

                if ( currentLine == null )
                {
                    nextToken.set( "", 0, 0 );  // Signals end of file.
                    return;
                }

                ++currentLineNumber;
                currentLineColumnNumber = 1;

                // System.out.printf("Line %d: %s\n", currentLineNumber, currentLine );

                continue; // The new line may be empty, so start from the
                          // beginning.
            }

            // Are we at the start of the next token?
            // BufferedReader.readLine() uses isWhitespace(), so use the same here.
            if ( Character.isWhitespace( currentLine.charAt( currentLineColumnNumber - 1 ) ) )
            {
                ++currentLineColumnNumber;
                continue;
            }

            break;
        }

        // We found the start position of the next token, now find the end position.
        int startColumnNumber = currentLineColumnNumber;

        do
        {
            ++currentLineColumnNumber;
        }
        while ( currentLineColumnNumber <= currentLine.length()
                && !Character.isWhitespace( currentLine.charAt( currentLineColumnNumber - 1 ) ) );

        nextToken.set( currentLine.substring( startColumnNumber - 1, currentLineColumnNumber - 1 ),
        		       currentLineNumber,
        		       startColumnNumber );

        assert nextToken.getTokenStr().length() > 0 : myutils.ErrUtils.assertionFailed();
    }
    

    // The VCD parser is decoupled from the file reading, its state machine gets
    // fed one token at a time. This is so that the parser can later be reused from
    // some other code that waits on several files/pipes/sockets at the same time.
    // If the parser read directly from the file, the thread would block every now
    // and then waiting for more data to be available. Such a parser is a pain to use,
    // especially if the user wants an asynchronous 'cancel' button.
    //
    // Returns true if the next timestamp has been reached.

    private boolean stepToken ( VcdToken tok, VcdTimePeriod ts )
    {
        boolean traceToken = false;
        if ( traceToken )
            tok.trace();

        // At the beginning, the parser looked up the token string here, in order to see
        // whether it was a VCD keyword like "$var". The trouble is, the variable identifiers
        // can actually start with '$'. You also want to allow words that start with '$'
        // inside $comment sections. Therefore, I moved the detection of the '$' keywords
        // to the points where one of them is actually expected.

        try
        {
            return stepToken2( tok, ts );
        }
        catch ( Throwable ex )
        {
            String prefix;

            if ( tok.isEndOfFile() )
            {
                prefix = "At end of file: ";
            }
            else
            {
                prefix = String.format( "Error in file \"%s\" at line %d, column %d: ",
                                        filenameForUserMessages,
                                        tok.getLineNumber(),
                                        tok.getColumnNumber() );
            }

            throw myutils.ErrUtils.asRuntimeException( ex, prefix );
        }
    }


    private boolean stepToken2 ( VcdToken tok, VcdTimePeriod ts )
    {
        if (  state == StateEnum.STATE_NORMAL )
        {
            if ( tok.isEndOfFile() )
            {
                // If the end of the file has been reached before the data part,
                // the header consistency check has not been done yet.
                if ( isInHeaderSection )
                    checkHeaderIsOk();
                
                ts.set( lastTimestampFound, VcdTimePeriod.TIMESTAMP_END_OF_FILE ); 

                // If the caller tries to read any further, it will get an error back.
                myutils.FileUtils.closeIgnoreErrors( this );
                
                return true;
            }

            if ( isInHeaderSection )
            {
                readingHeaderSection( tok );
                return false;  // There are no timestamps in the header section.
            }
            else
            {
                return readingValueChangeSection( tok, ts );
            }
        }

        
        checkNotEndOfFile( tok );

        switch ( state )
        {
        case STATE_EXPECTING_END_KEYWORD:
            expectingEnd( tok );
            state = StateEnum.STATE_NORMAL;
            break;

        case STATE_SKIPPING_UNTIL_END_KEYWORD:
            skipUntilEnd( tok );
            break;

        case STATE_EXPECTING_END_OF_ENDDEFINITIONS:
            expectingEnd( tok );

            // Transition from the header to the value change section.
            assert isInHeaderSection == true : myutils.ErrUtils.assertionFailed();
            isInHeaderSection = false;
            
            checkHeaderIsOk();
            
            // This variable is only used in the value change section.
            assert isInsideReadingDataVarDumpSection == false : myutils.ErrUtils.assertionFailed();
            
            state = StateEnum.STATE_NORMAL;
            break;

        case STATE_READING_TIMESCALE:
            readTimescale( tok );
            break;
        case STATE_EXPECTING_TIMESCALE_UNIT:
            readTimescaleUnit( tok );
            break;

        case STATE_EXPECTING_SCOPE_TYPE:
            readScopeType( tok );
            break;
        case STATE_EXPECTING_SCOPE_ID:
            readScopeId( tok );
            break;

        case STATE_EXPECTING_VAR_TYPE:
            readVarType( tok );
            break;
        case STATE_EXPECTING_VAR_SIZE:
            readVarSize( tok );
            break;
        case STATE_EXPECTING_VAR_ID:
            readVarId( tok );
            break;
        case STATE_EXPECTING_VAR_REF:
            readVarRef( tok );
            break;
        case STATE_EXPECTING_VAR_INDEX_RANGE:
            readVarIndexRange( tok );
            break;

        case STATE_EXPECTING_VECTOR_ID:
            readVectorId( tok );
            break;

        case STATE_READING_DATA_INSIDE_DUMP_SECTION:
            readVarDump( tok );
            break;

        default:
            assert false : myutils.ErrUtils.assertionFailed();
            break;
        }

        return false;  // Only readingHeader() can discover a timestamp.
    }

    
    private void checkHeaderIsOk ()
    {
        if ( timescale == null )
        {
            throw new RuntimeException( String.format( "The file header does not contain a \"%s\" section.",
                                                       VcdKeywords.TIMESCALE_KEYWORD ) );
        }
        
        if ( ! scopeTypeStack.isEmpty() )
        {
            throw new RuntimeException( String.format( "One of the \"%s\" sections is missing its counterpart \"%s\" section.",
                                                       VcdKeywords.SCOPE_KEYWORD,
                                                       VcdKeywords.UPSCOPE_KEYWORD ) );
        }
    }
    

    private void readingHeaderSection ( VcdToken tok )
    {
        TokenTypeEnum tokenType = startKeywordSection( tok );

        switch ( tokenType )
        {
        case TT_DATE:
        case TT_VERSION:
        case TT_COMMENT:
            // System.out.printf( "Skipping header section %s\n", fromTokenToStr.get( tokenType ) );
            state = StateEnum.STATE_SKIPPING_UNTIL_END_KEYWORD;
            break;

        case TT_TIMESCALE:
            state = StateEnum.STATE_READING_TIMESCALE;
            break;

        case TT_SCOPE:
            state = StateEnum.STATE_EXPECTING_SCOPE_TYPE;
            break;

        case TT_UPSCOPE:

            assert scopeTypeStack.size() == scopeIdStack.size() : myutils.ErrUtils.assertionFailed();

            if ( scopeTypeStack.size() <= 0 )
                throw new RuntimeException( "Unexpected \"$upscope\"." );

            scopeTypeStack.remove( scopeTypeStack.size() - 1 );
            scopeIdStack  .remove( scopeIdStack  .size() - 1 );

            state = StateEnum.STATE_EXPECTING_END_KEYWORD;
            break;

        case TT_VAR:
            assert scopeTypeStack.size() == scopeIdStack.size() : myutils.ErrUtils.assertionFailed();

            if ( scopeTypeStack.size() <= 0 )
                throw new RuntimeException( "Variable definition found outside a \"$scope\" block." );

            state = StateEnum.STATE_EXPECTING_VAR_TYPE;
            break;

        case TT_ENDDEFINITIONS:
            state = StateEnum.STATE_EXPECTING_END_OF_ENDDEFINITIONS;
            break;

        default:
            // Note that startKeywordSection() has already handled the TT_NORMAL_TOKEN case. 
            throw new RuntimeException( String.format( "Keyword \"%s\" is illegal at this position.",
                                                       tok.getTokenStr() ) );
        }
    }


    private TokenTypeEnum startKeywordSection ( VcdToken tok )
    {
        TokenTypeEnum tokenType = VcdKeywords.lookupToken( tok.getTokenStr() );
        
        if ( tokenType == TokenTypeEnum.TT_NORMAL_TOKEN )
        {
            throw new RuntimeException( String.format( "Unexpected literal \"%s\" when looking for a '$' keyword.",
                                                       tok.getTokenStr() ) );
        }

        if ( tokenType == TokenTypeEnum.TT_UNKNOWN_KEYWORD )
        {
            throw new RuntimeException( String.format( "Unknown keyword \"%s\".", tok.getTokenStr() ) );
        }

        if ( VcdKeywords.isExtendedVcdKeyword( tokenType ) )
        {
            throw new RuntimeException(
                    String.format( "Keyword \"%s\" is only used in the Extended VCD format, which is not supported yet.",
                                   tok.getTokenStr() ) );
        }

        lastSectionStartToken = new VcdToken( tok );

        return tokenType;
    }


    private boolean readingValueChangeSection ( VcdToken tok, VcdTimePeriod ts )
    {
        if ( readVariableChange( tok ) )
            return false;

        assert tok.getTokenStr().length() >= 1 : myutils.ErrUtils.assertionFailed();

        char firstChar = tok.getTokenStr().charAt( 0 );

        switch ( firstChar )
        {
        case '#':
            long prev = lastTimestampFound;
            
            readTimestamp( tok );

            // When a new timestamp is found, the time period reported
            // covers the span between the last timestamp found and this new one.
            ts.set( prev, lastTimestampFound );
            
            return true;

        case '$':
            readValueChangeSectionKeyword( tok );
            return false;

        default:
            throw new RuntimeException( String.format( "Invalid literal \"%s\".",
                                                       tok.getTokenStr() ) );
        }
    }


    private boolean readVariableChange ( VcdToken tok )
    {
        assert tok.getTokenStr().length() >= 1 : myutils.ErrUtils.assertionFailed();

        char firstChar = tok.getTokenStr().charAt( 0 );

        switch ( firstChar )
        {
        case '0':
        case '1':
        case 'x':
        case 'z':
            // System.out.printf( "Read variable change: %s\n", tok.tokenStr );
            vcdVarManager.updateScalarVarValue( firstChar, tok.getTokenStr().substring( 1 ) );
            return true;

        case 'b':
        case 'r':
            readVectorVariableChange( tok );
            return true;

        default:
            return false;
        }
    }


    private void readTimestamp ( VcdToken tok )
    {
        String timestampStr = tok.getTokenStr().substring( 1 );

        if ( timestampStr.isEmpty() )
        {
            throw new RuntimeException( String.format( "Invalid timestamp token \"%s\".", tok.getTokenStr() ) );
        }

        long prev = lastTimestampFound;
        
        lastTimestampFound = myutils.Misc.parsePositiveIntegerNl( timestampStr, "Error reading the timestamp value: " );

        if ( prev != VcdTimePeriod.TIMESTAMP_INITIAL_VALUES && lastTimestampFound <= prev )
        {
            throw new RuntimeException(
                    String.format( "Invalid timestamp value of %d, the time has not moved forward since the last value of %d.",
                                   lastTimestampFound,
                                   prev ) );
        }
    }


    private void readVectorVariableChange ( VcdToken tok )
    {
        // System.out.printf( "Read vector data: %s\n", tok.tokenStr );
        
        newValueForVcdVectorVar = tok.getTokenStr();
        
        state = StateEnum.STATE_EXPECTING_VECTOR_ID;
    }


    private void readVectorId ( VcdToken tok )
    {
        // System.out.printf( "Read vector ID: %s\n", tok.tokenStr );
        
        vcdVarManager.updateVectorVarValue( newValueForVcdVectorVar, tok.getTokenStr() );

        if ( isInsideReadingDataVarDumpSection )
        {
            state = StateEnum.STATE_READING_DATA_INSIDE_DUMP_SECTION;
        }
        else
        {
            state = StateEnum.STATE_NORMAL;
        }
    }


    private void readValueChangeSectionKeyword ( VcdToken tok )
    {
        TokenTypeEnum tokenType = startKeywordSection( tok );

        switch ( tokenType )
        {
        // I've seen a $comment section once in the value change section.
        // therefore it is allowed here.
        case TT_COMMENT:
            state = StateEnum.STATE_SKIPPING_UNTIL_END_KEYWORD;
            break;
        
        case TT_DUMPALL:
        case TT_DUMPOFF:
        case TT_DUMPON:
        case TT_DUMPVARS:
            isInsideReadingDataVarDumpSection = true;
            state = StateEnum.STATE_READING_DATA_INSIDE_DUMP_SECTION;
            break;
            
        default:
            // Note that startKeywordSection() has already handled the TT_NORMAL_TOKEN case. 
            throw new RuntimeException( String.format( "Keyword \"%s\" is illegal at this position.",
                                                       tok.getTokenStr() ) );
        }
    }


    private void readVarDump ( VcdToken tok )
    {
        if ( readVariableChange( tok ) )
            return;

        if ( !tok.isEndKeyword() )
        {
            throw new RuntimeException(
                    String.format( "Unexpected token \"%s\" when looking for the \"%s\" keyword of section \"%s\" at line %d.",
                                   tok.getTokenStr(),
                                   VcdKeywords.END_KEYWORD,
                                   lastSectionStartToken.getTokenStr(),
                                   lastSectionStartToken.getLineNumber() ) );
        }

        isInsideReadingDataVarDumpSection = false;
        state = StateEnum.STATE_NORMAL;
    }


    private void skipUntilEnd ( VcdToken tok )
    {
        TokenTypeEnum tokenType = VcdKeywords.lookupToken( tok.getTokenStr() );

        if ( tokenType == TokenTypeEnum.TT_END )
        {
            state = StateEnum.STATE_NORMAL;
        }
    }


    private void expectingEnd ( VcdToken tok )
    {
        TokenTypeEnum tokenType = VcdKeywords.lookupToken( tok.getTokenStr() );

        if ( tokenType != TokenTypeEnum.TT_END )
        {
            throw new RuntimeException(
                    String.format( "Unexpected token \"%s\" when looking for the \"%s\" keyword of section \"%s\" at line %d.",
                                   tok.getTokenStr(),
                                   VcdKeywords.END_KEYWORD,
                                   lastSectionStartToken.getTokenStr(),
                                   lastSectionStartToken.getLineNumber() ) );
        }
    }


    private void checkNotEndOfFile ( VcdToken tok )
    {
        if ( tok.isEndOfFile() )
        {
            throw new RuntimeException(
                    String.format( "Unexpected end of file while reading section \"%s\" at line %d.",
                                   lastSectionStartToken.getTokenStr(),
                                   lastSectionStartToken.getLineNumber() ) );
        }

        assert !tok.getTokenStr().isEmpty() : myutils.ErrUtils.assertionFailed();
    }


    private void readTimescale ( VcdToken tok )
    {
        // Sometimes the timescale is given in a single token like "10ms", sometimes
        // in two tokens like "10 ms".
        
        int endIndex;
        for ( endIndex = 0; endIndex < tok.getTokenStr().length(); ++endIndex )
        {
            if ( !Character.isDigit( tok.getTokenStr().charAt( endIndex ) ) )
                break;
        }

        timescale = new VcdTimescale();
        
        timescale.parsePeriod( tok.getTokenStr().substring( 0, endIndex ) );

        String rest = tok.getTokenStr().substring( endIndex, tok.getTokenStr().length() );

        // System.out.printf( "Timescale rest: %s\n", rest );

        if ( rest.isEmpty() )
        {
            state = StateEnum.STATE_EXPECTING_TIMESCALE_UNIT;
        }
        else
        {
            timescale.parseUnit( rest );
            state = StateEnum.STATE_EXPECTING_END_KEYWORD;
        }
    }

    private void readTimescaleUnit ( VcdToken tok )
    {
        timescale.parseUnit( tok.getTokenStr() );
        
        state = StateEnum.STATE_EXPECTING_END_KEYWORD;
    }


    private void readScopeType ( VcdToken tok )
    {
        // The scope type can be: begin | fork | function | module | task
        //
        // At the moment, the scope type is not validated, any string is accepted.

        scopeTypeStack.add( tok.getTokenStr() );

        state = StateEnum.STATE_EXPECTING_SCOPE_ID;
    }


    private void readScopeId ( VcdToken tok )
    {
        scopeIdStack.add( tok.getTokenStr() );

        // System.out.printf( "Entering scope type: %s, id: %s\n",
        // scopeTypeStack.get( scopeTypeStack.size() - 1 ),
        // scopeIdStack .get( scopeTypeStack.size() - 1 ) );

        state = StateEnum.STATE_EXPECTING_END_KEYWORD;
    }


    private void readVarType ( VcdToken tok )
    {
        vcdVarDefinition = new VcdVariableManager.VcdVariable();

        // var_type = event | integer | parameter | real | reg |
        //            supply0 | supply1 | time | tri | triand |
        //            trior | trireg | tri0 | tri1 | wand | wire | wor
        //
        // At the moment, the variable type is not validated, any string is accepted.

        vcdVarDefinition.type = tok.getTokenStr();

        state = StateEnum.STATE_EXPECTING_VAR_SIZE;
    }


    private void readVarSize ( VcdToken tok )
    {
        if ( !myutils.Misc.isPositiveInteger( tok.getTokenStr() ) )
        {
            throw new RuntimeException( String.format( "Invalid variable size value \"%s\".", tok.getTokenStr() ) );
        }

        vcdVarDefinition.size = myutils.Misc.parsePositiveIntegerNl( tok.getTokenStr(),
                                                                     "Error reading the variable size: " );

        if ( vcdVarDefinition.size == 0 )
        {
            throw new RuntimeException( String.format( "Invalid variable size value of %d.",
                                                       vcdVarDefinition.size ) );
        }

        state = StateEnum.STATE_EXPECTING_VAR_ID;
    }


    private void readVarId ( VcdToken tok )
    {
        for ( int i = 0; i < tok.getTokenStr().length(); ++i )
        {
            char c = tok.getTokenStr().charAt( i );

            if ( c < 33 || c > 126 )
            {
                throw new RuntimeException(
                        String.format( "Invalid variable identifier \"%s\", character '%c' it outside the allowed range.",
                                       tok.getTokenStr(),
                                       c ) );
            }
        }

        vcdVarDefinitionId = tok.getTokenStr();

        state = StateEnum.STATE_EXPECTING_VAR_REF;
    }


    private void readVarRef ( VcdToken tok )
    {
        // Sometimes a vector variable reference is given in one token, sometimes in two:
        //    myvector[31:0]
        //    myvector [31:0]
        
        // Split the token into identifier and index range (if present).
        assert tok.getTokenStr().length() >= 1 : myutils.ErrUtils.assertionFailed();

        int bracketPos;
        for ( bracketPos = 0; bracketPos < tok.getTokenStr().length(); ++bracketPos )
        {
            if ( tok.getTokenStr().charAt( bracketPos ) == '[' )
                break;
        }

        if ( bracketPos == 0 )
        {
            throw new RuntimeException( String.format( "Invalid variable reference \"%s\".", 
                                                       vcdVarDefinition.ref ) );
        }

        vcdVarDefinition.ref = tok.getTokenStr().substring( 0, bracketPos );
        vcdVarDefinition.indexRange = tok.getTokenStr().substring( bracketPos, tok.getTokenStr().length() );

        // System.out.printf( "Pass 1: Variable reference: \"%s\", optional range: \"%s\"\n",
        //                    vcdVarDefinition.ref,
        //                    vcdVarDefinition.indexRange );

        state = StateEnum.STATE_EXPECTING_VAR_INDEX_RANGE;
    }


    private void readVarIndexRange ( VcdToken tok )
    {
        if ( tok.isEndKeyword() )
        {
            state = StateEnum.STATE_NORMAL;
        }
        else
        {
            if ( !vcdVarDefinition.indexRange.isEmpty() )
                throw new RuntimeException( String.format( "Second variable index range \"%s\" found.", tok.getTokenStr() ) );

            vcdVarDefinition.indexRange = tok.getTokenStr();

            // System.out.printf( "Pass 2: Variable reference: \"%s\", optional range: \"%s\"\n",
            //                    vcdVarDefinition.ref,
            //                    vcdVarDefinition.indexRange );

            state = StateEnum.STATE_EXPECTING_END_KEYWORD;
        }

        StringBuilder fullRef = new StringBuilder();

        for ( int i = 0; i < scopeIdStack.size(); ++i )
        {
            fullRef.append( scopeIdStack.get( i ) );
            fullRef.append( VcdVariableManager.VAR_NAME_SEP );
        }

        fullRef.append( vcdVarDefinition.ref );
        fullRef.append( vcdVarDefinition.indexRange );

        vcdVarDefinition.fullRef = fullRef.toString();
        vcdVarManager.addVcdVariable( vcdVarDefinition, vcdVarDefinitionId );
        
        // The Variable Manager takes ownership, we don't need this variable here any more.
        vcdVarDefinition = null;
        vcdVarDefinitionId = "";  // Helps catch bugs.
    }

    
    public final VcdVariableManager getVarManager ()
    {
        return vcdVarManager;
    }
    
    public final String getFilenameForUserMessages ()
    {
        return filenameForUserMessages;
    }
    
    public final VcdTimescale getTimescale ()
    {
        return timescale;
    }

    @Override
    public void close () throws IOException
    {
        if ( br != null )
            br.close();
    }
}
