/**
 * @author R. Diez
 */

package vcdEngine;

import java.util.EnumMap;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;

class VcdKeywords
{
    public static final String END_KEYWORD       = "$end"; 
    public static final String TIMESCALE_KEYWORD = "$timescale"; 
    public static final String SCOPE_KEYWORD     = "$scope"; 
    public static final String UPSCOPE_KEYWORD   = "$upscope"; 
    
    public enum TokenTypeEnum
    {
    	// Header keywords.
        TT_DATE,
        TT_VERSION,
        TT_TIMESCALE,
        TT_SCOPE,
        TT_VAR,
        TT_UPSCOPE,
        TT_ENDDEFINITIONS, 
        
        // Dump keywords.
        TT_DUMPALL,
        TT_DUMPOFF,
        TT_DUMPON,
        TT_DUMPVARS,

        // Dumpports keywords, for Extended VCD files.
        TT_DUMPPORTS,
        TT_DUMPPORTSOFF,
        TT_DUMPPORTSON,
        TT_DUMPPORTSALL,
        TT_DUMPPORTSLIMIT,  // Probably never shows up in a VCD file, only in Verilog source files.
        TT_DUMPPORTSFLUSH,  // Probably never shows up in a VCD file, only in Verilog source files.
        
        // Other token types.
        TT_END,
        TT_COMMENT,
        
        TT_UNKNOWN_KEYWORD,  // The token does start with a '$' character, but is unknown.
        TT_NORMAL_TOKEN,     // Something other than a $keyword .
    };


    private final static EnumMap< TokenTypeEnum, String > fromTokenToStr = InitFromTokenToStr();
    private final static HashMap< String, TokenTypeEnum > fromStrToToken = InitFromStrToToken();
    
    
    private static EnumMap< TokenTypeEnum, String > InitFromTokenToStr ()
    {
        // Actually, we could write all this as a switch statement.
    	// It does not really matter, as the look-up speed from enum to string
    	// is not critical.
        
        EnumMap< TokenTypeEnum, String > m = new EnumMap< TokenTypeEnum, String >( TokenTypeEnum.class );

        m.put( TokenTypeEnum.TT_VAR,            "$var" );
        m.put( TokenTypeEnum.TT_END,            END_KEYWORD );
        m.put( TokenTypeEnum.TT_SCOPE,          SCOPE_KEYWORD );
        m.put( TokenTypeEnum.TT_UPSCOPE,        UPSCOPE_KEYWORD );
        m.put( TokenTypeEnum.TT_COMMENT,        "$comment" );
        m.put( TokenTypeEnum.TT_DATE,           "$date" );
        m.put( TokenTypeEnum.TT_DUMPALL,        "$dumpall" );
        m.put( TokenTypeEnum.TT_DUMPOFF,        "$dumpoff" );
        m.put( TokenTypeEnum.TT_DUMPON,         "$dumpon" );
        m.put( TokenTypeEnum.TT_DUMPVARS,       "$dumpvars" );
        m.put( TokenTypeEnum.TT_ENDDEFINITIONS, "$enddefinitions" );
        m.put( TokenTypeEnum.TT_DUMPPORTS,      "$dumpports" );
        m.put( TokenTypeEnum.TT_DUMPPORTSOFF,   "$dumpportsoff" );
        m.put( TokenTypeEnum.TT_DUMPPORTSON,    "$dumpportson" );
        m.put( TokenTypeEnum.TT_DUMPPORTSALL,   "$dumpportsall" );
        m.put( TokenTypeEnum.TT_DUMPPORTSLIMIT, "$dumpportslimit" );
        m.put( TokenTypeEnum.TT_DUMPPORTSFLUSH, "$dumpportsflush" );
        m.put( TokenTypeEnum.TT_TIMESCALE,      TIMESCALE_KEYWORD );
        m.put( TokenTypeEnum.TT_VERSION,        "$version" );

        assert TokenTypeEnum.values().length - 2 == m.size() : myutils.ErrUtils.assertionFailed();
        
        return m;
    }

    
    private static HashMap< String, TokenTypeEnum > InitFromStrToToken ()
    {
        assert fromTokenToStr.size() > 0 : myutils.ErrUtils.assertionFailed();
        
        // Favour speed over memory in this case. Maybe a factor of 10 is too big, it could
        // even cause more cache misses and slow things down. Unfortunately, there is no convenient
        // way to check how full the buckets are in a given hash object, you have to manually
        // find out by calling String.hashCode() on all the strings and checking yourself
        // if there are many collisions.
        // With a default hash, I've seen that 3 of the $dump commands land on the same bucket,
        // and that's not ideal. I just looked with the debugger, and there are no collisions
        // with Java 6 and a factor of 10.
        // Not that all this really matters much, as a VCD file does not have many keywords.
        // If it did, it may be worth implementing this as a perfect hash.
        int initialCapacity = fromTokenToStr.size() * 10;
        
        HashMap< String, TokenTypeEnum > m = new HashMap< String, TokenTypeEnum >( initialCapacity );
        
        Set<TokenTypeEnum> values = fromTokenToStr.keySet();
        Iterator<TokenTypeEnum> it = values.iterator();
        
        while ( it.hasNext() )
        {        
            TokenTypeEnum k = it.next();
            m.put( fromTokenToStr.get(k), k );
        }
       
        assert fromTokenToStr.size() == m.size() : myutils.ErrUtils.assertionFailed();
        
        return m;
    }
    

    public static TokenTypeEnum lookupToken ( String tokenStr )
    {
        assert ! tokenStr.isEmpty() : myutils.ErrUtils.assertionFailed();
            
        if ( tokenStr.charAt(0) != '$' )
            return TokenTypeEnum.TT_NORMAL_TOKEN;

        
        TokenTypeEnum tt = fromStrToToken.get( tokenStr );
        
        if ( tt == null )
            return TokenTypeEnum.TT_UNKNOWN_KEYWORD;
        
        
        return tt;

    }
    
    
    public static boolean isExtendedVcdKeyword ( TokenTypeEnum tokenType )
    {
        switch ( tokenType )
        {
        case TT_DUMPPORTS:
        case TT_DUMPPORTSOFF:
        case TT_DUMPPORTSON:
        case TT_DUMPPORTSALL:
        case TT_DUMPPORTSLIMIT:
        case TT_DUMPPORTSFLUSH:
            return true;
           
        default:
            return false;
        }
    }
}
