/**
 * @author R. Diez
 */

package vcdEngine;

public class VcdTimePeriod
{
    public static final long TIMESTAMP_ZERO           =  0;  // The minimum non-special timestamp value.
    public static final long TIMESTAMP_INVALID        = -1;
    public static final long TIMESTAMP_INITIAL_VALUES = -2;  // Only valid for currentTimestamp.
    public static final long TIMESTAMP_END_OF_FILE    = -3;  // Only valid for nextTimestamp.
    
    private long currentTimestamp;  // -1 means initial values before the first timestamp, which is often zero.
    private long nextTimestamp;     // -1 means that the end of the file has been reached.
    
    
    public VcdTimePeriod ()
    {
        reset();
    }
    
    void reset ()
    {
        currentTimestamp = TIMESTAMP_INVALID;
        nextTimestamp    = TIMESTAMP_INVALID;
    }
    
    public boolean isReset ()
    {
        return currentTimestamp == TIMESTAMP_INVALID &&
               nextTimestamp    == TIMESTAMP_INVALID;
    }
    
    void set ( long currentTimestamp, long nextTimestamp )
    {
    	this.currentTimestamp = currentTimestamp;
    	this.nextTimestamp    = nextTimestamp;
    	
    	assert isValid() : myutils.ErrUtils.assertionFailed();
    }
    
    public long getCurrentTimestamp()
    {
		return currentTimestamp;
	}

	public long getNextTimestamp()
	{
		return nextTimestamp;
	}

	
	public boolean isValid ()
    {
        //  There are 4 valid combinations:
        //    a)  ( TIMESTAMP_INITIAL_VALUES, TIMESTAMP_END_OF_FILE )  [there are no timestamps at all in the VCD file]
        //    b)  ( TIMESTAMP_INITIAL_VALUES, some val >= 0 )
        //    c)  ( some val  >= 0, TIMESTAMP_END_OF_FILE )
        //    d)  ( some val1 >= 0, some val2 > val1 )
        
        
        // Check if we forgot to set one of the fields.
        
        if ( currentTimestamp == TIMESTAMP_INVALID )
            return false;
        
        if ( nextTimestamp == TIMESTAMP_INVALID )
            return false;
        
        
        // No negative timestamp values are allowed, except for one specific exception
        // for each field.
        
        if ( currentTimestamp != TIMESTAMP_INITIAL_VALUES &&
             currentTimestamp < TIMESTAMP_ZERO )
        {
            return false;
        }
        
        if ( nextTimestamp != TIMESTAMP_END_OF_FILE &&
             nextTimestamp < TIMESTAMP_ZERO )
        {
            return false;
        }
        
        // For "normal" timestamps, the next time should be greater than the previous time.
        // The VCD file parser does check that the time always moves forwards.

        if ( currentTimestamp != TIMESTAMP_INITIAL_VALUES &&
             nextTimestamp    != TIMESTAMP_END_OF_FILE &&
             currentTimestamp >= nextTimestamp )
        {
            return false;
        }
        
        return true;
    }

    
    public String format ( VcdTimescale.UNIT_ENUM unit )
    {
        return String.format( "[current: %s, next: %s]",
                              formatTimestamp( currentTimestamp, unit ),
                              formatTimestamp( nextTimestamp, unit ) );
    }

    
    public static String formatTimestamp ( long timestamp, VcdTimescale.UNIT_ENUM unit )
    {
        // Java cannot switch() on a type 'long', thus all the if's here.
        
        if ( timestamp == TIMESTAMP_INVALID )
        {
        	assert false : myutils.ErrUtils.assertionFailed();
            return "<invalid>";
        }
            
        if ( timestamp == TIMESTAMP_INITIAL_VALUES )
            return "<initialisation>";
            
        if ( timestamp == TIMESTAMP_END_OF_FILE )
            return "<end of file>";
            
        if ( timestamp < 0 )
        {
        	assert false : myutils.ErrUtils.assertionFailed();
            return "<negative>";
        }
        
        return String.format( "%d %s", timestamp, VcdTimescale.formatUnit( unit ) ); 
    }
}
