/**
 * @author R. Diez
 */

package vcdEngine;

import java.lang.ref.WeakReference;
import java.util.AbstractMap;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;


public class VcdVariableManager
{
	// See the help text for an explanation about why I've chosen a period as separator.
    public static final String VAR_NAME_SEP = ".";

    
    public static class VcdVariable
    {
        String type;
        int    size;
        String ref;
        String indexRange;

        // Full reference name, like "module1.submodule2.variableName[31:0]".
        // At the moment, no duplicates are allowed in a VCD file.
        String fullRef;
        
        public String getFullRef ()  {  return fullRef;  }

        // Note that several variable definitions may share a single value ID.
        // The map further below is the master owner, use a weak reference to prevent
        // circular references and help the garbage collector.
        private WeakReference< VcdValue > value;
        
        public String getValue ()  {  return value.get().value;  }
        
        public String formatValue ()
        {
            String v = value.get().value;

            if ( v.length() <= 1 )
                return v;
            
            if ( v.equals( UNINITIALISED_MARKER ) )
                return v;
            
            if ( ! myutils.Misc.isPositiveInteger( v  ) )
                return v;
            
            if ( v.length() >= Long.SIZE - 1 )
                return v;
            
            long valueAsInteger = Long.parseLong( v, 2 );

            return String.format( "0b%s (0x%X, %d)", v, valueAsInteger, valueAsInteger );
        }
    }
    
    
    private static class VcdValue
    {
        private String id;
        private String value;
        
        // The map further below is the master owner, use a weak reference to prevent
        // circular references and help the garbage collector.
        ArrayList< WeakReference< VcdVariable > > variables = new ArrayList< WeakReference< VcdVariable > > ();
    }
    

    // The choice between maps and sets:
    //   The maps below are convenient, as they already know how to sort strings,
    //   so you don't need to supply a comparator function, but it would save
    //   a little space if there were sets.
    //
    // The choice betwen trees or hashes:
    //   Hashes would be generally faster when looking up, but maps don't give
    //   you performance surprises depending on your luck with the hash function.
    //   Besides, we do use the tree ordering when printing the variable list.
    //
    // Possible performance optimisation:
    //   The look-up speed from ID to VcdVariable is important. A possible optimisation would be
    //   to build an array-based look-up table for IDs up to 2 characters long (the most common IDs).
    //   That would be a table with less than 64 K entries. For any longer IDs, 
    //   a perfect hash (or a tree if that fails to build) should suffice.
    private TreeMap<String, VcdVariable> allVariablesByFullRef  = new TreeMap< String, VcdVariable >();
    private TreeMap<String, VcdValue   > allValuesById          = new TreeMap< String, VcdValue    >();

    private static final boolean traceVarChanges = false;
    private static final String LIST_INDENTATION = "  ";
    
    // I was tempted to use the British spelling "uninitialised", but I guess
    // most people would expect the American one in the program's output.  8-)
    private static final String UNINITIALISED_MARKER = "<uninitialized>";


    // This routine takes ownership of the vcdVar object, so the caller
    // should not be reusing or modifying it further.
    
    void addVcdVariable ( VcdVariable vcdVar, String id )
    {
        assert ! id.isEmpty() : myutils.ErrUtils.assertionFailed();
        
        // This should be an atomic operations: insert on both data structures, or leave them
        // like there were on entry.
        // The only scenario we cannot cope with is an out of memory situation. We wouldn't
        // be able to reinsert the old element then in order to leave the maps like there were on entry.
        boolean insertedVariable = false;
        boolean insertedValue    = false;
        VcdVariable prevVariable = null;
        VcdValue    prevValue    = null;
        
        try
        {
            // Step 1: Insert a new VcdValue.
            
            VcdValue newValue = new VcdValue();
            newValue.id       = id;
            newValue.value    = UNINITIALISED_MARKER;          
            
            prevValue = allValuesById.put( id, newValue );
            insertedValue = true;

            if ( prevValue != null )
            {
                assert prevValue.id.equals( id ) : myutils.ErrUtils.assertionFailed();  // Paranoia.
                assert prevValue.value.equals( UNINITIALISED_MARKER ) : myutils.ErrUtils.assertionFailed();
                
                // Note that we leave prevValue untouched, in case we need to roll the insertion back.
                
                newValue.variables.addAll( prevValue.variables );
            }
            
            newValue.variables.add( new WeakReference< VcdVariable >( vcdVar ) );
            
            
            // Step 2: Insert the VcdVariable.
            
            assert !vcdVar.fullRef.isEmpty() : myutils.ErrUtils.assertionFailed();
            assert vcdVar.value == null : myutils.ErrUtils.assertionFailed();
            
            vcdVar.value = new WeakReference<VcdVariableManager.VcdValue>( newValue );
            
            prevVariable = allVariablesByFullRef.put( vcdVar.fullRef, vcdVar );
            insertedVariable = true;

            if ( prevVariable != null )
            {
                throw new RuntimeException( String.format( "Duplicate variable name \"%s\".",
                                                           vcdVar.fullRef ) );
            }
        }
        catch ( Throwable ex )
        {
            rollbackMapInsertion( insertedVariable, allVariablesByFullRef, vcdVar.fullRef, prevVariable );
            rollbackMapInsertion( insertedValue   , allValuesById        , id            , prevValue    );
            
            throw myutils.ErrUtils.asRuntimeException( ex, "" );
        }

        
        // Make the old VcdVariables that were using the old VcdValue point to the new one.  
        // We do not rollback this change in case of failure, but this kind of operation
        // should never fail, barring we run out ouf memory. But we cannot cope with
        // out-of-memory scenarios anyway, see the comment above.
        
        if ( prevValue != null )
        {
           Iterator< WeakReference<VcdVariable> > it = prevValue.variables.iterator();
           
           while ( it.hasNext() )
           {
               WeakReference<VcdVariable> wref = it.next();
               wref.get().value = vcdVar.value;
           }
        }
    }
    

    // This routine assumes that a previous value of null means that the element was not there at all.
    // However, maps can hold null elements, so the caller needs to make sure that either this assumption
    // holds for its purposes, or it makes no difference that an element is not present or it has a null value.
    private static <K,V> void rollbackMapInsertion ( boolean shouldRollback,
                                                     AbstractMap<K,V> map,
                                                     K key,
                                                     V previousValue )
    {
        // This keeps the caller clean if statements, it's just visually more appealing.  8-)
        if ( !shouldRollback )
            return;
        
        V recentlyInserted;
        
        if ( previousValue == null )
        {
            recentlyInserted = map.remove( key );
        }
        else
        {
            recentlyInserted = map.put( key, previousValue );
        }

        // If we are rolling back an insertion, then there should be a previous object
        // that we just removed or replaced.
        assert recentlyInserted != null : myutils.ErrUtils.assertionFailed();
    }
    

    public void printVariableFullRefs ( String filenameForUserMessages )
    {
        System.out.printf( "Variables defined in file \"%s\":\n", filenameForUserMessages );

        Set<Entry<String, VcdVariable>> entries = allVariablesByFullRef.entrySet();

        Iterator<Entry<String, VcdVariable>> it = entries.iterator();

        if ( !it.hasNext() )
        {
            System.out.printf( "%s<no variables defined>\n", LIST_INDENTATION );
        }
        else
        {
            do
            {
                Entry<String, VcdVariable> entry = it.next();

                VcdVariable vcdVar = entry.getValue();

                // System.out.printf( "%s- Variable Name: %s, VCD ID: %s\n", margin, vcdVar.fullRef, vcdVar.id );
                System.out.printf( "%s- Variable name: %s\n", LIST_INDENTATION, vcdVar.fullRef );
            }
            while ( it.hasNext() );
        }
    }


    public void printVariableValues ( long timestamp,
    		                          VcdTimescale.UNIT_ENUM unit,
    		                          String filenameForUserMessages )
    {
        System.out.printf( "Variables values at timestamp %s for file \"%s\":\n",
                           VcdTimePeriod.formatTimestamp( timestamp, unit ),
                           filenameForUserMessages );

        Set<Entry<String, VcdVariable>> entries = allVariablesByFullRef.entrySet();

        Iterator<Entry<String, VcdVariable>> it = entries.iterator();

        if ( !it.hasNext() )
        {
            System.out.printf( "%s<no variables defined>\n", LIST_INDENTATION );
        }
        else
        {
            do
            {
                Entry<String, VcdVariable> entry = it.next();

                VcdVariable vcdVar = entry.getValue();

                System.out.printf( "%s- Variable name: %s, value: %s\n",
                		           LIST_INDENTATION, 
                		           vcdVar.fullRef,
                		           vcdVar.formatValue() );
            }
            while ( it.hasNext() );
        }
    }
    
    
    void updateScalarVarValue ( char newValue, String vcdVarId )
    {
        switch ( newValue )
        {
        case '0':
        case '1':
        case 'x':
        case 'z':
            break;

        default:
            throw new RuntimeException( String.format( "Invalid scalar variable value '%c'.", newValue ) );
        }

        VcdValue v = allValuesById.get( vcdVarId );

        if ( v == null )
        {
            throw new RuntimeException( String.format( "Unknown variable ID \"%s\".", vcdVarId ) );
        }
        
        v.value = Character.toString( newValue );

        if ( traceVarChanges )
        {
            for ( int i = 0; i < v.variables.size(); ++i )
            {
                VcdVariable var = v.variables.get( i ).get();
                
                System.out.printf( "Variable %s has changed to %s\n",
                                   var.fullRef,
                                   var.formatValue() );
            }
        }
    }


    void updateVectorVarValue ( String newValue, String vcdVarId )
    {
        assert newValue.length() >= 1 : myutils.ErrUtils.assertionFailed();

        char valueType = newValue.charAt( 0 );

        switch ( valueType )
        {
        case 'b':
            break;

        case 'r':
            throw new RuntimeException( "Variable values of type real are not supported yet." );

        default:
            throw new RuntimeException( String.format( "Invalid vector variable value type '%c'.", valueType ) );
        }

        VcdValue  v = allValuesById.get( vcdVarId );

        if ( v == null )
        {
            throw new RuntimeException( String.format( "Unknown variable ID \"%s\".", vcdVarId ) );
        }

        String valWithoutType = newValue.substring( 1 );

        for ( int i = 0; i < valWithoutType.length(); ++i )
        {
            char c = valWithoutType.charAt( i );

            switch ( c )
            {
            case '0':
            case '1':
                break;

            case 'x':
            case 'z':
                break;

            default:
                throw new RuntimeException( String.format( "Invalid vector variable value \"%s\".", valWithoutType ) );
            }
        }

        v.value = valWithoutType;
        
        if ( traceVarChanges )
        {
            for ( int i = 0; i < v.variables.size(); ++i )
            {
                VcdVariable var = v.variables.get( i ).get();
                System.out.printf( "Variable %s has changed to %s\n",
                                   var.fullRef,
                                   var.formatValue() );
            }
        }
    }


    public boolean isEmpty ()
    {
        // Either both maps should be empty, or none of them.
        assert ( allVariablesByFullRef.size() == 0 ) ==
               ( allValuesById        .size() == 0 )
               : myutils.ErrUtils.assertionFailed();
        
        return allVariablesByFullRef.size() == 0;
    }

    
    public final Set<String> getAllVarFullRefs ()
    {
        return allVariablesByFullRef.keySet();        
    }
    
    
    // Returns null if the variable is not found. 
    
    public final VcdVariable getVarByFullRef ( String varFullRef )
    {
        return allVariablesByFullRef.get( varFullRef );
    }

}
